/*
 *	B I N D . C
 *
 *	purpose:
 *		This file is for functions having to do with key bindings,
 *		descriptions and Help and Ascii commands.
 *
 *	to do:
 *		In CmdStr() - modify SPEC code to actually display the key pressed,
 *		not just some FNx type of display (which is completely meaningless).
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

#define	FOUND		0		/* instead of using SUCCESS */

#define PATHCHR ';'			/* PATH entries seperator */

/*
 * local function prototypes
 */

static char	* near	getfname(int (*func)());
static int near		unbindchar(int c);
static ushort near	getckey(int);
static ushort near	stock(char *);

/**
 * Bring up a fake buffer and read the Help file into it with view mode
 */
int Help(f, n)
  int		f, n;
{
  WINDOW	*wp; 		/* scanning pointer to windows */
  BUFFER	*bp; 		/* buffer pointer to Help */
  char		*fname;		/* ptr to file returned by FileLook() */

	/* first check if we are already here */
#ifdef EKBD
	bp = FindBuf("ce.hlp", FALSE, BFINVS);
#endif
#ifdef OKBD
	bp = FindBuf("ceo.hlp", FALSE, BFINVS);
#endif
	if (bp == NULLBPTR)
	{
#ifdef EKBD
		fname = FileLook("ce.hlp", TRUE);
#endif
#ifdef OKBD
		fname = FileLook("ceo.hlp", TRUE);
#endif
		if (fname == NULLCPTR)
		{
			MlWrite("Can't find Help file");
			return(FALSE);
		}
	}

	/* split the current window to make room for the Help stuff */
	if (SplitWind(FALSE, 1) == FALSE)
		return(FALSE);

	if (bp == NULLBPTR)
	{
		/* and read it in */
		if (GetFile(fname) == FALSE)
			return(FALSE);
	}
	else
		SwapBuffer(bp);

	/* put this window in VIEW mode; Update all mode lines */
	curwp->w_bufp->b_mode |= MDVIEW;
	curwp->w_bufp->b_flag |= BFINVS;
	UpMode();

	return(TRUE);
}

/**
 * Bring up a fake buffer and read the ASCII table file into it with view mode
 */
int AsciiTable(f, n)
  int		f, n;
{
  WINDOW	*wp; 		/* scanning pointer to windows */
  BUFFER	*bp; 		/* buffer pointer to Ascii table */
  char		*fname;		/* ptr to file returned by FileLook() */

	/* first check if we are already here */
	bp = FindBuf("ascii.tbl", FALSE, BFINVS);
	if (bp == NULLBPTR)
	{
		fname = FileLook("ascii.tbl", TRUE);

		if (fname == NULLCPTR)
		{
			MlWrite("Can't find ASCII table file");
			return(FALSE);
		}
	}

	/* split the current window to make room for the Help stuff */
	if (SplitWind(FALSE, 1) == FALSE)
		return(FALSE);

	if (bp == NULLBPTR)
	{
		/* and read it in */
		if (GetFile(fname) == FALSE)
			return(FALSE);
	}
	else
		SwapBuffer(bp);

	/* put this window in VIEW mode; Update all mode lines */
	curwp->w_bufp->b_mode |= MDVIEW;
	curwp->w_bufp->b_flag |= BFINVS;
	UpMode();

	return(TRUE);
}

/**
 * Describe the command for a certain key
 */
int DesKey(f, n)
  int	f, n;
{
  int	c;				/* key to describe */
  char	*ptr;			/* string pointer to scan output strings */
  char	outseq[NSTRING];/* output buffer for command sequence */

	/* prompt the user to type us a key to describe */
	MlWrite(": describe-key : ");

	/* get command sequence to describe, change it to something we can print */
	CmdStr(c = getckey(FALSE), &outseq[0]);

	/* and dump it out */
	OString(outseq);
	OString(" : ");

	/* find the right function pointer */
	if ((ptr = getfname(GetBind(c))) == NULLCPTR)
		ptr = "Key not bound";

	/* output the command sequence */
	OString(ptr);
	return(TRUE);
}

/**
 * Add a new key to the key binding table
 */
int BindToKey(f, n)
  int		f, n;
{
  ushort	c;					/* command key to bind */
  int		(*kfunc)();			/* ptr to requested function to bind to */
  KEYTAB	*ktp;				/* pointer into the command table */
  int		found;				/* matched command flag */
  char		outseq[NSTRING];	/* output buffer for keystroke sequence */

	/* prompt the user to type in a key to bind */
	MlWrite(": bind-to-key ");

	/* get the function name to bind it to */
	kfunc = GetName();
	if (kfunc == NULLFPTR)
	{
		MlWrite("No such function");
		return(FALSE);
	}
	OString(" ");

	/* get the command sequence to bind */
	c = getckey(kfunc == Meta || kfunc == Cex || kfunc == CmdAbort);

	/* change it to something we can print as well */
	CmdStr(c, &outseq[0]);

	/* and dump it out */
	OString(outseq);

	/* if the function is a prefix key */
	if (kfunc == Meta || kfunc == Cex || kfunc == CmdAbort)
	{
		/* search for an existing binding for the prefix key */
		ktp = &keytab[0];
		while (ktp->k_fp != NULLFPTR)
		{
			if (ktp->k_fp == kfunc)
				unbindchar(ktp->k_code);
			++ktp;
		}

		/* reset the appropriate global prefix variable */
		if (kfunc == Meta)
			metac = c;
		else if (kfunc == Cex)
			ctlxc = c;
		else if (kfunc == CmdAbort)
			abortc = c;
	}

	/* search the table to see if it exists */
	ktp = &keytab[0];
	found = FALSE;
	while (ktp->k_fp != NULLFPTR)
	{
		if (ktp->k_code == c)
		{
			found = TRUE;
			break;
		}
		++ktp;
	}

	if (found)				/* it exists, just change it then */
	{
		ktp->k_fp = kfunc;
 	}
	else					/* otherwise we need to add it to the end */
	{
		/* if we run out of binding room, bitch */
		if (ktp >= &keytab[NBINDS])
		{
			MlWrite("Binding table full");
			return(FALSE);
		}

		ktp->k_code = c;	/* add keycode */
		ktp->k_fp = kfunc;	/* and the function pointer */
		++ktp;				/* and make sure the next is null */
		ktp->k_code = 0;
		ktp->k_fp = NULLFPTR;
	}
	return(TRUE);
}

/**
 * Delete a key from the key binding table
 */
int UnbindKey(f, n)
  int	f, n;
{
  int	c;			/* command key to unbind */
  char	outseq[80]; /* output buffer for keystroke sequence */

	/* prompt the user to type in a key to unbind */
	MlWrite(": unbind-key ");

	/* get the command sequence to unbind */
	c = getckey(FALSE);

	/* change it to something we can print as well */
	CmdStr(c, &outseq[0]);

	/* and dump it out */
	OString(outseq);

	/* if it isn't bound, bitch */
	if (unbindchar(c) == FALSE)
	{
		MlWrite("Key not bound");
		return(FALSE);
	}
	return(TRUE);
}

/**
 * Unbind a command key
 */
static int near unbindchar(c)
  int		c;		/* command key to unbind */
{
  KEYTAB	*ktp;	/* pointer into the command table */
  KEYTAB	*sktp;	/* saved pointer into the command table */
  int		found;	/* matched command flag */

	/* search the table to see if the key exists */
	ktp = &keytab[0];
	found = FALSE;
	while (ktp->k_fp != NULLFPTR)
	{
		if (ktp->k_code == c)
		{
			found = TRUE;
			break;
		}
		++ktp;
	}

	/* if it isn't bound, bitch */
	if (!found)
		return(FALSE);

	/* save the pointer and scan to the end of the table */
	sktp = ktp;
	while (ktp->k_fp != NULLFPTR)
		++ktp;
	--ktp;					/* backup to the last legit entry */

	/* copy the last entry to the current one */
	sktp->k_code = ktp->k_code;
	sktp->k_fp = ktp->k_fp;

	/* null out the last one */
	ktp->k_code = 0;
	ktp->k_fp = NULLFPTR;
	return(TRUE);
}

/**
 * Get a command key sequence from the keyboard
 */
static ushort near getckey(mflag)
  int		mflag;			/* going for a Meta sequence? */
{
  char		tok[NSTRING];	/* command incoming */

	/* check to see if we are executing a command line */
	if (clexec)
	{
		MacArg(tok);		/* get the next Token */
		return(stock(tok));
	}

	/* or the normal way */
	return(mflag ? TermGetCh() : GetCmd());
}

/**
 * Look up the existance of a file along the normal or PATH environment
 * variable. Look first in the current directory, then look for the CODED
 * environment variable (if 'hflag' is TRUE), then try the various PATHs.
 */
char * FileLook(fname, hflag)
  char			*fname;			/* base file name to search for */
  int			hflag;			/* include CODED environment var in search? */
{
  char			envstr[NSTRING];/* environment string */
  char			*path;			/* PATH string */
  char			*sp; 			/* pointer into path spec */
  static char	fspec[NFILEN];	/* full path spec to search */

	/* always try the current directory first */
	if (FileExist(fname))
		return(fname);

	/* next, try CodEd's own environment variable */
	if (hflag)
	{
		if (syenvstr("CODED", envstr, NSTRING) == RC_SUCCESS)
		{
			/* build home dir file spec */
			strfmt(fspec, "%s\\%s", envstr, fname);

			/* and try it out */
			if (FileExist(fspec))
				return(fspec);
		}
	}

	/* if all else fails, try the various PATHs */
 	if (syenvstr("PATH", envstr, NSTRING) == RC_SUCCESS)
	{
		path = envstr;
		while (*path)
		{
			/* build next possible file spec */
			sp = fspec;
			while (*path && (*path != PATHCHR))
				*sp++ = *path++;

			/* add a terminating dir separator if we need it */
			if (sp != fspec)
				*sp++ = '\\';
			*sp = 0;
			stracat(fspec, fname);

			/* and try it out */
			if (FileExist(fspec))
				return(fspec);

			if (*path == PATHCHR)
				++path;
		}
	}
	return(NULLCPTR);	/* no such luck */
}

/**
 * Change a key command to a string we can print out
 */
void CmdStr(c, seq)
  int	c;		/* sequence to translate */
  char	*seq;	/* destination string for sequence */
{
  char	*ptr;	/* pointer into current position in sequence */

	ptr = seq;

	/* apply Meta sequence if needed */
	if (c & META)
	{
		*ptr++ = '<';
		*ptr++ = 'E';
		*ptr++ = 's';
		*ptr++ = 'c';
		*ptr++ = '>';
	}

	/* apply ^X sequence if needed */
	if (c & CTLX)
	{
		*ptr++ = 'c';
		*ptr++ = 'X';
		*ptr++ = ' ';
	}

	/* apply SPEC sequence if needed */
	if (c & SPEC)
	{
		*ptr++ = 'F';
		*ptr++ = 'N';
	}

	/* apply control sequence if needed */
	if (c & CTRL)
	{
		*ptr++ = 'c';
	}

	c &= 0xff;			/* strip the prefixes */

	/* and output the final sequence */

	*ptr++ = (char)c;
	*ptr = 0;			/* terminate the string */
}

/**
 * Look up a key binding in the binding table
 */
int (* GetBind(c))()
  int		c;			/* key to find what is bound to it */
{
  KEYTAB	*ktp;

	ktp = &keytab[0];	/* look in key table */
	while (ktp->k_fp != NULLFPTR)
	{
		if (ktp->k_code == c)
			return(ktp->k_fp);
		++ktp;
	}

	/* no such binding */
	return(NULLFPTR);
}

/**
 * Take a ptr to function and get the name associated with it
 */
static char * near getfname(func)
  int	(*func)();			/* ptr to the requested function to bind to */
{
  NBIND	*nptr;				/* pointer into the name binding table */

	/* scan the table, looking for a match */
	nptr = &names[0];
	while (nptr->n_func != NULLFPTR)
	{
		if (nptr->n_func == func)
			return(nptr->n_name);
		++nptr;
	}
	return(NULLCPTR);
}

/**
 * Match fname to a function in the names table and return it, or NULL if none
 */
int (* FuncMatch(fname))()
  char	*fname;			/* name to attempt to match */
{
#if	BINARY
  int	nval;			/* value of matched name */

	nval = Binary(fname, NamVal, numfuncs);
	return(nval == -1 ? NULLFPTR : names[nval].n_func);
#else
  NBIND	*ffp;			/* pointer to entry in name binding table */

	/* scan the table, returning any match */
	ffp = &names[0];
	while (ffp->n_func != NULLFPTR)
	{
		if (stracmp(fname, ffp->n_name) == FOUND)
			return(ffp->n_func);
		++ffp;
	}
	return(NULLFPTR);
#endif
}

#if	BINARY
char * NamVal(index)
  int	index;			/* index of name to fetch out of the name table */
{
	return(names[index].n_name);
}
#endif

/**
 * String key name to command key
 */
static ushort near stock(keyname)
  char		*keyname;	/* key to translate to command key form */
{
  ushort	c;			/* key sequence to return */

	/* parse it up */
	c = 0;

	/* first, the META prefix */
	if (*keyname == 'M' && *(keyname + 1) == '-')
	{
		c = META;
		keyname += 2;
	}

	/* next the function prefix */
	if (*keyname == 'F' && *(keyname + 1) == 'N')
	{
		c |= SPEC;
		keyname += 2;
	}

	/* control-x as well (but not with FN) */
	if (*keyname == '^' && *(keyname + 1) == 'X' && !(c & SPEC))
	{
		c |= CTLX;
		keyname += 2;
	}

	/* a control char? */
	if (*keyname == '^' && *(keyname + 1) != 0)
	{
		c |= CTRL;
		++keyname;
	}

	if (*keyname < 0x20)
	{
		c |= CTRL;
		*keyname |= 0x40;
	}

	/* make sure we are not lower case (not with function keys) */
	if (*keyname >= 'a' && *keyname <= 'z' && !(c & SPEC))
		*keyname ^= 0x20;

	/* the final sequence */
	c |= *keyname;
	return(c);
}

/**
 * String key name to binding name
 */
char * TransBind(skey)
  char		*skey;		/* name of key to get binding for */
{
  char		*bindname;

	bindname = getfname(GetBind(stock(skey)));
	if (bindname == NULLCPTR)
		bindname = "ERROR";
	return(bindname);
}
