/*
 *	B U F F E R . C
 *
 *	purpose:
 *		Buffer management. Some of the functions are internal, and some are
 *		actually attached to user keys. Like everyone else, they set hints
 *		for the display system.
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <alloc.h>
#include <dos.h>
#include <mem.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

/*
 * local function prototypes
 */

static int near		addline(char *);
static int near		makelist(int);
static void near	litoa(char *, int, long);

/**
 * Attach a buffer to a window. The values of dot and mark come from the buffer
 * if the use count is 0. Otherwise, they come from some other window.
 */
int UseBuffer(f, n)
  int		f, n;
{
  int		status;
  char		bufn[NBUFN];
  BUFFER	*bp;

	if ((status = MlReply("Use buffer: ", bufn, NBUFN)) != TRUE)
		return(status);
	if ((bp = FindBuf(bufn, TRUE, 0)) == NULLBPTR)
		return(FALSE);
	return(SwapBuffer(bp));
}

/**
 * Switch to the next buffer in the buffer list
 */
int NextBuffer(f, n)
  int		f;
  int		n;
{
  BUFFER	*bp1, *bp2;

	/* make sure the arg is legit */
	if (f == FALSE)
		n = 1;
	if (n < 1)
		return(FALSE);

	bp2 = curbp;
	while (n-- > 0)
	{
		/* advance to the next buffer */
		bp1 = bp2->b_fp;

		/* cycle through the buffers to find an eligible one */
		while (bp1 == NULLBPTR || bp1->b_flag & BFINVS)
		{
			if (bp1 == NULLBPTR)
				bp1 = bheadp;
			else
				bp1 = bp1->b_fp;

			/* don't get caught in an infinite loop! */
			if (bp1 == bp2)
				return(FALSE);
		}
		bp2 = bp1;
	}
	return(SwapBuffer(bp1));
}

/**
 * Switch to the previous buffer in the buffer list
 */
int PrevBuffer(f, n)
  int		f;
  int		n;
{
  BUFFER	*bp1, *bp2;

	/* make sure the arg is legit */
	if (f == FALSE)
		n = 1;
	if (n < 1)
		return(FALSE);

	bp2 = curbp;
	while (n-- > 0)
	{
		/* backup to the previous buffer */
		bp1 = bp2->b_bp;

		/* cycle through the buffers to find an eligible one */
		while (bp1 == NULLBPTR || bp1->b_flag & BFINVS)
		{
			if (bp1 == NULLBPTR)
				bp1 = btailp;
			else
				bp1 = bp1->b_bp;

			/* don't get caught in an infinite loop! */
			if (bp1 == bp2)
				return(FALSE);
		}
		bp2 = bp1;
	}
	return(SwapBuffer(bp1));
}

/**
 * Make buffer BP current
 */
int SwapBuffer(bp)
  BUFFER	*bp;
{
  WINDOW	*wp;
#if BMARKS
  int		bmark;
#endif

	if (--curbp->b_nwnd == 0)		/* last use */
	{
		curbp->b_dotp = curwp->w_dotp;
		curbp->b_doto = curwp->w_doto;
		curbp->b_markp = curwp->w_markp;
		curbp->b_marko = curwp->w_marko;
		curbp->b_anchorp = curwp->w_anchorp;
		curbp->b_anchoro = curwp->w_anchoro;
		curbp->b_fcol = curwp->w_fcol;
#if BMARKS
		for (bmark = 0; bmark < NBMARKS; bmark++)
		{
			curbp->b_bmarkp[bmark] = curwp->w_bmarkp[bmark];
			curbp->b_bmarko[bmark] = curwp->w_bmarko[bmark];
		}
#endif
	}
	curbp = bp; 					/* switch */
	if (curbp->b_active != TRUE)	/* buffer not active, so read and activate*/
	{
		ReadIn(curbp->b_fname);
		curbp->b_dotp = lforw(curbp->b_linep);
		curbp->b_doto = 0;
		curbp->b_active = TRUE;
	}
	curwp->w_bufp = bp;
	curwp->w_linep = bp->b_linep;				/* for macros, ignored */
	curwp->w_flag |= WFMODE | WFFORCE | WFHARD; /* quite nasty */
	if (bp->b_nwnd++ == 0)						/* first use */
	{
		curwp->w_dotp = bp->b_dotp;
		curwp->w_doto = bp->b_doto;
		curwp->w_markp = bp->b_markp;
		curwp->w_marko = bp->b_marko;
		curwp->w_anchorp = bp->b_anchorp;
		curwp->w_anchoro = bp->b_anchoro;
		curwp->w_fcol = bp->b_fcol;
#if BMARKS
		for (bmark = 0; bmark < NBMARKS; bmark++)
		{
			curwp->w_bmarkp[bmark] = bp->b_bmarkp[bmark];
			curwp->w_bmarko[bmark] = bp->b_bmarko[bmark];
		}
#endif
	}
	else
	{
		wp = wheadp;					/* look for old */
		while (wp != NULLWPTR)
		{
			if (wp != curwp && wp->w_bufp == bp)
			{
				curwp->w_dotp = wp->w_dotp;
				curwp->w_doto = wp->w_doto;
				curwp->w_markp = wp->w_markp;
				curwp->w_marko = wp->w_marko;
				curwp->w_fcol = wp->w_fcol;
#if BMARKS
				for (bmark = 0; bmark < NBMARKS; bmark++)
				{
					curwp->w_bmarkp[bmark] = wp->w_bmarkp[bmark];
					curwp->w_bmarko[bmark] = wp->w_bmarko[bmark];
				}
#endif
				break;
			}
			wp = wp->w_wndp;
		}
	}
	return(TRUE);
}

/**
 * Dispose of a buffer, by name. Ask for the name. Look it up (don't get too
 * upset if it isn't there at all!). Get quite upset if the buffer is being
 * displayed. Clear the buffer (ask, if the buffer has been changed). Then
 * free the header line and the buffer header.
 */
int KillBuffer(f, n)
  int		f, n;
{
  int		status;
  char		bufn[NBUFN];
  BUFFER	*bp;

	if ((status = MlReply("Kill buffer: ", bufn, NBUFN)) != TRUE)
		return(status);
	if ((bp = FindBuf(bufn, FALSE, 0)) == NULLBPTR)	/* easy if unknown */
		return(TRUE);
	if (bp->b_flag & BFINVS) 		/* deal with special buffers */
		return(TRUE);				/*	 by doing nothing */
	return(ZotBuf(bp));
}

/**
 * Kill the buffer pointed to by argument bp
 */
int ZotBuf(bp)
  BUFFER	*bp;
{
  BUFFER	*bp1, *bp2;

	if (bp->b_nwnd != 0)			/* error if on screen */
	{
		MlWrite("Buffer is being displayed");
		PcPutC(BELL);
		return(FALSE);
	}
	if (!ClrBuf(bp))				/* blow text away */
		return(FALSE);
	free((char *)bp->b_linep);		/* release header line */
	bp1 = NULLBPTR;					/* find the header */
	bp2 = bheadp;
	while (bp2 != bp)
	{
		bp1 = bp2;
		bp2 = bp2->b_fp;
	}
	bp2 = bp2->b_fp;				/* next one in chain */
	if (bp1 == NULLBPTR)			/* bp == bheadp */
	{
		bheadp = bp2;
		bheadp->b_bp = NULLBPTR;
	}
	else
	{
		bp1->b_fp = bp2;
		if (bp2 == NULLBPTR)		/* end of buffer list */
			btailp = bp1;
		else
			bp2->b_bp = bp1;
	}
	free((char *)bp);				/* release buffer block */
	return(TRUE);
}

/**
 * Rename the current buffer
 */
int NameBuffer(f, n)
  int		f, n;
{
  int		status;
  char		bufn[NBUFN];	/* buffer to hold buffer name */
  BUFFER	*bp; 			/* pointer to scan through all buffers */

	/* prompt for and get the new buffer name */
ask:
	if ((status = MlReply("Change buffer name to: ", bufn, NBUFN)) != TRUE)
		return(status);

	strlower(bufn);			/* force to lower case */

	/* and check for duplicates */
	bp = bheadp;
	while (bp != NULLBPTR)
	{
		if (bp != curbp)
		{
			/* if the names the same */
			if (stracmp(bufn, bp->b_bname) == 0)
				goto ask;			/* try again */
		}
		bp = bp->b_fp;				/* onward */
	}

	stracpy(curbp->b_bname, bufn);	/* copy buffer name to structure */
	curwp->w_flag |= WFMODE;		/* make mode line replot */
	MlErase();
	return(TRUE);
}

/**
 * List all of the active buffers.	First Update the special buffer that holds
 * the list.  Next make sure at least 1 window is displaying the buffer list,
 * splitting the screen if this is what it takes.  Lastly, repaint all of the
 * windows that are displaying the list. A numeric argument forces it to list
 * invisible buffers as well.
 */
int ListBuffers(f, n)
  int		f, n;
{
  WINDOW	*wp;
  BUFFER	*bp;
#if BMARKS
  int		bmark;
#endif

	if (!makelist(f))
		return(FALSE);
	if (blistp->b_nwnd == 0)		/* not on screen yet */
	{
		if ((wp = WindPopUp()) == NULLWPTR)
			return(FALSE);
		bp = wp->w_bufp;
		if (--bp->b_nwnd == 0)
		{
			bp->b_dotp = wp->w_dotp;
			bp->b_doto = wp->w_doto;
			bp->b_markp = wp->w_markp;
			bp->b_marko = wp->w_marko;
			bp->b_anchorp = wp->w_anchorp;
			bp->b_anchoro = wp->w_anchoro;
			bp->b_fcol = wp->w_fcol;
#if BMARKS
			for (bmark = 0; bmark < NBMARKS; bmark++)
			{
				bp->b_bmarkp[bmark] = wp->w_bmarkp[bmark];
				bp->b_bmarko[bmark] = wp->w_bmarko[bmark];
			}
#endif
		}
		wp->w_bufp = blistp;
		++blistp->b_nwnd;
	}
	wp = wheadp;
	while (wp != NULLWPTR)
	{
		if (wp->w_bufp == blistp)
		{
			wp->w_linep = lforw(blistp->b_linep);
			wp->w_dotp = lforw(blistp->b_linep);
			wp->w_doto = 0;
			wp->w_markp = NULLLPTR;
			wp->w_marko = 0;
			wp->w_anchorp = NULLLPTR;
			wp->w_anchoro = 0;
			wp->w_fcol = 0;
#if BMARKS
			for (bmark = 0; bmark < NBMARKS; bmark++)
			{
				wp->w_bmarkp[bmark] = NULLLPTR;
				wp->w_bmarko[bmark] = 0;
			}
#endif
			wp->w_flag |= WFMODE | WFHARD;
		}
		wp = wp->w_wndp;
	}
	return(TRUE);
}

/**
 * This routine rebuilds the text in the special secret buffer that holds the
 * buffer list. It is called by the ListBuffers(). Return TRUE if everything
 * works. Return FALSE if there is an error (no memory).  'iflag' indicates
 * whether to list hidden buffers.
 */
static int near makelist(iflag)
  int		iflag;			/* list hidden buffer flag */
{
  BUFFER	*bp;
  LINE		*lp;
  char		*cp1, *cp2;
  char		buf[6 + 1];		/* 6 digits plus NULLCHR */
  char		c;
  int		i;
  long		nbytes; 		/* # of bytes in current buffer */
  ushort	ramfree;		/* number of paragraphs of RAM available */
  char		line[NSTRING];	/* working storage

	blistp->b_flag &= ~BFCHG;			/* don't complain! */
	if (!ClrBuf(blistp))				/* blow old text away */
		return(FALSE);
	stracpy(blistp->b_fname, "");
	if (
			!addline("ATCN  MODES  Size    Buffer            File")
		||
			!addline("        ")
		)
		return(FALSE);

	/* display the global modes on the first line */
	cp1 = &line[0];						/* start at left edge */
	for (i = 0; i < 6; i++)				/* space over to first MODES column */
		*cp1++ = ' ';
	for (i = 0; i < NUMMODES; i++)		/* output the Global Modes codes */
	{
		if (gmode & (1 << i))
			*cp1++ = modecode[i];
		else
			*cp1++ = '.';
	}
	for (i = 0; i < 10; i++)			/* space over to first Buffer column */
		*cp1++ = ' ';
	*cp1 = NULLCHR;
	stracat(cp1, "[Global Modes]");
	if (!addline(line))
		return(FALSE);

	bp = bheadp;						/* for all buffers */
	while (bp != NULLBPTR)				/* output the list of buffers */
	{
		/* skip invisible buffers if iflag is false */
		if ((bp->b_flag & BFINVS) && (iflag != TRUE))
		{
			bp = bp->b_fp;
			continue;
		}
		cp1 = &line[0]; 				/* start at left edge */

		if (bp->b_active == TRUE)		/* "@" if activated */
			*cp1++ = '@';
		else
			*cp1++ = ' ';

		if (bp->b_flag & BFTRUNC)		/* "#" if truncated */
			*cp1++ = '#';
		else
			*cp1++ = ' ';

		if (bp->b_flag & BFCHG)			/* "*" if changed */
			*cp1++ = '*';
		else
			*cp1++ = ' ';

		if (bp->b_flag & BFNAROW)		/* "%" if narrowed */
			*cp1++ = '%';
		else
			*cp1++ = ' ';

		*cp1++ = ' ';
		*cp1++ = ' ';


		for (i = 0; i < NUMMODES; i++)	/* output the mode codes */
		{
			if (bp->b_mode & (1 << i))
				*cp1++ = modecode[i];
			else
				*cp1++ = '.';
		}

		*cp1++ = ' ';
		*cp1++ = ' ';

		nbytes = 0L;					/* count bytes in buf */
		lp = lforw(bp->b_linep);
		while (lp != bp->b_linep)
		{
			nbytes += (long)(llength(lp) + 1);
			lp = lforw(lp);
		}
		litoa(buf, 6, nbytes);			/* base 0 number of digits + NULLCHR */
		cp2 = &buf[0];
		while ((c = *cp2++) != NULLCHR)
			*cp1++ = c;

		*cp1++ = ' ';
		*cp1++ = ' ';

		cp2 = &bp->b_bname[0];			/* buffer name */
		while ((c = *cp2++) != NULLCHR)
			*cp1++ = c;

		cp2 = &bp->b_fname[0];			/* file name */
		if (*cp2 != 0)
		{
			/* if we have a file name, we need to fill out to it with spaces */
			while (cp1 < &line[4 + 2 + 5 + 2 + 6 + 2 + NBUFN + 2])
				*cp1++ = ' ';
			while ((c = *cp2++) != NULLCHR)
			{
				if (cp1 < &line[NSTRING - 1])
					*cp1++ = c;
			}
		}
		*cp1 = NULLCHR;					/* add to the buffer */
		if (!addline(line))
			return(FALSE);
		bp = bp->b_fp;
	}

	/* display amount of RAM available */
	ramfree = memfartest();				/* get number of paragraphs available */
	nbytes = (long)ramfree * 16L;		/* convert it to bytes */
	cp1 = &line[0];						/* start at left edge */
	for (i = 0; i < 13; i++)			/* space over to first Size column */
		*cp1++ = ' ';
	litoa(buf, 6, nbytes);				/* base 0 number of digits + NULLCHR */
	cp2 = &buf[0];
	while ((c = *cp2++) != NULLCHR)
		*cp1++ = c;
	*cp1++ = ' ';						/* space over to first Buffer column */
	*cp1++ = ' ';
	*cp1 = NULLCHR;
	stracat(cp1, "[RAM Available]");
	if (!addline(line))
		return(FALSE);

	return(TRUE);						/* all done */
}

/**
 * Convert a long to ascii
 */
static void near litoa(buf, width, num)
  char	buf[];			/* where to stuff the resulting string */
  int	width;			/* number of digits (base 0 (includes the NULLCHR)) */
  long	num;			/* the number to convert */
{
	buf[width] = 0; 					/* end of string */
	while (num >= 10)					/* conditional digits */
	{
		buf[--width] = (char)(num % 10L) + '0';
		num /= 10L;
	}
	buf[--width] = (char)num + '0';		/* always 1 digit */
	while (width)						/* pad with blanks */
		buf[--width] = ' ';
}

/**
 * The argument "text" points to a string. Append this line to the buffer list
 * buffer. Handcraft the EOL on the end. Return TRUE if it worked and FALSE if
 * you ran out of room.
 */
static int near addline(text)
  char	*text;
{
  LINE	*lp;
  int	ntext;

	ntext = stralen(text);
	if ((lp = LAlloc(ntext)) == NULLLPTR)
		return(FALSE);
	memcpy(lp->l_text, text, ntext);
	blistp->b_linep->l_bp->l_fp = lp;		/* hook onto the end */
	lp->l_bp = blistp->b_linep->l_bp;
	blistp->b_linep->l_bp = lp;
	lp->l_fp = blistp->b_linep;
	if (blistp->b_dotp == blistp->b_linep)	/* if dot is at the end, */
		blistp->b_dotp = lp;				/*	 move it to new line */
	return(TRUE);
}

/**
 * Find a buffer, by name. Return a pointer to the BUFFER structure associated
 * with it. If the buffer is not found and the "cflag" is TRUE, create it. The
 * "bflag" is the settings for the flags in the buffer.
 */
BUFFER * FindBuf(bname, cflag, bflag)
  char		*bname;
  int		cflag, bflag;
{
  BUFFER	*bp;
  BUFFER	*sb;    					/* buffer to insert after */
  LINE		*lp;
#if BMARKS
  int		bmark;
#endif

	strlower(bname);					/* let's make 'em all lower case */
	bp = bheadp;
	while (bp != NULLBPTR)
	{
		if (stracmp(bname, bp->b_bname) == 0)
			return(bp);
		bp = bp->b_fp;
	}
	if (cflag != FALSE)
	{
		if ((bp = (BUFFER *)malloc(sizeof(BUFFER))) == NULLBPTR)
			return(NULLBPTR);
		if ((lp = LAlloc(0)) == NULLLPTR)
		{
			free((char *)bp);
			return(NULLBPTR);
		}
		/* find the place in the list to insert this buffer */
		if (bheadp == NULLBPTR || stracmp(bheadp->b_bname, bname) > 0)
		{
			/* insert at the beginning */
			bp->b_bp = NULLBPTR;
			bp->b_fp = bheadp;			/* could be NULLBPTR (empty list) */
			if (bheadp != NULLBPTR)
				bheadp->b_bp = bp;
			else
				btailp = bp;
			bheadp = bp;
		}
		else
		{
			sb = bheadp;
			while (sb->b_fp != NULLBPTR)
			{
				if (stracmp(sb->b_fp->b_bname, bname) > 0)
					break;
				sb = sb->b_fp;
			}

			/* and insert it */
			bp->b_bp = sb;
			bp->b_fp = sb->b_fp;		/* could be NULLBPTR (end of list) */
			if (bp->b_fp != NULLBPTR)	/* if in middle, update forward ptr */
				bp->b_fp->b_bp = bp;
			else						/* else we've got a new tail */
				btailp = bp;
			sb->b_fp = bp;
		}

		/* and set up the other buffer fields */
		bp->b_active = TRUE;
		bp->b_dotp = lp;
		bp->b_doto = 0;
		bp->b_markp = NULLLPTR;
		bp->b_marko = 0;
		bp->b_anchorp = NULLLPTR;
		bp->b_anchoro = 0;
		bp->b_fcol = 0;
#if BMARKS
		for (bmark = 0; bmark < NBMARKS; bmark++)
		{
			bp->b_bmarkp[bmark] = NULLLPTR;
			bp->b_bmarko[bmark] = 0;
		}
#endif
		bp->b_topline = NULLLPTR;
		bp->b_botline = NULLLPTR;
		bp->b_flag = (char)bflag;
		bp->b_mode = gmode;
		bp->b_nwnd = 0;
		bp->b_linep = lp;
		stracpy(bp->b_fname, "");
		stracpy(bp->b_bname, bname);

		lp->l_fp = lp;
		lp->l_bp = lp;
	}
	return(bp);
}

/**
 * This routine blows away all of the text in a buffer. If the buffer is marked
 * as changed then we ask if it is ok to blow it away; this is to save the user
 * the grief of losing text. The window chain is nearly always wrong if this
 * gets called; the caller must arrange for the updates that are required.
 * Return TRUE if everything looks good.
 */
int ClrBuf(bp)
  BUFFER	*bp;
{
  LINE		*lp;
#if BMARKS
  int		bmark;
#endif

	if	(
				!(bp->b_flag & BFINVS)		/* not scratch buffer */
			&&
				bp->b_flag & BFCHG			/* something changed */
			&&
				!MlYesNo("Discard changes")
		)
		return(FALSE);
	bp->b_flag &= ~BFCHG;					/* not changed */
	while ((lp = lforw(bp->b_linep)) != bp->b_linep)
		LFree(lp);
	bp->b_dotp = bp->b_linep;				/* fix dot */
	bp->b_doto = 0;
	bp->b_markp = NULLLPTR;					/* invalidate mark */
	bp->b_marko = 0;
	bp->b_anchorp = NULLLPTR;				/* invalidate anchor */
	bp->b_anchoro = 0;
	bp->b_fcol = 0;
#if BMARKS
	for (bmark = 0; bmark < NBMARKS; bmark++)
	{
		bp->b_bmarkp[bmark] = NULLLPTR;		/* invalidate bookmarks */
		bp->b_bmarko[bmark] = 0;
	}
#endif
	return(TRUE);
}
