/*
 *	E V A L . C
 *
 *	purpose:
 *		Expression evaluation functions for CodEd
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <alloc.h>
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "var.h"
#include "protos.h"

/*
 * local function prototypes
 */

static char * near	getkill(void);
static char * near	gtenv(char *);
static char * near	gtfun(char *);
static char * near	gtusr(char *);
static char * near	ltos(int);
static int near		svar(struct vdesc *, char *);
static void near	findvar(char *, struct vdesc *, int);

/**
 * Initialize the user variable list
 */
void VarInit()
{
  int	i;

	for (i = 0; i < MAXVARS; i++)
		uv[i].u_name[0] = 0;
}

/**
 * Evaluate a function
 */
static char * near gtfun(fname)
  char			*fname;				/* name of function to evaluate */
{
  int			fnum;				/* index to function to eval */
  int			arg;				/* working storage */
  char			*tsp;				/* temporary string pointer */
  char			arg1[NSTRING];		/* value of first argument */
  char			arg2[NSTRING];		/* value of second argument */
  char			arg3[NSTRING];		/* value of third argument */
  static char	result[2 * NSTRING];/* string result */

	/* look the function up in the function table */
	fname[3] = 0;						/* only first 3 chars significant */
	strlower(fname); 					/* and let it be upper or lower case */

#if	BINARY
	fnum = Binary(fname, FunVal, NFUNCS);

	/* return errorm on a bad reference */
	if (fnum == -1)
		return(errorm);
#else
	for (fnum = 0; fnum < NFUNCS; fnum++)
		if (stracmp(fname, funcs[fnum].f_name) == 0)
			break;

	/* return errorm on a bad reference */
	if (fnum == NFUNCS)
		return(errorm);
#endif

	/* if needed, retrieve the first argument */
	if (funcs[fnum].f_type >= MONAMIC)
	{
		if (!MacArg(arg1))
			return(errorm);

		/* if needed, retrieve the second argument */
		if (funcs[fnum].f_type >= DYNAMIC)
		{
			if (!MacArg(arg2))
				return(errorm);
	    
			/* if needed, retrieve the third argument */
			if (funcs[fnum].f_type >= TRINAMIC)
				if (!MacArg(arg3))
					return(errorm);
		}
	}

	/* and now evaluate it! */
	switch (fnum)
	{
		case UFABS:
			return(i2a((arg = cvtatoi(arg1)) < 0 ? -arg : arg));
		case UFADD:
			return(i2a(cvtatoi(arg1) + cvtatoi(arg2)));
		case UFAND:
			return(ltos(SToL(arg1) && SToL(arg2)));
		case UFASCII:
			return(i2a((int)arg1[0]));
		case UFBAND:
			return(i2a(cvtatoi(arg1) & cvtatoi(arg2)));
		case UFBIND:
			return(TransBind(arg1));
		case UFBNOT:
			return(i2a(~cvtatoi(arg1)));
		case UFBOR:
			return(i2a(cvtatoi(arg1) | cvtatoi(arg2)));
		case UFBXOR:
			return(i2a(cvtatoi(arg1) ^ cvtatoi(arg2)));
		case UFCAT:
			stracpy(result, arg1);
			return(stracat(result, arg2));
		case UFCHR:
			result[0] = (char)cvtatoi(arg1);
			result[1] = 0;
			return(result);
		case UFDIV:
			return(i2a(cvtatoi(arg1) / cvtatoi(arg2)));
		case UFENV:
			arg = syenvstr(arg1, arg3, NSTRING);		/* using 'arg3' to */
			return(arg == RC_SUCCESS ? arg3 : "");		/* store env. string */
		case UFEQUAL:
			return(ltos(cvtatoi(arg1) == cvtatoi(arg2)));
		case UFEXIST:
			return(ltos(FileExist(arg1)));
		case UFFIND:
			tsp = FileLook(arg1, TRUE);
			return(tsp == NULLCPTR ? "" : tsp);
		case UFGREATER:
			return(ltos(cvtatoi(arg1) > cvtatoi(arg2)));
		case UFGTKEY:
			result[0] = (char)ExtChToCh(TermGetCh());
			result[1] = 0;
			return(result);
		case UFIND:
			return(stracpy(result, GetVal(arg1)));
		case UFLEFT:
			return(strancpy(result, arg1, cvtatoi(arg2)+1));
		case UFLENGTH:
			return(i2a(stralen(arg1)));
		case UFLESS:
			return(ltos(cvtatoi(arg1) < cvtatoi(arg2)));
		case UFLOWER:
			return(strlower(arg1));
		case UFMID:
			arg = cvtatoi(arg2);
			if (arg > stralen(arg1))
				arg = stralen(arg1);
			return(strancpy(result, &arg1[arg - 1], cvtatoi(arg3) + 1));
		case UFMOD:
			return(i2a(cvtatoi(arg1) % cvtatoi(arg2)));
		case UFNEG:
			return(i2a(-cvtatoi(arg1)));
		case UFNOT:
			return(ltos(SToL(arg1) == FALSE));
		case UFOR:
			return(ltos(SToL(arg1) || SToL(arg2)));
		case UFRIGHT:
			arg = cvtatoi(arg2);
			if (arg > stralen(arg1))
				arg = stralen(arg1);
			return(stracpy(result, &arg1[stralen(arg1) - arg]));
		case UFSEQUAL:
			return(ltos(stracmp(arg1, arg2) == 0));
		case UFSGREAT:
			return(ltos(stracmp(arg1, arg2) > 0));
		case UFSINDEX:
			return(i2a(stridxs(arg2, arg1)));
		case UFSLESS:
			return(ltos(stracmp(arg1, arg2) < 0));
		case UFSUB:
			return(i2a(cvtatoi(arg1) - cvtatoi(arg2)));
		case UFTIMES:
			return(i2a(cvtatoi(arg1) * cvtatoi(arg2)));
		case UFTRUTH:
			return(ltos(cvtatoi(arg1) == 42));
		case UFUPPER:
			return(strupper(arg1));
	}
	Quit(TRUE, FAILURE);		/* never should get here */
	return(NULLPTR);			/* make the compiler happy */
}

/**
 * Look up a user variable's value
 */
static char * near gtusr(vname)
  char	*vname;	   /* name of user variable to fetch */
{
  int	vnum;	   /* ordinal number of user var */

	/* scan the list looking for the user var name */
	for (vnum = 0; vnum < MAXVARS; vnum++)
	{
		if (uv[vnum].u_name[0] == 0)
			return(errorm);
		if (stracmp(vname, uv[vnum].u_name) == 0)
			return(uv[vnum].u_value);
	}
	/* return errorm if we run off the end */
	return(errorm);
}

#if	BINARY
char * FunVal(i)
  int	i;
{
	return(funcs[i].f_name);
}

char * EnvVal(i)
  int	i;
{
	return(envars[i]);
}

int Binary(key, tval, tlength)
  char	*key;		/* key string to look for */
  char	*(*tval)();	/* pointer to function to fetch table value with */
  int	tlength;	/* length of table to search */
{
  int	l, u;		/* lower and upper limits of Binary search */
  int	i;			/* current search index */
  int	cresult;	/* result of comparison */

	/* set current search limit as entire list */
	l = 0;
	u = tlength - 1;

	/* get the midpoint! */
	while (u >= l)
	{
		i = (l + u) >> 1;

		/* do the comparison */
		cresult = stracmp(key, (*tval)(i));
		if (cresult == 0)
			return(i);
		if (cresult < 0)
			u = i - 1;
		else
			l = i + 1;
	}
	return(-1);
}
#endif

/**
 * Look up an environment variable's value
 */
static char * near gtenv(vname)
  char	*vname;		/* name of environment variable to retrieve */
{
  int	vnum;		/* ordinal number of var referenced */

#if	BINARY
	vnum = Binary(vname, EnvVal, NEVARS);

	/* return errorm on a bad reference */
	if (vnum == -1)
		return(errorm);
#else
	/* scan the list, looking for the referenced name */
	for (vnum = 0; vnum < NEVARS; vnum++)
		if (stracmp(vname, envars[vnum]) == 0)
			break;

	/* return errorm on a bad reference */
	if (vnum == NEVARS)
		return(errorm);
#endif

	/* otherwise, fetch the appropriate value */
	switch (vnum)
	{
		case EVCBUFNAME:
			return(curbp->b_bname);
		case EVCFNAME:
			return(curbp->b_fname);
		case EVCMODE:
			return(i2a(curbp->b_mode));
		case EVCURCHAR:
			return(curwp->w_dotp->l_used == curwp->w_doto ?
					 i2a('\n') : i2a(lgetc(curwp->w_dotp, curwp->w_doto)));
		case EVCURCOL:
			return(i2a(GetCurCol(FALSE)));
		case EVCURLINE:
			return(i2a(GetCurLine()));
		case EVCWLINE:
			return(i2a(GetWindPos()));
		case EVDEBUG:
			return(ltos(macbug));
		case EVDISCMD:
			return(ltos(discmd));
		case EVDISINP:
			return(ltos(disinp));
		case EVGMODE:
			return(i2a(gmode));
		case EVHARDTAB:
			return(i2a(tabsize));
		case EVHJUMP:
			return(i2a(hjump));
		case EVHSCROLL:
			return(ltos(hscroll));
		case EVKILL:
			return(getkill());
		case EVLASTKEY:
			return(i2a(lastkey));
		case EVLINE:
			return(GetCText());
		case EVLWIDTH:
			return(i2a(llength(curwp->w_dotp)));
		case EVMATCH:
			return((patmatch == NULLCPTR) ? "" : patmatch);
		case EVPAGELEN:
			return(i2a(term.t_nrow + 1));
		case EVPENDING:
			return(falsem);
		case EVREPLACE:
			return(rpat);
		case EVRVAL:
			return(i2a(rval));
		case EVSAFESAVE:
			return(ltos(safesave));
		case EVSEARCH:
			return(pat);
		case EVSOFTTAB:
			return(i2a(stabsize));
		case EVSRES:
			return(sres);
		case EVSTATUS:
			return(ltos(cmdstatus));
		case EVTARGET:
			saveflag = lastflag;
			return(i2a(curgoal));
		case EVTPAUSE:
			return(i2a(term.t_pause));
		case EVWLINE:
			return(i2a(curwp->w_ntrows));
	}
	Quit(TRUE, FAILURE);		/* again, we should never get here */
	return(NULLPTR);			/* make the compiler happy */
}

/**
 * Return some of the contents of the kill buffer
 */
static char * near getkill()
{
  int	size;			/* max number of chars to return */
  char	value[NSTRING]; /* temp buffer for value */

	if (kbufh == NULLKPTR)
		value[0] = 0;	/* no kill buffer, just a null string */
	else
	{
		/* copy in the contents */
		if (kused < NSTRING)
			size = kused;
		else
			size = NSTRING - 1;
		strancpy(value, kbufh->d_chunk, size + 1);
	}
	/* and return the constructed value */
	return(value);
}

/**
 * Set a variable
 */
int SetVar(f, n)
  int	f, n;
{
  int	status; 		/* status return */
  VDESC vd; 			/* variable num/type */
  char	var[NVSIZE+1];	/* name of variable to fetch */
  char	value[NSTRING]; /* value to set variable to */
  char	buf[NSTRING];

	/* first get the variable to set */
	if (!clexec)
	{
		if ((status = MlReply("Variable to set: ", var, NVSIZE)) != TRUE)
			return(status);
	}
	else					/* macro line argument; grab Token and skip it */
	{
		execstr = Token(execstr, var, NVSIZE + 1);
	}

	/* check the legality and find the var */
	findvar(var, &vd, NVSIZE + 1);
	    
	/* if its not legal, bitch */
	if (vd.v_type == -1)
	{
		MlWrite("No such variable");
		return(FALSE);
	}

	/* get the value for that variable */
	if (f == TRUE)
		stracpy(value, i2a(n));
	else
	{
		if ((status = MlReply("Value: ", &value[0], NSTRING)) != TRUE)
			return(status);
	}

	/* and set the appropriate value */
	status = svar(&vd, value);

	/* if $debug == TRUE every assignment will echo a statement */
	if (macbug)
	{
		/*
		 * (((assignment status:variable name:value)))
		 */
		strfmt(outline, "(((%s:%s:%s)))", ltos(status), var, value);

		/* write out the debug line */
		MlForce(outline);
		Update(TRUE);

		/* and get the keystroke to hold the output */
		if (TermGetCh() == abortc)
		{
			MlForce("Macro aborted");
			status = FALSE;
		}
	}

	/* and return it */
	return(status);
}

/**
 * Find a variable's type and name
 */
static void near findvar(var, vd, size)
  char	*var;		/* name of var to get */
  VDESC *vd; 		/* structure to hold type and ptr */
  int	size;		/* size of var array */
{
  int	vnum;		/* subscript in variable arrays */
  int	vtype;		/* type to return */

fvar:
	vtype = -1;
	switch (var[0])
	{
		case '$':					/* check for legal enviromnent var */
			for (vnum = 0; vnum < NEVARS; vnum++)
			{
				if (stracmp(&var[1], envars[vnum]) == 0)
				{
					vtype = TKENV;
					break;
				}
			}
			break;
		case '%':					/* check for existing legal user variable */
			for (vnum = 0; vnum < MAXVARS; vnum++)
			{
				if (stracmp(&var[1], uv[vnum].u_name) == 0)
				{
					vtype = TKVAR;
					break;
				}
			}
			if (vnum < MAXVARS)		/* found an existing one */
				break;

			    
			/* got room for a new one??? */
			for (vnum = 0; vnum < MAXVARS; vnum++)
			{
				if (uv[vnum].u_name[0] == 0)
				{
					vtype = TKVAR;
					stracpy(uv[vnum].u_name, &var[1]);
					break;
				}
			}
			break;
		case '&':					/* indirect operator? */
			var[4] = 0;
			if (stracmp(&var[1], "ind") == 0)
			{
				/* grab Token, and eval it */
				execstr = Token(execstr, var, size);
				stracpy(var, GetVal(var));
				goto fvar;
			}
			break;
	}

	/* return the results */
	vd->v_num = vnum;
	vd->v_type = vtype;
	return;
}

/**
 * Set a variable
 */
static int near svar(var, value)
  VDESC	*var;		/* variable to set */
  char	*value;		/* value to set to */
{
  int	vnum;		/* ordinal number of var referenced */
  int	status; 	/* status return */
  int	c;			/* translated character */
  char	*sp; 		/* scratch string pointer */

	/* simplify the vd structure (we are gonna look at it a lot) */
	vnum = var->v_num;

	/* and set the appropriate value */
	status = TRUE;
	switch (var->v_type)
	{
		case TKVAR: 			/* set a user variable */
			if (uv[vnum].u_value != NULLCPTR)
				free(uv[vnum].u_value);
			sp = (char *)malloc(stralen(value) + 1);
			if (sp == NULLCPTR)
				return(FALSE);
			stracpy(sp, value);
			uv[vnum].u_value = sp;
			break;
		case TKENV: 			/* set an environment variable */
			status = TRUE;	/* by default */
			switch (vnum)
			{
				case EVCBUFNAME:
					stracpy(curbp->b_bname, value);
					curwp->w_flag |= WFMODE;
					break;
				case EVCFNAME:
					stracpy(curbp->b_fname, value);
					curwp->w_flag |= WFMODE;
					break;
				case EVCMODE:
					curbp->b_mode = cvtatoi(value);
					curwp->w_flag |= WFMODE;
					break;
				case EVCURCHAR:
					LDelete(1L, FALSE); 	/* delete 1 char */
					c = cvtatoi(value);
					if (c == '\n')
						LNewline();
					else
						LInsert(1, c);
					BackChar(FALSE, 1);
					break;
				case EVCURCOL:
					status = SetCurCol(cvtatoi(value));
					break;
				case EVCURLINE:
					status = GotoLine(TRUE, cvtatoi(value));
					break;
				case EVCWLINE:
					status = ForwLine(TRUE, cvtatoi(value) - GetWindPos());
					break;
				case EVDEBUG:
					macbug = SToL(value);
					break;
				case EVDISCMD:
					discmd = SToL(value);
					break;
				case EVDISINP:
					disinp = SToL(value);
					break;
				case EVGMODE:
					gmode = cvtatoi(value);
					break;
				case EVHARDTAB:
					tabsize = cvtatoi(value);
					UpWind();
					break;
				case EVHJUMP:
					hjump = cvtatoi(value);
					if (hjump < 1)
						hjump = 1;
					else if (hjump > term.t_ncol - 1)
						hjump = term.t_ncol - 1;
					break;
				case EVHSCROLL:
					hscroll = SToL(value);
					lbound = 0;
					break;
				case EVKILL:
					break;
				case EVLASTKEY:
					lastkey = cvtatoi(value);
					break;
				case EVLINE:
					PutCText(value);
				case EVLWIDTH:
					break;
				case EVMATCH:
					break;
				case EVPAGELEN:
					status = NewSize(TRUE, cvtatoi(value));
					break;
				case EVPENDING:
					break;
				case EVREPLACE:
					stracpy(rpat, value);
					break;
				case EVRVAL:
					break;
				case EVSAFESAVE:
					safesave = SToL(value);
					break;
				case EVSEARCH:
					stracpy(pat, value);
					SetJmpTbl(pat);
#if MAGIC
					McClear();
#endif
					break;
				case EVSOFTTAB:
					stabsize = cvtatoi(value);
					UpWind();
					break;
				case EVSRES:
					status = PcRez(value);
					sgarbf = TRUE;			/* screen is garbage */
					Update(TRUE);			/* so repaint */
					break;
				case EVSTATUS:
					cmdstatus = SToL(value);
					break;
				case EVTARGET:
					curgoal = cvtatoi(value);
					thisflag = saveflag;
					break;
				case EVTPAUSE:
					term.t_pause = cvtatoi(value);
					break;
				case EVWLINE:
					status = Resize(TRUE, cvtatoi(value));
					break;
			}
			break;
	}
	return(status);
}

/**
 * Integer to ascii string (this is too inconsistant to use the compiler's)
 */
char * i2a(i)
  int			i;					/* integer to translate to a string */
{
  int			digit;				/* current digit being used */
  char			*sp; 				/* pointer into result */
  int			sign;				/* sign of resulting number */
  static char	result[INTWIDTH+1]; /* resulting string */

	/* record the sign */
	sign = 1;
	if (i < 0)
	{
		sign = -1;
		i = -i;
	}

	/* and build the string (backwards!) */
	sp = result + INTWIDTH;
	*sp = 0;
	do
	{
		digit = i % 10;
		*(--sp) = (char)digit + '0';	/* and install the new digit */
		i = i / 10;
	} while (i);

	/* and fix the sign */
	if (sign == -1)
	{
		*(--sp) = '-';	/* and install the minus sign */
	}

	return(sp);
}

/**
 * Find the type of a passed Token
 */
int GetType(Token)
  char	*Token;	  /* Token to analyze */
{
  char	c;		  /* first char in Token */

	/* grab the first char (this is all we need) */
	c = *Token;

	/* no blanks!!! */
	if (c == 0)
		return(TKNUL);

	/* a numeric literal? */
	if (c >= '0' && c <= '9')
		return(TKLIT);

	switch (c)
	{
		case '"':
			return(TKSTR);
		case '!':
			return(TKDIR );
		case '@':
			return(TKARG);
		case '#':
			return(TKBUF);
		case '$':
			return(TKENV);
		case '%':
			return(TKVAR);
		case '&':
			return(TKFUN);
		case '*':
			return(TKLBL);
		default:
			return(TKCMD);
	}
}

/**
 * Find the value of a Token
 */
char * GetVal(Token)
  char			*Token;			/* Token to evaluate */
{
  int			status; 		/* error return */
  BUFFER		*bp; 			/* temp buffer pointer */
  int			blen;			/* length of buffer argument */
  int			distmp; 		/* temporary discmd flag */
  static char	buf[NSTRING];	/* string buffer for some returns */

	switch (GetType(Token))
	{
		case TKNUL:
			return("");
		case TKARG: 				/* interactive argument */
			stracpy(Token, GetVal(&Token[1]));
			distmp = discmd;		/* echo it always! */
			discmd = TRUE;
			status = GetString(Token, buf, NSTRING, ChToExtCh('\n'));
			discmd = distmp;
			if (status == ABORT)
				return(errorm);
			return(buf);
		case TKBUF: 				/* buffer contents fetch */
			/* grab the right buffer */
			stracpy(Token, GetVal(&Token[1]));
			bp = FindBuf(Token, FALSE, 0);
			if (bp == NULLBPTR)
				return(errorm);
			    
			/* if the buffer is displayed, get the window
			 *	 vars instead of the buffer vars
			 */
			if (bp->b_nwnd > 0)
			{
				curbp->b_dotp = curwp->w_dotp;
				curbp->b_doto = curwp->w_doto;
			}

			/* make sure we are not at the end */
			if (bp->b_linep == bp->b_dotp)
				return(errorm);
			    
			/* grab the line as an argument */
			blen = bp->b_dotp->l_used - bp->b_doto;
			if (blen > NSTRING)
				blen = NSTRING;
			strancpy(buf, bp->b_dotp->l_text + bp->b_doto, blen + 1);
			buf[blen] = 0;
			    
			/* and step the buffer's line ptr ahead a line */
			bp->b_dotp = bp->b_dotp->l_fp;
			bp->b_doto = 0;

			/* if displayed buffer, reset window ptr vars */
			if (bp->b_nwnd > 0)
			{
				curwp->w_dotp = curbp->b_dotp;
				curwp->w_doto = 0;
				curwp->w_flag |= WFMOVE;
			}

			/* and return the spoils */
			return(buf);		    
		case TKVAR:
			return(gtusr(Token + 1));
		case TKENV:
			return(gtenv(Token + 1));
		case TKFUN:
			return(gtfun(Token + 1));
		case TKDIR:
			return(errorm);
		case TKLBL:
			return(errorm);
		case TKLIT:
			return(Token);
		case TKSTR:
			return(Token+1);
		case TKCMD:
			return(Token);
	}
	return(NULLPTR);			/* make the compiler happy */
}

/**
 * Convert a string to a numeric logical
 */
int SToL(val)
  char	*val;	  /* value to check for SToL */
{
	/* check for logical values */
	if (val[0] == 'F')
		return(FALSE);
	if (val[0] == 'T')
		return(TRUE);

	/* check for numeric truth (!= 0) */
	return(cvtatoi(val) != 0);
}

/**
 * Numeric logical to string logical
 */
static char * near ltos(val)
  int	val;		/* value to translate */
{
	if (val)
		return(truem);
	else
		return(falsem);
}
