/*
 *	E X E C . C
 *
 *	purpose:
 *		This file is for functions dealing with execution of commands,
 *		command lines, buffers, files and startup files.
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <alloc.h>
#include <dos.h>
#include <mem.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

#define	MAXMACROS		40	/* maximum number of macros */

/*
 * local function prototypes
 */

static int near		cbuf(int, int, int);
static int near		dobuf(struct buffer *);
static int near		docmd(char *);
static void near	freewhile(struct whblock *);

/**
 * Execute a named command even if it is not bound
 */
int NamedCmd(f, n)
  int	f, n;
{
  int	(*kfunc)();		/* ptr to the requested function to bind to */

	/* prompt the user to type a named command */
	MlWrite(": ");

	/* and now get the function name to execute */
	kfunc = GetName();
	if (kfunc == NULLFPTR)
	{
		MlWrite("No such function");
		return(FALSE);
	}

	/* and then execute the command */
	return((*kfunc)(f, n));
}

/**
 * Execute a command line typed in by the user
 */
int ExecCmd()
{
  int	status;
  char	CmdStr[NSTRING];		/* string holding command to execute */

	/* get the line wanted */
	if ((status = MlReply(": ", CmdStr, NSTRING)) != TRUE)
		return(status);

	execlevel = 0;
	return(docmd(CmdStr));
}

/**
 * Take a passed string as a command line and translate it to be executed as
 * a command. This function will be used by exec-cmnd-line and by all source
 * and startup files. 'lastflag' and 'thisflag' are also updated.
 *
 * format of the command line is:
 *
 *		{# arg} <command-name> {<argument string(s)>}
 */
static int near docmd(cline)
  char	*cline;			/* command line to execute */
{
  int	f;				/* default argument flag */
  int	n;				/* numeric repeat value */
  int	(*fnc)();		/* function to execute */
  int	status; 		/* return status of function */
  int	oldcle; 		/* old contents of clexec flag */
  char	*oldestr;		/* original Exec string */
  char	tkn[NSTRING];	/* next Token off of command line */
			    
	/* if we are scanning and not executing, go back here */
	if (execlevel)
		return(TRUE);

	oldestr = execstr;		/* save last ptr to string to execute */
	execstr = cline;		/* and set this one as current */

	/* first set up the default command values */
	f = FALSE;
	n = 1;
	lastflag = thisflag;
	thisflag = 0;

	if (!MacArg(tkn))		/* and grab the first Token */
	{
		execstr = oldestr;
		return(FALSE);
	}

	/* process leadin argument */
	if (GetType(tkn) != TKCMD)
	{
		f = TRUE;
		stracpy(tkn, GetVal(tkn));
		n = cvtatoi(tkn);

		/* and now get the command to execute */
		if (!MacArg(tkn))
		{
			execstr = oldestr;
			return(FALSE); 
		}	    
	}

	/* and match the Token to see if it exists */
	if ((fnc = FuncMatch(tkn)) == NULLFPTR)
	{
		MlWrite("No such function");
		execstr = oldestr;
		return(FALSE);
	}
	    
	/* save the arguments and go execute the command */
	oldcle = clexec;			/* save old clexec flag */
	clexec = TRUE;				/* in cline execution */
	status = (*fnc)(f, n);		/* call the function */
	cmdstatus = status; 		/* save the status */
	clexec = oldcle;			/* restore clexec flag */
	execstr = oldestr;
	return(status);
}

/**
 * Chop a Token off a string. Return a pointer past the Token.
 */
char * Token(src, tok, size)
  char	*src, *tok;	/* source string, destination Token string */
  int	size;		/* maximum size of Token */
{
  int	quotef; 	/* is the current string quoted? */
  char	c;			/* temporary character */

	/* first scan past any whitespace in the source string */
	while (*src == ' ' || *src == '\t')
		++src;

	/* scan through the source string */
	quotef = FALSE;
	while (*src)
	{
		/* process special characters */
		if (*src == '~')
		{
			++src;
			if (*src == 0)
				break;
			switch (*src++)
			{
				case 'r':			/* carriage return */
					c = 13;
					break;
				case 'n':			/* line feed */
					c = 10;
					break;
				case 't':			/* tab */
					c = 9;
					break;
				case 'b':			/* backspace */
					c = 8;
					break;
				case 'f':			/* form feed */
					c = 12;
					break;
				default:
					c = *(src - 1);
			}
			if (--size > 0)
			{
				*tok++ = c;
			}
		}
		else
		{
			/* check for the end of the Token */
			if (quotef)
			{
				if (*src == '"')
					break;
			}
			else
			{
				if (*src == ' ' || *src == '\t')
					break;
			}

			/* set Quote mode if Quote found */
			if (*src == '"')
			{
				quotef = TRUE;
				src++;
				continue;
			}

			c = *src++;				/* record the character */
			if (--size > 0)
				*tok++ = c;
		}
	}

	/* terminate the Token and exit */
	if (*src)
		++src;
	*tok = 0;
	return(src);
}

/**
 * Get a macro line argument
 */
int MacArg(tok)
  char	*tok;		/* buffer to place argument */
{
  int	savcle; 	/* buffer to store original clexec */
  int	status;

	savcle = clexec;		/* save execution mode */
	clexec = TRUE;			/* get the argument */
	status = NextArg("", tok, NSTRING, ChToExtCh('\n'));
	clexec = savcle;		/* restore execution mode */
	return(status);
}

/**
 * Get the next argument
 */
int NextArg(prompt, buffer, size, terminator)
  char	*prompt; 	/* prompt to use if we are interactive */
  char	*buffer; 	/* buffer to put Token into */
  int	size;		/* size of the buffer */
  int	terminator; /* terminating char to use on interactive fetch */
{
	/* if we are interactive, go get it! */
	if (!clexec)
		return(GetString(prompt, buffer, size, terminator));

	/* grab Token and advance past */
	execstr = Token(execstr, buffer, size);

	/* evaluate it */
	stracpy(buffer, GetVal(buffer));
	return(TRUE);
}

/**
 * Set up a macro buffer and flag to store all executed command lines there
 */
int StoreMac(f, n)
  int			f;
  int			n;
{
  struct buffer	*bp; 			/* pointer to macro buffer */
  char			bname[NBUFN];	/* name of buffer to use */

	/* must have a numeric argument to this function */
	if (f == FALSE)
	{
		MlWrite("No macro specified");
		return(FALSE);
	}

	/* range check the macro number */
	if (n < 1 || n > MAXMACROS)
	{
		MlWrite("Macro number out of range");
		return(FALSE);
	}

	/* construct the macro buffer name */
	stracpy(bname, "macro xx");
	bname[6] = (char)(n / 10) + '0';
	bname[7] = (char)(n % 10) + '0';

	/* set up the new macro buffer */
	if ((bp = FindBuf(bname, TRUE, BFINVS)) == NULLBPTR)
	{
		MlWrite("Can't create macro");
		return(FALSE);
	}

	/* and make sure it is empty */
	ClrBuf(bp);

	/* and set the macro store pointers to it */
	mstore = TRUE;
	bstore = bp;
	return(TRUE);
}

/**
 * Set up a procedure buffer and flag to store all executed command lines
 */
int StoreProc(f, n)
  int		f, n;
{
  int		status;
  char		bname[NBUFN];	/* name of buffer to use */
  BUFFER	*bp; 			/* pointer to macro buffer */

	/* a numeric argument means its a numbered macro */
	if (f == TRUE)
		return(StoreMac(f, n));

	/* get the name of the procedure */
	if ((status = MlReply("Procedure name: ", &bname[1], NBUFN - 2)) != TRUE)
		return(status);

	/* construct the macro buffer name */
	bname[0] = '[';
	stracat(bname, "]");

	/* set up the new macro buffer */
	if ((bp = FindBuf(bname, TRUE, BFINVS)) == NULLBPTR)
	{
		MlWrite("Can't create macro");
		return(FALSE);
	}

	/* and make sure it is empty */
	ClrBuf(bp);

	/* and set the macro store pointers to it */
	mstore = TRUE;
	bstore = bp;
	return(TRUE);
}

/**
 * Execute a procedure
 */
int ExecProc(f, n)
  int		f, n;
{
  int		status;
  char		bufn[NBUFN+2];	/* name of buffer to execute */
  BUFFER	*bp; 			/* ptr to buffer to execute */

	/* find out what buffer the user wants to execute */
	if ((status = MlReply("Execute procedure: ", &bufn[1], NBUFN)) != TRUE)
		return(status);

	/* construct the buffer name */
	bufn[0] = '[';
	stracat(bufn, "]");

	/* find the pointer to that buffer */
	if ((bp = FindBuf(bufn, FALSE, 0)) == NULLBPTR)
	{
		MlWrite("No such procedure");
		return(FALSE);
	}

	/* and now execute it as asked */
	while (n-- > 0)
	{
		if (!dobuf(bp))
			return(FALSE);
	}
	return(TRUE);
}

/**
 * Execute the contents of a buffer of commands
 */
int ExecBuf(f, n)
  int		f, n;
{
  int		status;
  char		bufn[NSTRING];	/* name of buffer to execute */
  BUFFER	*bp; 			/* ptr to buffer to execute */

	/* find out what buffer the user wants to execute */
	if ((status = MlReply("Execute buffer: ", bufn, NBUFN)) != TRUE)
		return(status);

	/* find the pointer to that buffer */
	if ((bp = FindBuf(bufn, FALSE, 0)) == NULLBPTR)
	{
		MlWrite("No such buffer");
		return(FALSE);
	}

	/* and now execute it as asked */
	while (n-- > 0)
	{
		if (!dobuf(bp))
			return(FALSE);
	}
	return(TRUE);
}

/**
 * Execute the contents of a buffer
 *
 *	Directives start with a "!" and include:
 *
 *		!endm			end a macro
 *		!if (cond)		conditional execution
 *		!else
 *		!endif
 *		!return 		return (terminating current macro)
 *		!goto <label>	jump to a label in the current macro
 *		!force			force macro to continue even if command fails
 *		!while (cond)	execute a loop if the condition is true
 *		!endwhile
 *		 
 *	Line labels begin with a "*" as the first nonblank char, like:
 *
 *		*LBL01
 */
static int near dobuf(bp)
  BUFFER	*bp; 			/* buffer to execute */
{
  int		status; 		/* status return */
  LINE		*lp; 			/* pointer to line to execute */
  LINE		*hlp;			/* pointer to line header */
  LINE		*glp;			/* line to goto */
  LINE		*mp; 			/* macro line storage temp */
  int		dirnum; 		/* directive index */
  int		linlen; 		/* length of line to execute */
  int		i;				/* index */
  int		c;				/* temp character */
  int		force;			/* force TRUE result? */
  WINDOW	*wp; 			/* ptr to windows to scan */
  WHBLOCK	*whlist; 		/* ptr to !WHILE list */
  WHBLOCK	*scanner;		/* ptr during scan */
  WHBLOCK	*whtemp; 		/* temporary ptr to a WHBLOCK */
  char		*einit;			/* initial value of eline */
  char		*eline;			/* text of line to execute */
  char		tkn[NSTRING];	/* buffer to evaluate an expression in */
  char		buf[NSTRING];

	/* clear IF level flags/while ptr */
	execlevel = 0;
	whlist = NULLWHPTR;
	scanner = NULLWHPTR;

	/* scan the buffer to execute, building WHILE header blocks */
	hlp = bp->b_linep;
	lp = hlp->l_fp;
	while (lp != hlp)
	{
		/* scan the current line */
		eline = lp->l_text;
		i = lp->l_used;

		/* trim leading whitespace */
		while (i-- > 0 && (*eline == ' ' || *eline == '\t'))
			++eline;

		/* if theres nothing here, don't bother */
		if (i <= 0)
			goto nxtscan;

		/* if it's a WHILE directive, make a block */
		if (eline[0] == '!' && eline[1] == 'w' && eline[2] == 'h')
		{
			whtemp = (WHBLOCK *)malloc(sizeof(WHBLOCK));
			if (whtemp == NULLWHPTR)
			{
noram:
				OutOfMem();
failexit:
				freewhile(scanner);
				freewhile(whlist);
				return(FALSE);
			}
			whtemp->w_begin = lp;
			whtemp->w_type = BTWHILE;
			whtemp->w_next = scanner;
			scanner = whtemp;
		}

		/* if it's a BREAK directive, make a block */
		if (eline[0] == '!' && eline[1] == 'b' && eline[2] == 'r')
		{
			if (scanner == NULLWHPTR)
			{
				MlWrite("!BREAK outside of any !WHILE loop");
				goto failexit;
			}
			whtemp = (WHBLOCK *)malloc(sizeof(WHBLOCK));
			if (whtemp == NULLWHPTR)
				goto noram;
			whtemp->w_begin = lp;
			whtemp->w_type = BTBREAK;
			whtemp->w_next = scanner;
			scanner = whtemp;
		}

		/* if it's an ENDWHILE directive, record the spot */
		if (eline[0] == '!' && strancmp(&eline[1], "endw", 4) == 0)
		{
			if (scanner == NULLWHPTR)
			{
				strfmt(buf, "!ENDWHILE with no preceding !WHILE in '%s'",
						bp->b_bname);
				MlWrite(buf);
				goto failexit;
			}

			/* move top records from the scanner list to the whlist until we
			 * have moved all BREAK records and one WHILE record
			 */
			do
			{
				scanner->w_end = lp;
				whtemp = whlist;
				whlist = scanner;
				scanner = scanner->w_next;
				whlist->w_next = whtemp;
			} while (whlist->w_type == BTBREAK);
		}

nxtscan:
		/* on to the next line */
		lp = lp->l_fp;
	}

	/* while and endwhile should match! */
	if (scanner != NULLWHPTR)
	{
		strfmt(buf, "!WHILE with no matching !ENDWHILE in '%s'", bp->b_bname);
		MlWrite(buf);
		goto failexit;
	}

	/* let the first command inherit the flags from the last one */
	thisflag = lastflag;

	/* starting at the beginning of the buffer */
	hlp = bp->b_linep;
	lp = hlp->l_fp;
	while (lp != hlp)
	{
		/* allocate eline and copy macro line to it */
		linlen = lp->l_used;
		if ((einit = eline = (char *)malloc(linlen + 1)) == NULLCPTR)
		{
			OutOfMem();
			freewhile(whlist);
			return(FALSE);
		}
		strancpy(eline, lp->l_text, linlen + 1);
		eline[linlen] = 0;		/* make sure it ends */

		/* trim leading whitespace */
		while (*eline == ' ' || *eline == '\t')
			++eline;

		/* dump comments and blank lines */
		if (*eline == ';' || *eline == 0)
			goto onward;

		/* if $debug == TRUE, every line to execute gets echoed and a key
		 * needs to be pressed to continue (Alt-A will abort the command)
		 */
		if (macbug)
		{
			/*
			 * <<<macro name:if levels:the line>>>
			 */
			strfmt(outline, "<<<%s:%s:%s>>>", bp->b_bname,i2a(execlevel),eline);

			/* write out the debug line */
			MlForce(outline);
			Update(TRUE);
    
			/* and get the keystroke */
			if ((c = TermGetCh()) == abortc)
			{
				MlForce("Macro aborted");
				freewhile(whlist);
				return(FALSE);
			}

			if (c == metac)
				macbug = FALSE;
		}

		/* parse directives here */
		dirnum = -1;
		if (*eline == '!')
		{
			/* Find out which directive this is */
			++eline;
			for (dirnum = 0; dirnum < NUMDIRS; dirnum++)
				if (strancmp(eline, dname[dirnum], stralen(dname[dirnum])) == 0)
					break;

			/* and bitch if it's illegal */
			if (dirnum == NUMDIRS)
			{
				MlWrite("Unknown directive");
				freewhile(whlist);
				return(FALSE);
			}

			/* service only the !ENDM macro here */
			if (dirnum == DENDM)
			{
				mstore = FALSE;
				bstore = NULLBPTR;
				goto onward;
			 }
			 /* restore the original eline */
			 --eline;
		 }

		 /* if macro store is on, just salt this away */
		 if (mstore)
		 {
			/* allocate the space for the line */
			linlen = stralen(eline);
			if ((mp = LAlloc(linlen)) == NULLLPTR)
			{
				OutOfMem();
				return(FALSE);
			}
	    
			/* copy the text into the new line */
			memcpy(mp->l_text, eline, linlen);
	    
			/* attach the line to the end of the buffer */
			bstore->b_linep->l_bp->l_fp = mp;
			mp->l_bp = bstore->b_linep->l_bp;
			bstore->b_linep->l_bp = mp;
			mp->l_fp = bstore->b_linep;
			goto onward;
		}

		force = FALSE;

		/* dump comments */
		if (*eline == '*')
			goto onward;

		/* now, execute directives */
		if (dirnum != -1)
		{
			/* skip past the directive */
			while (*eline && *eline != ' ' && *eline != '\t')
				++eline;
			execstr = eline;

			switch (dirnum)
			{
				case DIF:		/* IF directive */
					/* grab the value of the logical exp */
					if (execlevel == 0)
					{
						if (MacArg(tkn) != TRUE)
							goto eexec;
						if (SToL(tkn) == FALSE)
							++execlevel;
					}
					else
						++execlevel;
					goto onward;
				case DWHILE:	/* WHILE directive */
					/* grab the value of the logical exp */
					if (execlevel == 0)
					{
						if (MacArg(tkn) != TRUE)
							goto eexec;
						if (SToL(tkn) == TRUE)
							goto onward;
					}
					/* drop down and act just like !BREAK */
				case DBREAK:	/* BREAK directive */
					if (dirnum == DBREAK && execlevel)
						goto onward;

					/* jump down to the endwhile */
					/* find the right while loop */
					whtemp = whlist;
					while (whtemp)
					{
						if (whtemp->w_begin == lp)
							break;
						whtemp = whtemp->w_next;
					}
					    
					if (whtemp == NULLWHPTR)
					{
looperr:
						MlWrite("Internal while loop error");
						freewhile(whlist);
						return(FALSE);
					}
					    
					/* reset the line pointer back */
					lp = whtemp->w_end;
					goto onward;
				case DELSE: 	/* ELSE directive */
					if (execlevel == 1)
						--execlevel;
					else if (execlevel == 0 )
						++execlevel;
					goto onward;
				case DENDIF:	/* ENDIF directive */
					if (execlevel)
						--execlevel;
					goto onward;
				case DGOTO: 	/* GOTO directive */
					/* only if we are currently executing */
					if (execlevel == 0)
					{
						/* grab label to jump to */
						eline = Token(eline, golabel, NPAT);
						linlen = stralen(golabel);
						glp = hlp->l_fp;
						while (glp != hlp)
						{
							if	(
										*glp->l_text == '*'
									&&
										(
											strancmp	(
															&glp->l_text[1],
															golabel,
															linlen
														) == 0
										)
								)
							{
								lp = glp;
								goto onward;
							}
							glp = glp->l_fp;
						}
						MlWrite("No such label");
						freewhile(whlist);
						return(FALSE);
					}
					goto onward;
				case DRETURN:	/* RETURN directive */
					if (execlevel == 0)
						goto eexec;
					goto onward;
				case DENDWHILE: /* ENDWHILE directive */
					if (execlevel)
					{
						--execlevel;
						goto onward;
					}
					else
					{
						/* find the right while loop */
						whtemp = whlist;
						while (whtemp)
						{
							if (whtemp->w_type == BTWHILE && whtemp->w_end ==lp)
								break;
							whtemp = whtemp->w_next;
						}
			    
						if (whtemp == NULLWHPTR)
							goto looperr;
			    
						/* reset the line pointer back */
						lp = whtemp->w_begin->l_bp;
						goto onward;
					}
				case DFORCE:	/* FORCE directive */
					force = TRUE;
			}
		}

		/* execute the statement */
		status = docmd(eline);
		if (force)				/* force the status */
			status = TRUE;

		/* check for a command error */
		if (status != TRUE)
		{
			/* look if buffer is showing */
			wp = wheadp;
			while (wp != NULLWPTR)
			{
				if (wp->w_bufp == bp)
				{
					/* and point it */
					wp->w_dotp = lp;
					wp->w_doto = 0;
					wp->w_flag |= WFHARD;
				}
				wp = wp->w_wndp;
			}
			/* in any case set the buffer '.' */
			bp->b_dotp = lp;
			bp->b_doto = 0;
			free(einit);
			execlevel = 0;
			freewhile(whlist);
			return(status);
		}
onward:
		/* on to the next line */
		free(einit);
		lp = lp->l_fp;
	}

eexec:
	/* exit the current function */
	execlevel = 0;
	freewhile(whlist);
	return(TRUE);
}

/**
 * Free a list of while block pointers
 */
static void near freewhile(wp)
  WHBLOCK	*wp;    /* head of structure to free */
{
	while (wp != NULLWHPTR)
	{
		freewhile(wp->w_next);
		free(wp);
	}
}

/**
 * Execute a series of commands in a file
 */
int ExecFile(f, n)
  int	f, n;
{
  int	status;
  char	fname[NSTRING]; /* name of file to execute */
  char	*fspec;			/* full file spec */

	if ((status = MlReply("File to execute: ", fname, NSTRING - 1)) != TRUE)
		return(status);

	/* look up the path for the file and Quit if it isn't around */
	if ((fspec = FileLook(fname, TRUE)) == NULLCPTR)
		return(FALSE);

	/* otherwise, execute it */
	while (n-- > 0)
	{
		if (!DoFile(fspec))
			return(FALSE);
	}
	return(TRUE);
}

/**
 * Yank a file into a buffer and execute it. If there are no errors,
 * delete the buffer on exit.
 */
int DoFile(fname)
  char		*fname;			/* file name to execute */
{
  BUFFER	*bp; 			/* buffer to place file to exeute */
  BUFFER	*cb; 			/* temp to hold current buf while we read */
  char		bname[NBUFN];	/* name of buffer */

	if (!MakeName(bname, fname))	/* derive the name of the buffer */
		return(FALSE);				/* invalid DOS filename; user said abort */
	UniqName(bname); 				/* make sure we don't stomp things */
	if ((bp = FindBuf(bname, TRUE, 0)) == NULLBPTR)	/* get the needed buffer */
		return(FALSE);

	bp->b_mode = MDVIEW;	/* mark the buffer as read only */
	cb = curbp; 			/* save the old buffer */
	curbp = bp; 			/* make this one current */
	/* and try to read in the file to execute */
	if (!ReadIn(fname))
	{
		curbp = cb; 		/* restore the current buffer */
		return(FALSE);
	}

	/* go execute it! */
	curbp = cb; 			/* restore the current buffer */
	if (!dobuf(bp))
		return(FALSE);

	/* if not displayed, remove the now unneeded buffer and exit */
	if (bp->b_nwnd == 0)
		ZotBuf(bp);
	return(TRUE);
}

/**
 * Execute the contents of a numbered buffer
 */
static int near cbuf(f, n, bufnum)
  int			f, n;
  int			bufnum; 	/* number of buffer to execute */
{
  BUFFER		*bp; 		/* ptr to buffer to execute */
  static char	bufname[] = "macro xx";

	/* make the buffer name */
	bufname[6] = (char)(bufnum / 10) + '0';
	bufname[7] = (char)(bufnum % 10) + '0';

	/* find the pointer to that buffer */
	if ((bp = FindBuf(bufname, FALSE, 0)) == NULLBPTR)
	{
		MlWrite("Macro not defined");
		return(FALSE);
	}

	/* and now execute it as asked */
	while (n-- > 0)
	{
		if (!dobuf(bp))
			return(FALSE);
	}
	return(TRUE);
}

int ExecMac1(f, n)
  int	f, n;
{
	return(cbuf(f, n, 1));
}

int ExecMac2(f, n)
  int	f, n;
{
	return(cbuf(f, n, 2));
}

int ExecMac3(f, n)
  int	f, n;
{
	return(cbuf(f, n, 3));
}

int ExecMac4(f, n)
  int	f, n;
{
	return(cbuf(f, n, 4));
}

int ExecMac5(f, n)
  int	f, n;
{
	return(cbuf(f, n, 5));
}

int ExecMac6(f, n)
  int	f, n;
{
	return(cbuf(f, n, 6));
}

int ExecMac7(f, n)
  int	f, n;
{
	return(cbuf(f, n, 7));
}

int ExecMac8(f, n)
  int	f, n;
{
	return(cbuf(f, n, 8));
}

int ExecMac9(f, n)
  int	f, n;
{
	return(cbuf(f, n, 9));
}

int ExecMac10(f, n)
  int	f, n;
{
	return(cbuf(f, n, 10));
}

int ExecMac11(f, n)
  int	f, n;
{
	return(cbuf(f, n, 11));
}

int ExecMac12(f, n)
  int	f, n;
{
	return(cbuf(f, n, 12));
}

int ExecMac13(f, n)
  int	f, n;
{
	return(cbuf(f, n, 13));
}

int ExecMac14(f, n)
  int	f, n;
{
	return(cbuf(f, n, 14));
}

int ExecMac15(f, n)
  int	f, n;
{
	return(cbuf(f, n, 15));
}

int ExecMac16(f, n)
  int	f, n;
{
	return(cbuf(f, n, 16));
}

int ExecMac17(f, n)
  int	f, n;
{
	return(cbuf(f, n, 17));
}

int ExecMac18(f, n)
  int	f, n;
{
	return(cbuf(f, n, 18));
}

int ExecMac19(f, n)
  int	f, n;
{
	return(cbuf(f, n, 19));
}

int ExecMac20(f, n)
  int	f, n;
{
	return(cbuf(f, n, 20));
}

int ExecMac21(f, n)
  int	f, n;
{
	return(cbuf(f, n, 21));
}

int ExecMac22(f, n)
  int	f, n;
{
	return(cbuf(f, n, 22));
}

int ExecMac23(f, n)
  int	f, n;
{
	return(cbuf(f, n, 23));
}

int ExecMac24(f, n)
  int	f, n;
{
	return(cbuf(f, n, 24));
}

int ExecMac25(f, n)
  int	f, n;
{
	return(cbuf(f, n, 25));
}

int ExecMac26(f, n)
  int	f, n;
{
	return(cbuf(f, n, 26));
}

int ExecMac27(f, n)
  int	f, n;
{
	return(cbuf(f, n, 27));
}

int ExecMac28(f, n)
  int	f, n;
{
	return(cbuf(f, n, 28));
}

int ExecMac29(f, n)
  int	f, n;
{
	return(cbuf(f, n, 29));
}

int ExecMac30(f, n)
  int	f, n;
{
	return(cbuf(f, n, 30));
}

int ExecMac31(f, n)
  int	f, n;
{
	return(cbuf(f, n, 31));
}

int ExecMac32(f, n)
  int	f, n;
{
	return(cbuf(f, n, 32));
}

int ExecMac33(f, n)
  int	f, n;
{
	return(cbuf(f, n, 33));
}

int ExecMac34(f, n)
  int	f, n;
{
	return(cbuf(f, n, 34));
}

int ExecMac35(f, n)
  int	f, n;
{
	return(cbuf(f, n, 35));
}

int ExecMac36(f, n)
  int	f, n;
{
	return(cbuf(f, n, 36));
}

int ExecMac37(f, n)
  int	f, n;
{
	return(cbuf(f, n, 37));
}

int ExecMac38(f, n)
  int	f, n;
{
	return(cbuf(f, n, 38));
}

int ExecMac39(f, n)
  int	f, n;
{
	return(cbuf(f, n, 39));
}

int ExecMac40(f, n)
  int	f, n;
{
	return(cbuf(f, n, 40));
}
