/*
 *	F I L E . C
 *
 *	purpose:
 *		These routines handle the reading, writing and lookup of disk files.
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <dir.h>
#include <dos.h>
#include <mem.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

/*
 * local function prototypes
 */

static int near		ffgetline(void);
static int near		ifile(char *);
static int near		validfilename(char *);

/*
 * local (module global) variable definitions
 */

static FILE	*fp;			/* file pointer */
static char	linebuf[NLINE];	/* buffer to read a line into */

/**
 * Read a file into the current buffer. This is really easy; all you do it
 * find the name of the file, and call the standard "read a file into the
 * current buffer" code.
 */
int FileRead(f, n)
  int	f, n;
{
  int	status;
  char	fname[NFILEN];

	if ((status = MlReply("Read file: ", fname, NFILEN)) != TRUE)
		return(status);
	return(ReadIn(fname));
}

/**
 * Insert a file into the current buffer. This is really easy; all you do is
 * find the name of the file, and call the standard "insert a file into the
 * current buffer" code.
 */
int InsFile(f, n)
  int	f, n;
{
  int	status;
  char	fname[NFILEN];

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if ((status = MlReply("Insert file: ", fname, NFILEN)) != TRUE)
		return(status);
	return(ifile(fname));
}

/**
 * Select a file for editing. Look around to see if you can find the
 * file in another buffer; if you can find it, just switch to the buffer.
 * If you cannot find the file, create a new buffer, read in the text, and
 * switch to the new buffer.
 */
int FileFind(f, n)
  int	f, n;
{
  int	status;
  char	fname[NFILEN];	/* file user wishes to find */

	if ((status = MlReply("Find file: ", fname, NFILEN)) != TRUE)
		return(status);
	return(GetFile(fname));
}

/**
 * Visit a file in VIEW mode
 */
int ViewFile(f, n)
  int		f, n;
{
  int		status;
  char		fname[NFILEN];	/* file user wishes to find */
  WINDOW	*wp; 			/* scan for windows that need updating */

	if ((status = MlReply("View file: ", fname, NFILEN)) != TRUE)
		return(status);
	if (GetFile(fname))		/* if we succeed, put it in view mode */
	{
		curwp->w_bufp->b_mode |= MDVIEW;

		/* Update mode lines of all windows */
		UpMode();
		return(TRUE);
	}
	return(FALSE);
}

/**
 * Create a buffer (if necessary), switch to it, then read a file into it
 */
int GetFile(fname)
  char		*fname;			/* file name to find */
{
  BUFFER	*bp;
  LINE		*lp;
  int		i;
  int		s;
#if BMARKS
  int		bmark;
#endif
  char		bname[NBUFN];	/* buffer name to put file */

	for (bp = bheadp; bp != NULLBPTR; bp = bp->b_fp)
	{
		if ((bp->b_flag & BFINVS) == 0 && stracmp(bp->b_fname, fname) == 0)
		{
			SwapBuffer(bp);
			lp = curwp->w_dotp;
			i = curwp->w_ntrows / 2;
			while (i-- && lback(lp) != curbp->b_linep)
				lp = lback(lp);
			curwp->w_linep = lp;
			curwp->w_flag |= WFMODE | WFHARD;
			MlWrite("Old buffer");
			return(TRUE);
		}
	}
	if (!MakeName(bname, fname))	/* new buffer name */
		return(FALSE);				/* invalid DOS filename; user said abort */
	while ((bp = FindBuf(bname, FALSE, 0)) != NULLBPTR)
	{
		/* old buffer name conflict code */
		s = MlReply("Buffer name: ", bname, NBUFN);
		if (s == ABORT) 				/* Alt-A to just Quit */
			return(s);
		if (s == FALSE) 				/* CR to clobber it */
		{
			if (!MakeName(bname, fname))
				return(FALSE);
			break;
		}
	}
	if (bp == NULLBPTR && (bp = FindBuf(bname, TRUE, 0)) == NULLBPTR)
	{
		MlWrite("Can't create buffer");
		return(FALSE);
	}
	if (--curbp->b_nwnd == 0)			/* undisplay */
	{
		curbp->b_dotp = curwp->w_dotp;
		curbp->b_doto = curwp->w_doto;
		curbp->b_markp = curwp->w_markp;
		curbp->b_marko = curwp->w_marko;
		curbp->b_anchorp = curwp->w_anchorp;
		curbp->b_anchoro = curwp->w_anchoro;
		curbp->b_fcol = curwp->w_fcol;
#if BMARKS
		for (bmark = 0; bmark < NBMARKS; bmark++)
		{
			curbp->b_bmarkp[bmark] = curwp->w_bmarkp[bmark];
			curbp->b_bmarko[bmark] = curwp->w_bmarko[bmark];
		}
#endif
	}
	curbp = bp; 						/* switch to it */
	curwp->w_bufp = bp;
	curbp->b_nwnd++;
	return(ReadIn(fname));				/* read it in */
}

/**
 * Read file "fname" into the current buffer, blowing away any text found there.
 * Called by both the read and find commands.  Return the final status of the
 * read.  Also called to read in a file specified on the DOS command line.
 */
int ReadIn(fname)
  char		*fname;			/* name of file to read */
{
  LINE		*lp1;
  LINE		*lp2;
  WINDOW	*wp;
  BUFFER	*bp;
  int		s;
  int		i;
  char		*cp1;
  int		nbytes;
  int		nline;
#if BMARKS
  int		bmark;
#endif
  char		buf[NSTRING];

  /* Cmode FileName EXTensions */
  static char	*cfnext[] = { "c", "h", "cpp", "hpp", "asm", "slt" };

#define	NCFNES	sizeof(cfnext) / sizeof(char *)

	bp = curbp; 			/* cheap */
	if (!ClrBuf(bp))		/* might be old */
		return(FALSE);
	bp->b_flag &= ~(BFINVS | BFCHG);
	strlower(fname); 		/* let's make 'em all lower case */
	stracpy(bp->b_fname, fname);

	/* auto CMODE check */
	cp1 = fname;
	while (*cp1 != NULLCHR)				/* go to the end of the FileName */
		++cp1;
	while (*cp1 != '.' && cp1 != fname)	/* back up till we get a dot */
		--cp1;
	if (*cp1 == '.' && cp1 != fname)	/* FileName had an extension */
	{
		++cp1;							/* first char past the dot */
		for (i = 0; i < NCFNES; i++)
		{
			if (stracmp(cp1, cfnext[i]) == 0)
			{
				curbp->b_mode |= MDCMODE;
				curwp->w_flag |= WFMODE;
			}
		}
	}
	
	if ((fp = fopen(fname, "rb")) == (FILE *)NULLCPTR)
	{
		MlWrite("New file");			/* file not found */
		goto out;
	}

	/* read the file in */
	MlWrite("Reading file");
	nline = 0;
	while ((s = ffgetline()) == FIOSUC)
	{
		nbytes = stralen(linebuf);
		if ((lp1 = LAlloc(nbytes)) == NULLLPTR)
		{
			s = FIOMEM; 				/* keep message on the display */
			break;
		}
		lp2 = lback(curbp->b_linep);
		lp2->l_fp = lp1;
		lp1->l_fp = curbp->b_linep;
		lp1->l_bp = lp2;
		curbp->b_linep->l_bp = lp1;
		memcpy(lp1->l_text, linebuf, nbytes);
		++nline;
	}
	fclose(fp);							/* ignore errors */

	buf[0] = NULLCHR;					/* in case there is no file i/o error */
	if (s == FIOERR || s == FIOMEM)
	{
		switch(s)
		{
			case FIOERR:
				stracpy(buf, "I/O Error, ");
				break;

			case FIOMEM:
				stracpy(buf, "Out of memory, ");
				break;
		}
		curbp->b_flag |= BFTRUNC;
		PcPutC(BELL);
	}
	strfmt(&buf[stralen(buf)], "Read %d lines", nline);
	MlWrite(buf);

out:
	for (wp = wheadp; wp != NULLWPTR; wp = wp->w_wndp)
	{
		if (wp->w_bufp == curbp)
		{
			wp->w_linep = lforw(curbp->b_linep);
			wp->w_dotp = lforw(curbp->b_linep);
			wp->w_doto = 0;
			wp->w_markp = NULLLPTR;
			wp->w_marko = 0;
			wp->w_anchorp = NULLLPTR;
			wp->w_anchoro = 0;
			wp->w_fcol = 0;
#if BMARKS
			for (bmark = 0; bmark < NBMARKS; bmark++)
			{
				wp->w_bmarkp[bmark] = NULLLPTR;
				wp->w_bmarko[bmark] = 0;
			}
#endif
			wp->w_flag |= WFMODE | WFHARD;
		}
	}
	return(s == FIOERR || s == FIOFNF ? FALSE : TRUE);	/* false if error */
}

/**
 * Take a file name, and from it fabricate a buffer name. This routine knows
 * about the syntax of file names.
 */
int MakeName(bname, fname)
  char	bname[];
  char	fname[];
{
  char	*cp1;
  char	*cp2;
  char	buf[NSTRING];

	if (!validfilename(fname))
	{
		PcPutC(BELL);
		strfmt(buf, "Invalid DOS file name (%s). Use it", fname);
		if (MlYesNo(buf) == FALSE)
		{
			CmdAbort(FALSE, 1);
			return(FALSE);
		}
	}

	cp1 = &fname[0];
	while (*cp1)
		++cp1;

	while	(
					cp1 != &fname[0]
				&&
					*(cp1-1) != ':'
				&&
					*(cp1-1) != '\\'
				&&
					*(cp1-1) != '/'
			)
		--cp1;

	cp2 = &bname[0];
	while (cp2 != &bname[NBUFN - 1] && *cp1 != 0)
		*cp2++ = *cp1++;
	*cp2 = 0;
	return(TRUE);
}

/**
 * Make sure a buffer name is unique
 */
int UniqName(name)
  char	*name;	  /* name to check on */
{
  char	*sp;

	/* check to see if it is in the buffer list */
	while (FindBuf(name, 0, FALSE) != NULLBPTR)
	{
		/* go to the end of the name */
		sp = name;
		while (*sp)
			++sp;
		if (sp == name || (*(sp - 1) < '0' || *(sp - 1) > '8'))
		{
			*sp++ = '0';
			*sp = 0;
		}
		else
			*(--sp) += 1;
	}
	return(TRUE);
}

/**
 * Ask for a file name, and write the contents of the current buffer to that
 * file. Update the remembered file name and clear the buffer changed flag.
 */
int FileWrite(f, n)
  int		f, n;
{
  int		status;
  char		fname[NFILEN];
  WINDOW	*wp;

	if ((status = MlReply("Write file: ", fname, NFILEN)) != TRUE)
		return(status);
	if (WriteOut(fname))
	{
		stracpy(curbp->b_fname, fname);
		curbp->b_flag &= ~BFCHG;

		/* Update all mode lines */
		UpMode();
		return(TRUE);
	}
	return(FALSE);
}

/**
 * Save the contents of the current buffer in its associated file. Do nothing
 * if nothing has changed (this may be a bug, not a feature). Error if there
 * is no remembered file name for the buffer.
 */
int FileSave(f, n)
  int		f, n;
{
  WINDOW	*wp;

	if (curbp->b_mode & MDVIEW) 		/* don't allow this command if */
		return(ReadOnly());				/* we are in read only mode */
	if ((curbp->b_flag & BFCHG) == 0)	/* return, no changes */
		return(TRUE);
	if (curbp->b_fname[0] == NULLCHR)	/* must have a name */
	{
		MlWrite("No file name");
		return(FALSE);
	}

	/* complain about truncated files */
	if ((curbp->b_flag & BFTRUNC) != 0)
	{
		if (MlYesNo("Truncated file. Write it out") == FALSE)
		{
			CmdAbort(FALSE, 1);
			return(FALSE);
		}
	}

	/* complain about narrowed files */
	if ((curbp->b_flag & BFNAROW) != 0)
	{
		if (MlYesNo("Narrowed buffer. Write it out") == FALSE)
		{
			CmdAbort(FALSE, 1);
			return(FALSE);
		}
	}

	if (WriteOut(curbp->b_fname))
	{
		curbp->b_flag &= ~BFCHG;

		/* Update all mode lines */
		UpMode();
		return(TRUE);
	}
	return(FALSE);
}

/**
 * This function performs the details of file writing. The number of lines
 * written is displayed.
 */
int WriteOut(filnam)
  char	*filnam;
{
  char	*tfn;			/* temp file name for safe save */
  char	*cp1,*cp2,*cp3;	/* working pointers */
  char	fn[NFILEN];		/* file name to open */
  HDL	ffh;			/* PforCe file handle */
  LINE	*lp;
  int	nline;
  char	buf[NSTRING];
  char	tmpfnam[9];		/* IMPORTANT!	9 = sizeof("ceXXXXXX") */

	if (safesave)
	{
		stracpy(tmpfnam, "ceXXXXXX");
		tfn = mktemp(tmpfnam);

		/* get just the path part of filnam */
		cp1 = &filnam[stralen(filnam) - 1];
		while	(
						cp1 != &filnam[0]
					&&
						*(cp1-1) != ':'
					&&
						*(cp1-1) != '\\'
					&&
						*(cp1-1) != '/'
				)
			--cp1;

		/* if there was a path, NULL out the file name part */
		if (cp1 != &filnam[0])
		{
			cp2 = &filnam[0];
			cp3 = &fn[0];
			while (cp2 != cp1)
				*cp3++ = *cp2++;
			*cp3 = NULLCHR;			/* NULL terminate it */
		}
		else
			fn[0] = NULLCHR;		/* init it for the stracat() below */

		/* make the fully qualified temp file name */
		stracat(fn, tfn);
	}
	else
		stracpy(fn, filnam);

	if ((ffh = flopen(fn, "w")) == NULLHDL)
	{
		MlWrite("Can't open file for writing");
		return(FALSE);
	}

	MlWrite("Writing");				/* tell us we're writing */
	lp = lforw(curbp->b_linep); 	/* first line */
	nline = 0;						/* number of lines */
	while (lp != curbp->b_linep)
	{
		flput(&lp->l_text[0], llength(lp), ffh);
		flput("\r\n", 2, ffh);
		++nline;
		lp = lforw(lp);
	}
	flclose(ffh);

	if (safesave)
	{
		if (FileExist(filnam))					/* if the old version exists */
		{
			if (drunlink(filnam) == RC_FAIL)	/* delete it */
			{
				strfmt(buf, "Can't unlink %s, %s not renamed", filnam, fn);
				MlWrite(buf);
				return(FALSE);
			}
		}
		if (drrename(fn, filnam) == RC_FAIL)	/* rename temp to real file */
		{
			strfmt(buf, "Can't rename %s to %s", fn, filnam);
			MlWrite(buf);
			return(FALSE);
		}
	}

	strfmt(buf, "Wrote %d lines", nline);
	MlWrite(buf);

	return(TRUE);
}

/**
 * This function allows the user to modify the file name associated with the
 * current buffer. The operation is simple; just zap the name in the BUFFER
 * structure, and mark the windows as needing an Update. You can type a blank
 * line at the prompt if you wish.
 */
int FileName(f, n)
  int		f, n;
{
  WINDOW	*wp;
  int		s;
  char		fname[NFILEN];

	if ((s = MlReply("Name: ", fname, NFILEN)) == ABORT)
		return(s);
	if (s == FALSE)
		stracpy(curbp->b_fname, "");
	else
		stracpy(curbp->b_fname, fname);

	/* update all mode lines */
	UpMode();
	curbp->b_mode &= ~MDVIEW;		/* no longer read only mode */
	return(TRUE);
}

/**
 * See if a file exists
 */
int FileExist(fname)
  char	*fname;
{
  HDL	fl;

	if ((fl = flopen(fname, "r+")) == NULLHDL)	/* file not found */
		return(FALSE);
	flclose(fl);								/* else it exists */
	return(TRUE);								/* so return TRUE */
}

/**
 * Insert file "fname" into the current buffer (called by insert file
 * function). Return the final status of the read.
 */
static int near ifile(fname)
  char		fname[];
{
  LINE		*lp0;
  LINE		*lp1;
  LINE		*lp2;
  BUFFER	*bp;
  int		s;
  int		nbytes;
  int		nline;
#if BMARKS
  int		bmark;
#endif
  char		buf[NSTRING];

	bp = curbp; 						/* cheap */
	bp->b_flag |= BFCHG;				/* we have changed */
	bp->b_flag &= ~BFINVS;				/* and are not temporary */

	if ((fp = fopen(fname, "rb")) == (FILE *)NULLCPTR)
	{
		MlWrite("No such file");		/* file not found */
		return(FALSE);
	}
	MlWrite("Inserting file");

	/* back up a line and save the mark here */
	curwp->w_dotp = lback(curwp->w_dotp);
	curwp->w_doto = 0;
	curwp->w_markp = curwp->w_dotp;
	curwp->w_marko = 0;

	nline = 0;
	while ((s = ffgetline()) == FIOSUC)
	{
		nbytes = stralen(linebuf);
		if ((lp1 = LAlloc(nbytes)) == NULLLPTR)
		{
			s = FIOMEM; 		/* keep message on the display */
			break;
		}
		lp0 = curwp->w_dotp;	/* line previous to insert */
		lp2 = lp0->l_fp;		/* line after insert */

		/* re-link new line between lp0 and lp2 */
		lp2->l_bp = lp1;
		lp0->l_fp = lp1;
		lp1->l_bp = lp0;
		lp1->l_fp = lp2;

		/* and advance and write out the current line */
		curwp->w_dotp = lp1;
		memcpy(lp1->l_text, linebuf, nbytes);
		++nline;
	}
	fclose(fp);				/* ignore errors */

	curwp->w_markp = lforw(curwp->w_markp);

	buf[0] = NULLCHR;			/* in case there is no file i/o error */
	if (s == FIOMEM)
	{
		stracpy(buf, "Out of memory, ");
		curbp->b_flag |= BFTRUNC;
		PcPutC(BELL);
	}
	strfmt(&buf[stralen(buf)], "Inserted %d lines", nline);
	MlWrite(buf);

	/* advance to the next line and mark the window for changes */
	curwp->w_dotp = lforw(curwp->w_dotp);
	curwp->w_flag |= WFHARD | WFMODE;

	/* copy window parameters back to the buffer structure */
	curbp->b_dotp = curwp->w_dotp;
	curbp->b_doto = curwp->w_doto;
	curbp->b_markp = curwp->w_markp;
	curbp->b_marko = curwp->w_marko;
	curbp->b_anchorp = curwp->w_anchorp;
	curbp->b_anchoro = curwp->w_anchoro;
	curbp->b_fcol = curwp->w_fcol;
#if BMARKS
	for (bmark = 0; bmark < NBMARKS; bmark++)
	{
		curbp->b_bmarkp[bmark] = curwp->w_bmarkp[bmark];
		curbp->b_bmarko[bmark] = curwp->w_bmarko[bmark];
	}
#endif
	return(TRUE);
}

/**
 * Read a line from a file, and store the bytes in the supplied buffer
 */
static int near ffgetline()
{
  char	*lbuf;
  int	lbufsiz;
  int	c;

	if (!feof(fp))
	{
		lbuf = &linebuf[0];
		lbufsiz = sizeof(linebuf);
		while	(
						(c = getc(fp)) != EOF
					&&
						c != (int)'\r'
					&&
						c != (int)'\n'
					&&
						--lbufsiz != 0	/* pre-dec it (leave room for NULLCHR)*/
				)
			*lbuf++ = (char)c;

		/* check for EOF */
		if (c == EOF)
		{
			if (lbufsiz == sizeof(linebuf))		/* no chars read */
				return(FIOEOF);
			else
			{
				*lbuf = NULLCHR;		/* null terminate it */
				return(FIOSUC);			/* and return (feof fails next time) */
			}
		}

		/*
		 * handle all combinations of <cr> and <lf> as line terminators
		 *
		 *	1.	<cr> <lf>
		 *	2.	<lf> <cr>
		 *	3.	<cr>
		 *	4.	<lf>
		 */
		if ((c = getc(fp)) != (int)'\n' && c != (int)'\r')
			ungetc(c, fp);

		*lbuf = NULLCHR;
		return(FIOSUC);
	}
	return(FIOEOF);
}

/**
 * Determine whether or not we have a valid DOS file name
 */
static int near validfilename(fname)
  char	*fname;
{
  int	fc, ec;
  char	*cp1;

	fc = ec = 0;

	cp1 = fname;

	while (*cp1 && *cp1 == '.')	/* bump past any leading . or .. chars */
		++cp1;

	while (*cp1 && *cp1 != '.')
		++cp1;

	if (*cp1 == '.')			/* we have a file extension */
	{
		while (*cp1)			/* point at the terminating NULL of fname */
			++cp1;

		while (*(cp1-1) != '.')
		{
			if (++ec > 3)		/* too many characters in the extension */
				return(FALSE);
			--cp1;
		}

		if (*(cp1-1) == '.')	/* bump past the '.' */
			--cp1;
	}

	while	(
					cp1 != fname
				&&
					*(cp1-1) != ':'
				&&
					*(cp1-1) != '\\'
				&&
					*(cp1-1) != '/'
			)
	{
		if (++fc > 8)			/* too many characters in the file name */
			return(FALSE);
		--cp1;
	}

	return(TRUE);			/* got a valid file name */
}
