/*
 *	F U N C . H
 *
 *	purpose:
 *		CodEd function declarations and names.
 *
 *	technical notes:
 *		This file lists the externally available C code functions used by CodEd
 *		and the names to use to bind keys to them. To add functions, declare it
 *		here in both the extern function list and the name binding table.
 *
 *		Include this file AFTER protos.h !!!
 */

/* Name to function binding table
 *
 *	 This table gives the names of all the bindable functions and their C
 *	 function address. These are used for the bind-to-key function.
 */

NBIND	names[] =
{
	{"abort-cmnd",			CmdAbort},
	{"add-gmode", 			AddGMode},
	{"add-mode",			AddMode},
	{"ascii-table",			AsciiTable},
	{"back-char",			BackChar},
	{"begin-mac", 			BegKbdMac},
	{"bind-to-key", 		BindToKey},
	{"case-region-lower",	LowerRegion},
	{"case-region-upper",	UpperRegion},
	{"case-word-cap",		CapWord},
	{"case-word-lower", 	LowerWord},
	{"case-word-upper", 	UpperWord},
	{"change-file-name",	FileName},
	{"clear-and-redraw",	Refresh},
	{"clear-msg-line",		ClrMsg},
	{"copy-region", 		CopyRegion},
	{"ctlx-prefix", 		Cex},
	{"del-blank-lines",		Deblank},
	{"del-bufr",			KillBuffer},
	{"del-gmode",			DelGMode},
	{"del-mode", 			DelMode},
	{"del-next-char",		ForwDel},
	{"del-next-word",		DelFWord},
	{"del-other-winds",		OnlyWind},
	{"del-prev-char",		BackDel},
	{"del-prev-word",		DelBWord},
	{"del-wind",			DelWind},
	{"desc-key",			DesKey},
	{"detab-line",			Detab},
	{"end-mac",				EndKbdMac},
	{"end-of-file", 		GotoEOB},
	{"end-of-line", 		GotoEOL},
	{"entab-line",			Entab},
	{"exec-bufr",			ExecBuf},
	{"exec-cmnd-line",		ExecCmd},
	{"exec-file",			ExecFile},
	{"exec-mac",			ExecKbdMac},
	{"exec-mac-1", 			ExecMac1},
	{"exec-mac-10",			ExecMac10},
	{"exec-mac-11",			ExecMac11},
	{"exec-mac-12",			ExecMac12},
	{"exec-mac-13",			ExecMac13},
	{"exec-mac-14",			ExecMac14},
	{"exec-mac-15",			ExecMac15},
	{"exec-mac-16",			ExecMac16},
	{"exec-mac-17",			ExecMac17},
	{"exec-mac-18",			ExecMac18},
	{"exec-mac-19",			ExecMac19},
	{"exec-mac-2", 			ExecMac2},
	{"exec-mac-20",			ExecMac20},
	{"exec-mac-21",			ExecMac21},
	{"exec-mac-22",			ExecMac22},
	{"exec-mac-23",			ExecMac23},
	{"exec-mac-24",			ExecMac24},
	{"exec-mac-25",			ExecMac25},
	{"exec-mac-26",			ExecMac26},
	{"exec-mac-27",			ExecMac27},
	{"exec-mac-28",			ExecMac28},
	{"exec-mac-29",			ExecMac29},
	{"exec-mac-3", 			ExecMac3},
	{"exec-mac-30",			ExecMac30},
	{"exec-mac-31",			ExecMac31},
	{"exec-mac-32",			ExecMac32},
	{"exec-mac-33",			ExecMac33},
	{"exec-mac-34",			ExecMac34},
	{"exec-mac-35",			ExecMac35},
	{"exec-mac-36",			ExecMac36},
	{"exec-mac-37",			ExecMac37},
	{"exec-mac-38",			ExecMac38},
	{"exec-mac-39",			ExecMac39},
	{"exec-mac-4", 			ExecMac4},
	{"exec-mac-40",			ExecMac40},
	{"exec-mac-5", 			ExecMac5},
	{"exec-mac-6", 			ExecMac6},
	{"exec-mac-7", 			ExecMac7},
	{"exec-mac-8", 			ExecMac8},
	{"exec-mac-9", 			ExecMac9},
	{"exec-named-cmnd",		NamedCmd},
	{"exec-proc",			ExecProc},
	{"exec-prog",			Exec},
	{"exit-coded",			Quit},
	{"find-file",			FileFind},
	{"for-char",			ForwChar},
	{"goto-line",			GotoLine},
	{"goto-mark",			GotoMark},
	{"goto-matching-fence", GetFence},
	{"grow-wind", 			EnlargeWind},
	{"help",				Help},
	{"hunt-next",			ForwHunt},
	{"hunt-prev",			BackHunt},
	{"i-shell", 			DosShell},
	{"ins-file", 			InsFile},
	{"ins-space",			InsSpace},
	{"ins-str",				InsString},
	{"kill-region", 		KillRegion},
	{"kill-to-eol", 		KillText},
	{"list-bufrs",			ListBuffers},
	{"meta-prefix", 		Meta},
	{"move-wind-down",		MoveWindDn},
	{"move-wind-up",		MoveWindUp},
	{"name-bufr", 			NameBuffer},
	{"narrow-to-region",	Narrow},
	{"newline", 			NewLine},
	{"newline-and-indent",	Indent},
	{"next-bufr", 			NextBuffer},
	{"next-line",			ForwLine},
	{"next-page",			ForwPage},
	{"next-wind", 			NextWind},
	{"next-word",			ForwWord},
	{"nop", 				NullProc},
	{"open-line",			OpenLine},
	{"over-str",			OvString},
	{"pipe-cmd",			Pipe},
	{"prev-line",			BackLine},
	{"prev-page",			BackPage},
	{"prev-wind", 			PrevWind},
	{"prev-word",			BackWord},
	{"query-replace-str",	QReplace},
	{"quick-exit",			QuickExit},
	{"quote-char",			Quote},
	{"read-file",			FileRead},
	{"redraw-display",		Reposition},
	{"rem-mark",			DelMark},
	{"replace-str",			SReplace},
	{"resize-wind",			Resize},
	{"restore-wind",		RestWind},
	{"save-file",			FileSave},
	{"save-wind", 			SaveWind},
	{"scroll-next-down",	NextWindDn},
	{"scroll-next-up",		NextWindUp},
	{"select-bufr",			UseBuffer},
	{"set", 				SetVar},
	{"set-mark",			SetMark},
	{"shell-cmnd",			Spawn},
	{"shrink-wind",			ShrinkWind},
	{"split-current-wind",	SplitWind},
	{"srch-for",			ForwSearch},
	{"srch-rev",			BackSearch},
	{"store-mac", 			StoreMac},
	{"store-proc",			StoreProc},
	{"swap-dot-and-mark", 	SwapMark},
	{"tab",					Tab},
	{"top-of-file",			GotoBOB},
	{"top-of-line",			GotoBOL},
	{"transpose-chars",		Twiddle},
	{"trim-line",			Trim},
	{"unbind-key",			UnbindKey},
	{"unmark-bufr",			UnmarkBuf},
	{"update-screen",		UpdateScr},
	{"view-file",			ViewFile},
	{"widen-from-region",	Widen},
	{"write-file",			FileWrite},
	{"write-msg",			WriteMsg},
	{"yank",				Yank},
	{"",					NULLFPTR}
};

#define	NNAMES	(sizeof(names) / sizeof(NBIND)) - 1
