/*
 *	I N P U T . C
 *
 *	purpose:
 *		Various input routines
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

/*
 * local function prototypes
 */

static void near	delleft(int);

/**
 * Ask a yes or no question in the message line. Return either TRUE, FALSE, or
 * ABORT. The ABORT status is returned if the user bumps out of the question
 * with an alt-A. Used any time a confirmation is required.
 */
int MlYesNo(prompt)
  char	*prompt;
{
  int	c;				   			/* input character */
  char	buf[NSTRING];

	/* build and prompt the user */
	strcpycat(buf, prompt, " [y/n]? ");
	MlWrite(buf);

	/* get the response */
	c = TermGetCh();

	if (c == ExtChToCh(abortc)) 	/* bail out! */
		return(ABORT);

	if (c == 'y' || c == 'Y')
		return(TRUE);

	return(FALSE);					/* if not a yes */
}

/**
 * Write a prompt into the message line, then read back a response. Keep
 * track of the physical position of the cursor. If we are in a keyboard
 * macro throw the prompt away, and return the remembered response. This
 * lets macros run at full speed. The reply is always terminated by a
 * carriage return. Handle erase, kill, and abort keys.
 */
int MlReply(prompt, buf, nbuf)
  char	*prompt;
  char	*buf;
{
	return(NextArg(prompt, buf, nbuf, ChToExtCh('\n')));
}

/**
 * Same as MlReply() except that you get to specify the terminating character
 */
int MlReplyT(prompt, buf, nbuf, eolchar)
  char	*prompt;
  char	*buf;
  int	eolchar;
{
	return(NextArg(prompt, buf, nbuf, eolchar));
}

/**
 * Convert extended character to ascii character. Collapse the CTRL and SPEC
 * flags back into an ascii code.
 */
int ExtChToCh(c)
  int	c;
{
	if (c & CTRL)
		c &= ~(CTRL | 0x40);
	if (c & SPEC)
		c &= ~SPEC;
	return(c);
}

/**
 * Convert character to extended character
 */
int ChToExtCh(c)
  int	c;
{
	if (c >= 0x00 && c <= 0x1f)
		c = CTRL | (c + 0x40);
	return(c);
}
 
/**
 * Get a command name from the command line
*/
int (* GetName())()
{
  int	cpos;			/* current column on screen output */
  int	c;
  char	*sp; 			/* pointer to string for output */
  NBIND	*ffp;			/* first ptr to entry in name binding table */
  NBIND *cffp;			/* current ptr to entry in name binding table */
  NBIND *lffp;			/* last ptr to entry in name binding table */
  char	buf[NSTRING];	/* buffer to hold tentative command name */

	/* start at the beginning of the string buffer */
	cpos = 0;

	/* if we are executing a command line, get the next arg and match it */
	if (clexec)
	{
		if (MacArg(buf) != TRUE)
			return(FALSE);
		return(FuncMatch(&buf[0]));
	}

	/* build a name string from the keyboard */
	while (TRUE)
	{
		c = ExtChToCh(TermGetCh());

		/* if we are at the end, just match it */
		if (c == 0x0d)
		{
			buf[cpos] = 0;

			/* and match it off */
			return(FuncMatch(&buf[0]));

		}
		else if (c == ExtChToCh(abortc))	/* bell, abort */
		{
			CmdAbort(TRUE, 1);
			return(NULLFPTR);
		}
		else if (c == 0x7f || c == 0x08)	/* rubout/erase */
		{
			if (cpos != 0)
			{
				delleft(1);
				--cpos;
			}
		}
		else
		{
			if (cpos < NSTRING - 1 && c > 0x20)
			{
				buf[cpos++] = (char)c;
				PcPutC(c);
			}
			++ttcol;
		}
	}
}

/**
 * Get a key from the terminal driver, resolve any keyboard macro action
 */
int TermGetCh()
{
  int	c;			/* fetched character */

	/* if we are playing a keyboard macro back */
	if (kbdmode == PLAY)
	{
		/* if there is some left */
		if (kbdptr < kbdend)
			return((int)*kbdptr++);

		/* if at the end of last repetition */
		if (--kbdrep < 1)
		{
			kbdmode = STOP;
			/* force a screen Update after all is done */
			Update(FALSE);
		}
		else
		{
			/* reset the macro to the beginning for the next rep */
			kbdptr = &kbdm[0];
			return((int)*kbdptr++);
		}
	}

	/* fetch a character from the terminal driver */
	c = PcGetC();

	/* record it for $lastkey */
	lastkey = c;

	/* save it if we need to */
	if (kbdmode == RECORD)
	{
		*kbdptr++ = c;
		kbdend = kbdptr;

		/* don't overrun the buffer */
		if (kbdptr == &kbdm[NKBDM - 1])
		{
			kbdmode = STOP;
			PcPutC(BELL);
		}
	}

	/* and finally give the char back */
	return(c);
}

/**
 * Get a command from the keyboard. Process all applicable prefix keys.
 */
int GetCmd()
{
  int	c;			/* fetched keystroke */

	/* get initial character */
	c = TermGetCh();

	/* process META prefix */
	if (c == metac)
	{
		c = TermGetCh();
		if (c >= 'a' && c <= 'z')
			c ^= 0x20;					/* force to upper */
		return(META | c);
	}

	/* process CTLX prefix */
	if (c == ctlxc)
	{
		c = TermGetCh();
		if (c >= 'a' && c <= 'z')
			c ^= 0x20;					/* force to upper */
		return(CTLX | c);
	}

	/* otherwise, just return it */
	return(c);
}

/**
 * A more generalized prompt/reply function allowing the caller to specify the
 * proper terminator. If the terminator is not a return ('\n'), it will echo as
 * "<NL>"
 */
int GetString(prompt, buf, nbuf, eolchar)
  char	*prompt;
  char	*buf;
  int	eolchar;
{
  int	cpos;		/* current character position in string */
  int	c;
  int	quotef; 	/* are we quoting the next char? */

	cpos = 0;
	quotef = FALSE;

	/* prompt the user for the input string */
	MlWrite(prompt);

	for (;;)
	{
		/* get a character from the user */
		c = TermGetCh();

		/* if it is a <CR>, change it to a <NL> */
		if (c == (CTRL | 0x40 | '\r'))
			c = CTRL | 0x40 | '\n';

		/* if we hit the line terminator, wrap it up */
		if (c == eolchar && quotef == FALSE)
		{
			/* null terminate it */
			buf[cpos] = 0;

			/* clear the message line */
			MlWrite("");

			/* if we default the buffer, return FALSE */
			if (buf[0] == 0)
				return(FALSE);
			return(TRUE);
		}

		/* change from command form back to character form */
		c = ExtChToCh(c);

		if (c == ExtChToCh(abortc) && quotef == FALSE)
		{
			/* abort the input? */
			CmdAbort(TRUE, 1);
			return(ABORT);
		}
		else if (c == 0x7f && quotef == FALSE)
		{
			buf[0] = 0;
			return(TRUE);
		}
		else if (c == 0x08 && quotef == FALSE)
		{
			/* rubout/erase */
			if (cpos != 0)
			{
				delleft(1);

				if (buf[--cpos] < 0x20)
					delleft(1);

				if (buf[cpos] == '\n')
					delleft(2);
			}
		}
		else if (c == quotec && quotef == FALSE)
		{
			quotef = TRUE;
		}
		else
		{
			quotef = FALSE;
			if (cpos < nbuf - 1)
			{
				buf[cpos++] = (char)c;
				if ((c < ' ') && (c != '\n'))
				{
					OutString("^");
					++ttcol;
					c ^= 0x40;
				}
				if (c != '\n')
				{
					if (disinp)
						PcPutC(c);
				}
				else
				{
					/* put out <NL> for <ret> */
					OutString("<NL>");
					ttcol += 3;
				}
				++ttcol;
			}
		}
	}
}

/**
 * Output a string of input characters
 */ 
int OutString(s)
  char	*s;					/* string to output */
{
	if (disinp)
		while (*s)
			PcPutC(*s++);
	return(TRUE);
}

/**
 * Output a string of output characters
 */
int OString(s)
  char	*s;					/* string to output */
{
	if (discmd)
		while (*s)
			PcPutC(*s++);
	return(TRUE);
}

/**
 * Delete 'nchars' to the left of the cursor and decrement 'ttcol' by 'nchars'
 */
static void near delleft(nchars)
  int	nchars;				/* number of ouput chars to the left to delete */
{
	if (nchars > 0)
	{
		ttcol -= nchars;	/* must do this first */
		while (nchars--)
			OutString("\b \b");
	}
}
