/*
 *  K D E F . H
 *
 *  purpose:
 *		Definitions for the various special keys on the keyboard.  A special
 *		key is defined as having a main byte of 0x00 or 0xe0, followed by an
 *		auxilliary byte != 0.
 *
 *	technical notes:
 *		The enhanced special keys (from the 101/102 key keyboard) can be
 *		retrieved via a BIOS call to INT 22 (16 hex) using service codes (ah=)
 *		16, 17 and 18 (10, 11 and 12 hex) instead of 0, 1 and 2 (see bios.asm).
 */

/* definitions for special and extended special keys */

#define	_F1			59			/* F1 to F12 */
#define	_F2			60
#define	_F3			61
#define	_F4			62
#define	_F5			63
#define	_F6			64
#define	_F7			65
#define	_F8			66
#define	_F9			67
#define	_F10		68
#define	_F11		133
#define _F12		134

#define	_sF1		84			/* Shift-F1 to Shift-F12 */
#define	_sF2		85
#define	_sF3		86
#define	_sF4		87
#define	_sF5		88
#define	_sF6		89
#define	_sF7		90
#define	_sF8		91
#define	_sF9		92
#define	_sF10		93
#define	_sF11		135
#define	_sF12		136

#define	_cF1		94			/* Control-F1 to Control-F12 */
#define	_cF2		95
#define	_cF3		96
#define	_cF4		97
#define	_cF5		98
#define	_cF6		99
#define	_cF7		100
#define	_cF8		101
#define	_cF9		102
#define	_cF10		103
#define	_cF11		137
#define	_cF12		138

#define	_aF1		104			/* Alt-F1 to Alt-F12 */
#define	_aF2		105
#define	_aF3		106
#define	_aF4		107
#define	_aF5		108
#define	_aF6		109
#define	_aF7		110
#define	_aF8		111
#define	_aF9		112
#define	_aF10		113
#define	_aF11		139
#define	_aF12		140

#define	_aA			30			/* Alt-A to Alt-Z */
#define	_aB			48
#define	_aC			46
#define	_aD			32
#define	_aE			18
#define	_aF			33
#define	_aG			34
#define	_aH			35
#define	_aI			23
#define	_aJ			36
#define	_aK			37
#define	_aL			38
#define	_aM			50
#define	_aN			49
#define	_aO			24
#define	_aP			25
#define	_aQ			16
#define	_aR			19
#define	_aS			31
#define	_aT			20
#define	_aU			22
#define	_aV			47
#define	_aW			17
#define	_aX			45
#define	_aY			21
#define	_aZ			44

#define	_a1			120			/* Alt-1 to Alt-0 */
#define	_a2			121
#define	_a3			122
#define	_a4			123
#define	_a5			124
#define	_a6			125
#define	_a7			126
#define	_a8			127
#define	_a9			128
#define	_a0			129

#define	_aHYPHEN	130			/* Alt-- */
#define	_aEQUALS	131			/* Alt-= */

#define	_INSERT		82
#define	_DELETE		83

#define	_HOME		71			/* home */
#define	_END		79			/* end */
#define	_PGUP		73			/* page up */
#define	_PGDN		81			/* page down */
#define	_LEFT		75			/* left arrow */
#define	_RIGHT		77			/* right arrow */
#define	_UP			72			/* up arrow */
#define	_DOWN		80			/* down arrow */

#define	_cHOME		119			/* Control-home */
#define	_cEND		117			/* Control-end */
#define	_cPGUP		132			/* Control-page up */
#define	_cPGDN		118			/* Control-page down */
#define	_cLEFT		115			/* Control-left arrow */
#define	_cRIGHT		116			/* Control-right arrow */
#define	_cUP		141			/* Control-up arrow */
#define	_cDOWN		145			/* Control-down arrow */

#define	_aHOME		151			/* Alt-home */
#define	_aEND		159			/* Alt-end */
#define	_aPGUP		153			/* Alt-page up */
#define	_aPGDN		161			/* Alt-page down */
#define	_aLEFT		155			/* Alt-left arrow */
#define	_aRIGHT		157			/* Alt-right arrow */
#define	_aUP		152			/* Alt-up arrow */
#define	_aDOWN		160			/* Alt-down arrow */

#define	_sTAB		15			/* Shift-Tab */

#define _cPRTSC		114			/* Control-PrtSc */

#define _cGREYPLUS	144			/* Control-Grey+ */
#define _cGREYMINUS	142			/* Control-Grey- */
#define _cGREYSTAR	150			/* Control-Grey* */
#define _cGREYSLASH	149			/* Control-Grey/ */

#define _aGREYPLUS	78			/* Alt-Grey+ */
#define _aGREYMINUS	74			/* Alt-Grey- */
#define _aGREYSTAR	55			/* Alt-Grey* */
#define _aGREYSLASH	164			/* Alt-Grey/ */
