/*
 *	M A I N . C
 *
 *	Copyright (C) 1988-1991 John W Grothman. All rights reserved.
 *
 *	purpose:
 *		This file contains the main driving routine, and some keyboard
 *		processing code, for CodEd.
 *
 *	to do:
 *		1.	add highlighting for the current region
 *		2.	use a seperate set of <dot><mark> to delineate a region (see above)
 *		3.	speed up scrolling
 *		4.	speed up macros
 *		5.	line number mode
 *
 *	technical notes:
 *
 *	other notes:
 *		CodEd is based, in part, on MicroEmacs 3.9 and 3.10 (by Daniel M.
 *		Lawrence, et. al). It was a good place to start, but the code needed
 *		some work and I wanted to take full advantage of the IBM-PC, add some
 *		new features, and speed up file I/O and searching. Note that all 3.9
 *		and 3.10 functions having nothing to do with writing code are toast.
 *		CodEd is a source code editor. 'Nough said?
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <stdlib.h>
#include <alloc.h>
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

/* make global definitions not external */

#define maindef

#include "struct.h"			/* global structures and defines */
#include "protos.h"			/* function prototypes */
#include "func.h"			/* name table (AFTER protos.h and BEFORE def.h) */

#ifdef EKBD
#include "bind.h"			/* default key bindings (include AFTER protos.h) */
#endif

#ifdef OKBD
#include "obind.h"			/* default key bindings (include AFTER protos.h) */
#endif

#include "def.h"			/* global definitions */

/*
 * local function prototypes
 */

static int near		anycb(void);
static int near		execute(int, int, int);
static int near		startup(char *);
static void near	edinit(char *);
static void near	editloop(void);
static void near	procline(int, char **);
static void near	putusage(void);

/*
 * let TURBO C++ know what the stack size is
 */

extern	unsigned int		_stklen = 16384;

/**
 * All good things have to start somewhere
 */
void main(argc, argv)
  int		argc;
  char		**argv;
{
	/* initialize the editor */
	VtInit();				/* display */
	edinit("new");			/* buffers, windows */
	VarInit();				/* user variables */

	procline(argc, argv);	/* process the DOS command line */
	editloop();				/* and go to work (never to return) */
}

/**
 * Initialize all of the buffers and windows. The buffer name is passed down
 * as an argument, because the main routine may have been told to read in a
 * file by default, and we want the buffer name to be right.
 */
static void near edinit(bname)
  char		bname[];
{
  BUFFER	*bp;
  WINDOW	*wp;

	bp = FindBuf(bname, TRUE, 0); 						/* first buffer */
	blistp = FindBuf("[buffer list]", TRUE, BFINVS);	/* make invisible */
	FindBuf("[function keys]", TRUE, BFINVS); 			/* make invisible */
	wp = (WINDOW *)malloc(sizeof(WINDOW));				/* first window */
	if (bp == NULLBPTR || wp == NULLWPTR || blistp == NULLBPTR)
	{
		PcClose();
		exit(FAILURE);
	}
	curbp = bp; 							/* make this current */
	wheadp = wp;
	curwp = wp;
	wp->w_wndp = NULLWPTR;					/* initialize window */
	wp->w_bufp = bp;
	bp->b_nwnd = 1; 						/* displayed */
	wp->w_linep = bp->b_linep;
	wp->w_dotp = bp->b_linep;
	wp->w_doto = 0;
	wp->w_markp = NULLLPTR;
	wp->w_marko = 0;
	wp->w_anchorp = NULLLPTR;
	wp->w_anchoro = 0;
	wp->w_toprow = 0;
	wp->w_fcolor = (char)gfcolor; 			/* init colors to global defaults */
	wp->w_bcolor = (char)gbcolor;
	wp->w_fcol = 0;
	wp->w_ntrows = (char)(term.t_nrow - 1); /* "-1" for mode line */
	wp->w_force = 0;
	wp->w_flag = WFMODE | WFHARD;			/* full */
	return;
}

/**
 * Process the DOS command line
 */
static void near procline(argc, argv)
  int		argc;
  char		**argv;
{
  char		*clarg;			/* ptr to command line argument */
  BUFFER	*bp; 			/* temp buffer pointer */
  int		firstfile;		/* first file flag */
  int		carg;			/* current arg to scan */
  int		startflag;		/* startup executed flag */
  BUFFER	*firstbp;		/* ptr to first buffer in cmd line */
  int		viewflag;		/* are we starting in view mode? */
  int		gotolflag;		/* need to goto a line at start? */
  int		gline;			/* if so, what line? */
  int		searchflag; 	/* do we need to search at start? */
  int		overflag;		/* put buffer in overwrite mode? */
  int		magicflag;		/* put buffer in magic search mode? */
  char		bname[NBUFN];	/* buffer name of file to read */
  char		buf[NSTRING];	/* message buffer */

	firstbp = NULLBPTR;

	viewflag = FALSE;
	gotolflag = FALSE;
	searchflag = FALSE;
	overflag = FALSE;
	magicflag = FALSE;
	firstfile = TRUE;		/* no file to edit yet */
	startflag = FALSE;		/* startup file not executed yet */

	/* parse the command line */
	for (carg = 1; carg < argc; ++carg)
	{
		clarg = argv[carg];

		/* process switches */
		if (*clarg == '-')
		{
			switch (*(clarg + 1) | 0x20)	/* force to lower case */
			{
				case 'e':		/* -e for edit file */
					viewflag = FALSE;
					break;
				case 'g':		/* -g for initial goto line */
					gotolflag = TRUE;
					gline = cvtatoi(clarg + 2);
					break;
				case 'i':		/* -i for insert mode */
					overflag = FALSE;
					break;
				case 'm':		/* -m for magic search mode */
					magicflag = TRUE;
					break;
				case 'n':		/* -n for normal search mode */
					magicflag = FALSE;
					break;
				case 'o':		/* -o for overwrite mode */
					overflag = TRUE;
					break;
				case 's':		/* -s for initial search string */
					searchflag = TRUE;
					strancpy(pat, clarg + 2, NPAT);	/* init search pattern */
					SetJmpTbl(pat);					/* init fast search arrays*/
					break;
				case 'v':		/* -v for view file */
					viewflag = TRUE;
					break;
				case 'h':		/* -h for print usage and exit */
				default:		/* unknown switch */
					putusage();
					PcClose();
					exit(FAILURE);
					break;
			}

		}
		else if (*clarg == '@')
		{
			/* process startup macros */
			if (startup(clarg + 1) == TRUE)
				startflag = TRUE;
		}
		else
		{
			/* process an input file */

			/* set up a buffer for this file */
			strlower(clarg);
			/* if filename good, or user said use it */
			if (MakeName(bname, clarg))
			{
				UniqName(bname);

				/* set this to inactive */
				bp = FindBuf(bname, TRUE, 0);
				stracpy(bp->b_fname, clarg);
				bp->b_active = FALSE;
				if (firstfile)
				{
					firstbp = bp;
					firstfile = FALSE;
				}

				/* set the modes appropriately */
				if (magicflag)
					bp->b_mode |= MDMAGIC;
				if (overflag)
					bp->b_mode |= MDOVER;
				if (viewflag)
					bp->b_mode |= MDVIEW;
			}
		}
	}

	/* if invoked with no other startup files, run system startup file here */
	if (startflag == FALSE)
		startup("");

	/* if there are any files to read, read the first one! */
	bp = FindBuf("new", FALSE, 0);
	if (firstfile == FALSE)
	{
		SwapBuffer(firstbp);
		curbp->b_mode |= gmode;
		Update(TRUE);
		MlWrite(lstmsg);
		ZotBuf(bp);
	}
	else
		bp->b_mode |= gmode;

	/* deal with startup gotos and searches */
	if (gotolflag && searchflag)
	{
		Update(FALSE);
		MlWrite("Can't SEARCH and GOTO at the same time");
	}
	else if (gotolflag)
	{
		if (GotoLine(TRUE, gline) == FALSE)
		{
			Update(FALSE);
			MlWrite("Bogus GOTO line argument");
		}
	}
	else if (searchflag)
	{
		if (ForwHunt(FALSE, 0) == FALSE)
			Update(FALSE);
	}
}

/**
 * Execute the startup file
 */
static int near startup(sfname)
  char	*sfname; 	/* name of startup file (null if default) */
{
  char	*fname;		/* resulting file name to execute */

	/* look up the startup file */
	if (*sfname != NULLCHR)
		fname = FileLook(sfname, TRUE);
	else
	{
#ifdef EKBD
		fname = FileLook("ce.sf", TRUE);
#endif
#ifdef OKBD
		fname = FileLook("ceo.sf", TRUE);
#endif
	}

	/* if it isn't around, don't sweat it */
	if (fname == NULLCPTR)
		return(TRUE);

	/* otherwise, execute the sucker */
	return(DoFile(fname));
}

/**
 * Let the user edit
 */
static void near editloop()
{
  int		c;				/* command character */
  int		f;				/* default flag */
  int		n;				/* numeric repeat count */
  int		mflag;			/* negative flag on repeat */
  int		basec;			/* c stripped of Meta character */
  char		buf[NSTRING];

	/* setup to process commands */
	lastflag = 0;							/* fake last flags */

loop:

	/* fix up the screen */
	Update(FALSE);

	/* get the next command from the keyboard */
	c = GetCmd();

	/* if there is something on the command line, clear it */
	if (mpresf != FALSE)
	{
		MlErase();
		Update(FALSE);
	}
	f = FALSE;
	n = 1;

	/* do META-# processing if needed */

	basec = c & ~META;						/* strip Meta char if there is one*/
	if ((c & META) && ((basec >= '0' && basec <= '9') || basec == '-'))
	{
		f = TRUE;							/* there is a # arg */
		n = 0;								/* start with a zero default */
		mflag = 1;							/* current minus flag */
		c = basec;							/* strip the META */
		while ((c >= '0' && c <= '9') || c == '-')
		{
			if (c == '-')
			{
				/* already hit a minus or digit? */
				if ((mflag == -1) || (n != 0))
					break;
				mflag = -1;
			}
			else
				n = n * 10 + (c - '0');

			if ((n == 0) && (mflag == -1))	/* lonely - */
				MlWrite("Arg: -");
			else
			{
				strfmt(buf, "Arg: %d", n * mflag);
				MlWrite(buf);
			}

			c = GetCmd();	/* get the next key */
		}
		n = n * mflag;		/* figure in the sign */
	}

	/* and execute the command */
	execute(c, f, n);
	goto loop;
}

/**
 * This is the general command execution routine. It handles the fake binding
 * of all the keys to "self-insert". It also clears out the "thisflag" word,
 * and arranges to move it to the "lastflag", so that the next command can
 * look at it. Return the status of command.
 */
static int near execute(c, f, n)
  int		c;
  int		f, n;
{
  int		status;
  int		(*execfunc)();	/* ptr to function to execute */
  boolean	incmode;

	/* if the keystroke is a bound function, do it */
	execfunc = GetBind(c);
	if (execfunc != NULLFPTR)
	{
		thisflag = 0;
		status = (*execfunc)(f, n);
		lastflag = thisflag;
		return(status);
	}

	if (c >= 0x20 && c <= 0xff)				/* self inserting */
	{
		if (n <= 0) 						/* fenceposts */
		{
			lastflag = 0;
			return(n < 0 ? FALSE : TRUE);
		}
		thisflag = 0;						/* for the future */

		/* if we are in overwrite mode, not at eol, and next char is not a Tab
		 * or we are at a Tab stop, delete a char forward
		 */
		if	(
					curwp->w_bufp->b_mode & MDOVER
				&&
					curwp->w_doto < curwp->w_dotp->l_used
				&&
					(
							lgetc(curwp->w_dotp, curwp->w_doto) != '\t'
						||
							curwp->w_doto % tabsize == tabsize - 1
					)
			)
			LDelete(1L, FALSE);

		/* do the appropriate insertion */
		incmode = curbp->b_mode & MDCMODE;
		if ((c == '}' || c == ')' || c == ']') && incmode)
			status = InsBrace(n, c);
		else if (c == '#' && incmode)
			status = InsPound();
		else
			status = LInsert(n, c);

		/* check for CMODE fence matching */
		if ((c == '}' || c == ')' || c == ']') && incmode)
			FMatch((char)c);

		lastflag = thisflag;
		return(status);
	}
	MlWrite("Key not bound");		/* complain */
	PcPutC(BELL);
	lastflag = 0;					/* fake last flags */
	return(FALSE);
}

/**
 * Fancy Quit command. If any buffer has changed, write that buffer out and
 * exit CodEd, otherwise simply exit.
 */
int QuickExit(f, n)
  int		f, n;
{
  BUFFER	*bp; 					/* scanning pointer to buffers */
  BUFFER	*oldcb;					/* original current buffer */
  char		buf[NSTRING];

	oldcb = curbp;					/* save in case we fail */

	bp = bheadp;
	while (bp != NULLBPTR)
	{
		/* changed && real */
		if ((bp->b_flag & BFCHG) && !(bp->b_flag & BFINVS))
		{
			curbp = bp; 			/* make that buffer current */
			strfmt(buf, "Saving %s", bp->b_fname);
			MlWrite(buf);
			MlWrite("\n");
			if (!FileSave(TRUE, 1))
			{
				curbp = oldcb;		/* restore curbp */
				sgarbf = TRUE;		/* the screen is garbage, so clean it up */
				return(FALSE);
			}
		}
		bp = bp->b_fp;				/* on to the next buffer */
	}
	Quit(f, n); 					/* conditionally Quit */
	return(TRUE);
}

/**
 * Quit command. If an argument, always Quit. Otherwise confirm if a buffer
 * has been changed and not written out.
 */
int Quit(f, n)
  int	f, n;
{
	/*
	 * Quit if the argument forces it OR all buffers clean OR user says it's OK
	 *
	 *	Note that we're counting on the compiler to force the evaluation of
	 *	anycb() BEFORE MlYesNo()
	 */
	if	(
				f != FALSE
			||
				anycb() == FALSE
			||
				MlYesNo("Modified buffers exist. Leave anyway") == TRUE
		)
	{
		VtTidy();
		if (f)
			exit(n);
		else
			exit(SUCCESS);
	}
	MlWrite("");
	return(FALSE);
}

/**
 * Look through the list of buffers. Return TRUE if there are any changed
 * buffers. Buffers that hold magic internal stuff are not considered; who
 * cares if the list of buffer names is hacked. Return FALSE if no buffers
 * have been changed.
 */
static int near anycb()
{
  BUFFER	*bp;

	bp = bheadp;
	while (bp != NULLBPTR)
	{
		if (!(bp->b_flag & BFINVS) && (bp->b_flag & BFCHG))
			return(TRUE);
		bp = bp->b_fp;
	}
	return(FALSE);
}

/**
 * Begin a keyboard macro. Error if not at the top level in keyboard
 * processing. Set up variables and return.
 */
int BegKbdMac(f, n)
  int	f, n;
{
	if (kbdmode != STOP)
	{
		MlWrite("Macro already active");
		return(FALSE);
	}
	MlWrite("Start macro");
	kbdptr = &kbdm[0];
	kbdend = kbdptr;
	kbdmode = RECORD;
	return(TRUE);
}

/**
 * End keyboard macro. Check for the same limit conditions as the above
 * routine. Set up the variables and return to the caller.
 */
int EndKbdMac(f, n)
  int	f, n;
{
	if (kbdmode == STOP)
	{
		MlWrite("Macro not active");
		return(FALSE);
	}
	if (kbdmode == RECORD)
	{
		MlWrite("End macro");
		kbdmode = STOP;
	}
	return(TRUE);
}

/**
 * Execute keyboard macro. The command argument is the number of times to loop.
 * Quit as soon as a command gets an error. Return TRUE if all ok, else FALSE.
 */
int ExecKbdMac(f, n)
  int	f, n;
{
	if (kbdmode != STOP)
	{
		MlWrite("Macro already active");
		return(FALSE);
	}
	if (n <= 0)
		return(TRUE);
	kbdrep = n; 			/* remember how many times to execute */
	kbdmode = PLAY; 		/* start us in play mode */
	kbdptr = &kbdm[0];		/* at the beginning */
	return(TRUE);
}

/**
 * Abort. Beep the beeper. Kill off any keyboard macro, etc., that is in
 * progress (if the flag is TRUE). Sometimes called as a routine, to do
 * general aborting of stuff.
 */
int CmdAbort(f, n)
  int	f, n;
{
	if (f == TRUE)
		kbdmode = STOP;
	MlWrite("Aborted");
	PcPutC(BELL);
	return(ABORT);
}

/**
 * Tell the user that this command is illegal while in VIEW (read-only) mode
 */
int ReadOnly()
{
	MlWrite("Key illegal in VIEW mode");
	PcPutC(BELL);
	return(FALSE);
}

int NullProc()	/* user function that does NOTHING */
{
	return(TRUE);
}

int Meta()		/* dummy function for binding to Meta prefix */
{
	return(TRUE);
}

int Cex()		/* dummy function for binding to control-x prefix */
{
	return(TRUE);
}

static void near putusage()
{
	vdsps("\r\nusage:\tCE [@sf] [options] [files] ... [options] [files]\r\n\n");
#ifdef EKBD
	vdsps("parms:\t@sf\talternate startup file (default:  CE.SF)\r\n\n");
#endif
#ifdef OKBD
	vdsps("parms:\t@sf\talternate startup file (default:  CEO.SF)\r\n\n");
#endif
	vdsps("\t-gn\tgoto line #n\r\n");
	vdsps("\t-sabc\tinitial search string is \"abc\"\r\n\n");
	vdsps("\t-e\topen following file(s) in edit mode - default\r\n");
	vdsps("\t-i\topen following file(s) in insert mode - default\r\n");
	vdsps("\t-m\topen following file(s) in magic search mode\r\n");
	vdsps("\t-n\topen following file(s) in normal search mode - default\r\n");
	vdsps("\t-o\topen following file(s) in overwrite mode\r\n");
	vdsps("\t-v\topen following file(s) in view mode\r\n\n");
	vdsps("\t-h\tprint this message and exit\r\n\n");
	vdsps("notes:\t[-g -s]\t\t\tuse one only\r\n");
	vdsps("\t[-e -i -m -n -o -v]\tuse as required\r\n");

	vdsps("\r\n\nPress any key to continue");
	/* wait until a key is pressed, and then consume it */
	PcGetC();
}
