/*
 *	R A N D O M . C
 *
 *	purpose:
 *		This file contains the command processing functions for a number of
 *		random commands. There is no functional grouping here, for sure.
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

#define nexttab(a)		(a&~(tabsize-1))+tabsize

/*
 * local function prototypes
 */

static int near		adjustmode(int, int);

/**
 * Display the current position of the cursor (in origin 1 X-Y coordinates)
 * the character that is under the cursor (in hex), and the fraction of the
 * text that is before the cursor. The displayed column is not the current
 * column, but the column that would be used on an infinite width display.
 */
int ShowCPos(f, n)
  int	f, n;
{
  LINE	*lp; 			/* current line */
  long	numchars;		/* # of chars in file */
  int	numlines;		/* # of lines in file */
  long	predchars;		/* # chars preceding point */
  int	predlines;		/* # lines preceding point */
  int	col;
  int	savepos;		/* temp save for current offset */
  int	ecol;			/* column pos/end of current line */
  char	buf[NSTRING];

	/* starting at the beginning of the buffer */
	lp = lforw(curbp->b_linep);

	/* start counting chars and lines (let's make things one-based) */
	numchars = 1L;
	numlines = 1;
	while (lp != curbp->b_linep)
	{
		/* if we are on the current line, record it */
		if (lp == curwp->w_dotp)
		{
			predlines = numlines;
			predchars = numchars + (long)curwp->w_doto;
		}
		/* on to the next line */
		++numlines;
		numchars += (long)(llength(lp) + 1);
		lp = lforw(lp);
	}

	/* if at end of file, record it */
	if (curwp->w_dotp == curbp->b_linep)
	{
		predlines = numlines;
		predchars = numchars;
	}

	/* get real column and end-of-line column */
	col = GetCurCol(FALSE);
	savepos = curwp->w_doto;
	curwp->w_doto = llength(curwp->w_dotp);
	ecol = GetCurCol(FALSE);
	curwp->w_doto = savepos;

	/* summarize and report the info */
	strfmt(buf, "line: %d/%d   col: %d/%d   char: %ld/%ld",
			 predlines, numlines, col + 1, ecol + 1, predchars, numchars);
	MlWrite(buf);
	return(TRUE);
}

/**
 * Get the current line number
 */
int GetCurLine()
{
  LINE	*lp; 			/* current line */
  int	numlines;		/* # of lines before point */

	/* starting at the beginning of the buffer */
	lp = lforw(curbp->b_linep);

	/* start counting lines */
	numlines = 0;
	while (lp != curbp->b_linep)
	{
		/* if we are on the current line, record it */
		if (lp == curwp->w_dotp)
			break;
		++numlines;
		lp = lforw(lp);
	}

	/* and return the resulting count */
	return(numlines + 1);
}

/**
 * Return current column (stop at first non-blank given TRUE argument)
 */
int GetCurCol(bflg)
  int	bflg;
{
  int	c;
  int	i;
  int	col;

	col = 0;
	for (i = 0; i < curwp->w_doto; ++i)
	{
		c = lgetc(curwp->w_dotp, i);
		if (c != ' ' && c != '\t' && bflg)
			break;
		if (c == '\t')
			col += -(col % tabsize) + (tabsize - 1);
		else if (c < 0x20 || c == 0x7F)
			++col;
		++col;
	}
	return(col);
}

/**
 * Set current column
 */
int SetCurCol(pos)
  int	pos;			/* position to set cursor */
{
  int	c;				/* character being scanned */
  int	i;				/* index into current line */
  int	col;			/* current cursor column */
  int	llen;			/* length of line in bytes */

	col = 0;
	llen = llength(curwp->w_dotp);

	/* scan the line until we are at or past the target column */
	for (i = 0; i < llen; ++i)
	{
		/* upon reaching the target, drop out */
		if (col >= pos)
			break;

		/* advance one character */
		c = lgetc(curwp->w_dotp, i);
		if (c == '\t')
			col += -(col % tabsize) + (tabsize - 1);
		else if (c < 0x20 || c == 0x7f)
			++col;
		++col;
	}

	/* set us at the new position */
	curwp->w_doto = i;

	/* and tell whether we made it */
	return(col >= pos);
}

/**
 * Twiddle the two characters on either side of dot. If dot is at the end of
 * the line Twiddle the two characters before it. Return with an error if dot
 * is at the beginning of line; it seems to be a bit pointless to make this
 * work. This fixes up a very common typo with a single stroke. This always
 * works within a line, so "WFEDIT" is good enough.
 */
int Twiddle(f, n)
  int		f, n;
{
  LINE		*dotp;
  short		doto;
  int		cl;
  int		cr;

	if (curbp->b_mode & MDVIEW) 	/* don't allow this command if */
		return(ReadOnly());			/* we are in read only mode */
	dotp = curwp->w_dotp;
	doto = curwp->w_doto;
	if (doto == llength(dotp) && --doto < 0)
		return(FALSE);
	cr = lgetc(dotp, doto);
	if (--doto < 0)
		return(FALSE);
	cl = lgetc(dotp, doto);
	lputc(dotp, doto + 0, (char)cr);
	lputc(dotp, doto + 1, (char)cl);
	LChange(WFEDIT);
	return(TRUE);
}

/**
 * Quote the next character, and insert it into the buffer. All the characters
 * are taken literally, with the exception of the NewLine, which always has
 * its line splitting meaning. The character is always read, even if it is
 * inserted 0 times, for regularity.
 */
int Quote(f, n)
  int	f;
  int	n;
{
  int	s;
  int	c;

	if (curbp->b_mode & MDVIEW) 	/* don't allow this command if */
		return(ReadOnly());			/* we are in read only mode */
	c = ExtChToCh(TermGetCh());
	if (n < 0)
		return(FALSE);
	if (n == 0)
		return(TRUE);
	if (c == '\n')
	{
		do
		{
			s = LNewline();
		} while (s == TRUE && --n);
		return(s);
	}
	return(LInsert(n, c));
}

/**
 * Set Tab size if given non-default argument (n > 1).  Otherwise, insert a
 * Tab into file.  This has to be done in this slightly funny way because the
 * Tab (in ASCII) has been turned into ^I (<Tab>) in 10 bit code already.
 */
int Tab(f, n)
  int	f;
  int	n;
{
	if (n < 1)
		return(FALSE);
	if (n != 1)
	{
		tabsize = n;
		UpWind();
		return(TRUE);
	}

	/*
	 * use a hard Tab if stabsize is 0, else use spaces (num spaces = stabsize)
	 */
	if (!stabsize)
		return(LInsert(1, '\t'));
	return(LInsert(stabsize - (GetCurCol(FALSE) % stabsize), ' '));
}

/**
 * Change tabs to spaces
 */
int Detab(f, n)
  int	f;
  int	n;
{
  int	inc;		/* increment to next line [sgn(n)] */

	if (curbp->b_mode & MDVIEW)	/* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */

	if (f == FALSE)
		n = RegionLines();

	/* loop thru detabbing n lines */
	inc = ((n > 0) ? 1 : -1);
	while (n)
	{
		curwp->w_doto = 0;		/* start at the beginning */

		/* Detab the entire current line */
		while (curwp->w_doto < llength(curwp->w_dotp))
		{
			/* if we have a Tab */
			if (lgetc(curwp->w_dotp, curwp->w_doto) == '\t')
			{
				LDelete(1L, FALSE);
				InsSpace(TRUE, tabsize - (curwp->w_doto % tabsize));
			}
			ForwChar(FALSE, 1);
		}

		/* advance/or back to the next line */
		ForwLine(TRUE, inc);
		n -= inc;
	}
	curwp->w_doto = 0;			/* to the beginning of the line */
	thisflag &= ~CFLMOV;		/* flag that this resets the goal column */
	LChange(WFEDIT);			/* yes, we have made at least an edit */
	return(TRUE);
}

/**
 * Change spaces to tabs where possible, but don't Entab within quoted strings.
 */
int Entab(f, n)
  int	f, n;
{
  int	inc;		/* increment to next line [sgn(n)] */
  int	fspace; 	/* pointer to first space if in a run */
  int	ccol;		/* current cursor column */
  char	cchar;		/* current character */
  char	pchar;		/* previous character */
  int	instr;		/* flag when in a quoted string */

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */

	if (f == FALSE)
		n = RegionLines();

	/* loop thru entabbing n lines */
	inc = ((n > 0) ? 1 : -1);
	while (n)
	{
		curwp->w_doto = 0;		/* start at the beginning */

		/* first, Detab the entire current line */
		while (curwp->w_doto < llength(curwp->w_dotp))
		{
			/* if we have a Tab */
			if (lgetc(curwp->w_dotp, curwp->w_doto) == '\t')
			{
				LDelete(1L, FALSE);
				InsSpace(TRUE, tabsize - (curwp->w_doto % tabsize));
			}
			ForwChar(FALSE, 1);
		}

		curwp->w_doto = 0;		/* start at the beginning */

		/* now, Entab the resulting spaced out line */
		fspace = -1;
		ccol = 0;
		cchar = '\0';			/* so pchar is defined on first pass */
		instr = FALSE;
		while (curwp->w_doto < llength(curwp->w_dotp))
		{
			/* see if it is time to compress */
			if ((fspace >= 0) && nexttab(fspace) <= ccol)
			{
				if (ccol - fspace < 2)
					fspace = -1;
				else
				{
					BackChar(TRUE, ccol - fspace);
					LDelete((long)(ccol - fspace), FALSE);
					LInsert(1, '\t');	    
					fspace = -1;
				}
			}

			/* get the current character */
			pchar = cchar;
			cchar = lgetc(curwp->w_dotp, curwp->w_doto);

			switch (cchar)
			{
				case '\t':		/* a Tab, so count em up */
					ccol = nexttab(ccol);
					break;
				case ' ':		/* a space, so compress? */
					if (!instr)
					{
						if (fspace == -1)
							fspace = ccol;
						ccol++;
					}
					else
					{
						ccol++;
						fspace = -1;
					}
					break;
				case '"':		/* don't Entab quoted strings */
					if (!instr)
					{
						if (pchar != '\\')	/* if not escaped quote */
							instr = TRUE;	/* opening quote */
					}
					else
					{
						if (pchar != '\\')	/* if not escaped quote */
							instr = FALSE;	/* then closing quote */
					}
					ccol++;
					fspace = -1;
					break;
				default:		/* any other char, just count */
					ccol++;
					fspace = -1;
					break;
			}
			ForwChar(FALSE, 1);
		}

		/* advance/backup to the next line */
		ForwLine(TRUE, inc);
		n -= inc;
	}
	curwp->w_doto = 0;			/* to the beginning of the line */
	thisflag &= ~CFLMOV;		/* flag that this resets the goal column */
	LChange(WFEDIT);			/* yes, we have made at least an edit */
	return(TRUE);
}

/**
 * Trim trailing whitespace from the end of the line
 */
int Trim(f, n)
  int	f;
  int	n;
{
  LINE	*lp; 		/* current line pointer */
  int	offset; 	/* original line offset position */
  int	length; 	/* current length */
  int	inc;		/* increment to next line [sgn(n)] */

	if (curbp->b_mode & MDVIEW)	/* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */

	if (f == FALSE)
		n = RegionLines();

	/* loop thru trimming n lines */
	inc = ((n > 0) ? 1 : -1);
	while (n)
	{
		lp = curwp->w_dotp; 	/* find current line text */
		offset = curwp->w_doto; /* save original offset */
		length = lp->l_used;	/* find current length */

		/* Trim the current line */
		while (length > offset)
		{
			if (lgetc(lp, length - 1) != ' ' && lgetc(lp, length - 1) != '\t')
				break;
			length--;
		}
		lp->l_used = length;

		/* advance/or back to the next line */
		ForwLine(TRUE, inc);
		n -= inc;
	}
	LChange(WFEDIT);
	thisflag &= ~CFLMOV;		/* flag that this resets the goal column */
	return(TRUE);
}

/**
 * Open up some blank space. The basic plan is to insert a bunch of newlines,
 * and then back up over them. Everything is done by the subcommand processors.
 * They even handle the looping.
 */
int OpenLine(f, n)
  int	f;
  int	n;
{
  int	i;
  int	s;

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if (n < 0)
		return(FALSE);
	if (n == 0)
		return(TRUE);
	i = n;						/* insert newlines */
	do
	{
		s = LNewline();
	} while (s == TRUE && --i);
	if (s == TRUE)				/* then back up over the top of them all */
		s = BackChar(f, n);
	return(s);
}

/**
 * Insert a NewLine. If we are in Cmode, do automatic indentation as specified.
 */
int NewLine(f, n)
  int	f;
  int	n;
{
	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if (n < 0)
		return(FALSE);

	/* if we are in C mode and this is a default <NL> */
	if (n == 1 && (curbp->b_mode & MDCMODE) && curwp->w_dotp != curbp->b_linep)
		return(InsChar());

	/* insert some lines */
	while (n--)
	{
		if (!LNewline())
			return(FALSE);
	}
	return(TRUE);
}

/**
 * Insert a NewLine and indentation for Cmode
 */
int InsChar()
{
  char		*cptr;			/* string pointer into text to copy */
  int		i;				/* index into line to copy Indent from */
  int		llen;			/* length of line to copy from */
  boolean	ofence; 		/* was there an opening fence char at the eol? */
  char		ofc;			/* opening fence character */
  LINE		*lp;			/* pointer to current line */
  int		offset;
  char		ichar[NSTRING]; /* buffer to hold Indent of last line */

	/* Trim the whitespace before the point */
	lp = curwp->w_dotp;
	offset = curwp->w_doto;
	while	(
					offset > 0
				&&
							lgetc(lp, offset - 1) == ' '
						||
							lgetc(lp, offset - 1) == '\t'
			)
	{
		BackDel(FALSE, 1);
	    	offset--;
	}

	/* check for a brace */
	ofc = lgetc(lp, offset - 1);
	ofence = (offset > 0 && (ofc == '{' || ofc == '(' || ofc == '['));

	/* put in the NewLine */
	if (!LNewline())
		return(FALSE);

	/* hunt for the last non-blank line to get indentation from */
	lp = lp->l_bp;
	while (lp->l_used == 0 && lp != curbp->b_linep)
		lp = lp->l_bp;

	/* grab a pointer to text to copy indentation from */
	cptr = &(lp->l_text[0]);
	llen = lp->l_used;

	/* save the Indent of the previous line */
	i = 0;
	while ((i < llen) && (cptr[i] == ' ' || cptr[i] == '\t') && (i < NSTRING-1))
	{
		ichar[i] = cptr[i];
		++i;
	}
	ichar[i] = 0;					/* terminate it */

	/* and the saved indentation */
	LInstr(ichar);

	/* and one more Tab for an opening symbol */
	if (ofence)
		Tab(FALSE, 1);

	return(TRUE);
}

/**
 * Insert a brace into the text here (we're in Cmode)
 */
int InsBrace(n, c)
  int	n;			/* repeat count */
  int	c;			/* brace to insert */
{
  int	ch; 		/* last character before input */
  int	oc; 		/* caractere oppose a c */
  int	i, count;
  int	target; 	/* column brace should go after */
  LINE	*odotp;
  short	odoto;

	/* if we aren't at the beginning of the line, scan it to see if
	 *	 all preceding space is white space
	 */
	if (curwp->w_doto != 0)
	{
		for (i = curwp->w_doto - 1; i >= 0; --i)
		{
			ch = lgetc(curwp->w_dotp, i);
			if (ch != ' ' && ch != '\t')
				return(LInsert(n, c));
		}
	}

	/* chercher le caractere oppose correspondant */
	switch (c)
	{
		case '}':
			oc = '{';
			break;
		case ']':
			oc = '[';
			break;
		case ')':
			oc = '(';
			break;
		default:
			return(FALSE);
	}
	    
	odotp = curwp->w_dotp;
	odoto = curwp->w_doto;
	    
	count = 1;

	while (count > 0)
	{
		BackChar(FALSE, 1);

		if (curwp->w_doto == llength(curwp->w_dotp))
			ch = '\n';
		else
			ch = lgetc(curwp->w_dotp, curwp->w_doto);

		if (ch == c)
			++count;
		if (ch == oc)
			--count;

		/* at beginning of file? */
		if (Boundry(curwp->w_dotp, curwp->w_doto, REVERSE))
			break;
	}
	    
	if (count != 0) 				/* no match */
	{
		curwp->w_dotp = odotp;
		curwp->w_doto = odoto;
		return(LInsert(n, c));
	}
	    
	curwp->w_doto = 0;				/* debut de ligne */
	/* aller au debut de la ligne apres la tabulation */
	while ((ch = lgetc(curwp->w_dotp, curwp->w_doto)) == ' ' || ch == '\t')
		ForwChar(FALSE, 1);

	/* delete back first */
	target = GetCurCol(FALSE);		/* c'est l'Indent que l'on doit avoir */
	curwp->w_dotp = odotp;
	curwp->w_doto = odoto;

	while (target != GetCurCol(FALSE))
	{
		if (target < GetCurCol(FALSE))/* on doit detruire des caracteres */
			while (GetCurCol(FALSE) > target)
				BackDel(FALSE, 1);
		else
		{							/* on doit en inserer */
			while (target - GetCurCol(FALSE) >= tabsize)
				LInsert(1, '\t');
			LInsert(target - GetCurCol(FALSE), ' ');
		}
	}

	/* and insert the required brace(s) */
	return(LInsert(n, c));
}

/**
 * Insert a # into the text here (we're in Cmode)
 */
int InsPound()
{
  int	ch; 		/* last character before input */
  int	i;

	/* if we are at the beginning of the line, just do it */
	if (curwp->w_doto == 0)
		goto insp_ret;

	/* scan to see if all space before this is white space */
	for (i = curwp->w_doto - 1; i >= 0; --i)
	{
		ch = lgetc(curwp->w_dotp, i);
		if (ch != ' ' && ch != '\t')
			goto insp_ret;
	}

	/* delete back first */
	while (GetCurCol(FALSE) >= 1)
		BackDel(FALSE, 1);

	/* and insert the required pound */
insp_ret:
	return(LInsert(1, '#'));
}

/**
 * Delete blank lines around dot. What this command does depends if dot is
 * sitting on a blank line. If dot is sitting on a blank line, this command
 * deletes all the blank lines above and below the current line. If it is
 * sitting on a non-blank line then it deletes all of the blank lines after
 * the line. Any argument is ignored.
 */
int Deblank(f, n)
  int	f, n;
{
  LINE	*lp1;
  LINE	*lp2;
  long	nld;

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	lp1 = curwp->w_dotp;
	while (llength(lp1) == 0 && (lp2 = lback(lp1)) != curbp->b_linep)
		lp1 = lp2;
	lp2 = lp1;
	nld = 0L;
	while ((lp2 = lforw(lp2)) != curbp->b_linep && llength(lp2) == 0)
		++nld;
	if (nld == 0L)
		return(TRUE);
	curwp->w_dotp = lforw(lp1);
	curwp->w_doto = 0;
	return(LDelete(nld, FALSE));
}

/**
 * Insert a NewLine, then enough tabs and spaces to duplicate the indentation
 * of the previous line. Quite simple. Figure out the indentation of the current
 * line. Insert a NewLine by calling the standard routine. Insert the Indent
 * by inserting the right number of tabs and spaces. Return TRUE if all OK.
 * Return FALSE if one of the subcomands failed.
 */
int Indent(f, n)
  int	f, n;
{
  int	nicol;
  int	c;
  int	i;

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if (n < 0)
		return(FALSE);
	while (n--)
	{
		nicol = 0;
		for (i = 0; i < llength(curwp->w_dotp); ++i)
		{
			c = lgetc(curwp->w_dotp, i);
			if (c != ' ' && c != '\t')
				break;
			if (c == '\t')
				nicol += -(nicol % tabsize) + (tabsize - 1);
			++nicol;
		}
		if	(
					LNewline() == FALSE
				||
					((i = nicol / tabsize) != 0 && LInsert(i, '\t') == FALSE)
				||
					((i = nicol % tabsize) != 0 && LInsert(i,  ' ') == FALSE)
			)
			return(FALSE);
	}
	return(TRUE);
}

/**
 * Delete forward. This is real easy, because the basic delete routine does
 * all of the work. Watches for negative arguments, and does the right thing.
 * If any argument is present, it kills rather than deletes, to prevent loss
 * of text if typed with a big argument.
 */
int ForwDel(f, n)
  int	f, n;
{
	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if (n < 0)
		return(BackDel(f, -n));
	if (f != FALSE) 			/* really a kill */
	{
		if ((lastflag & CFKILL) == 0)
			KDelete();
		thisflag |= CFKILL;
	}
	return(LDelete((long)n, f));
}

/**
 * Delete backwards. This is quite easy too, because it's all done with other
 * functions. Just move the cursor back, and delete forwards. Like delete
 * forward, this actually does a kill if presented with an argument.
 */
int BackDel(f, n)
  int	f, n;
{
	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if (n < 0)
		return(ForwDel(f, -n));
	if (f != FALSE) 			/* really a kill */
	{
		if ((lastflag & CFKILL) == 0)
			KDelete();
		thisflag |= CFKILL;
	}
	if (BackChar(f, n))
		return(LDelete((long)n, f));
	return(FALSE);
}

/**
 * Kill text. If called without an argument, it kills from dot to the end of
 * the line, unless it is at the end of the line, when it kills the NewLine.
 * If called with an argument of 0, it kills from the start of the line to dot.
 * If called with a positive argument, it kills from dot forward over that
 * number of newlines. If called with a negative argument it kills backwards
 * that number of newlines.
 */
int KillText(f, n)
  int	f;
  int	n;
{
  LINE	*nextp;
  long	chunk;

	if (curbp->b_mode & MDVIEW)		/* don't allow this command if */
		return(ReadOnly());			/* we are in read only mode */
	if ((lastflag & CFKILL) == 0)	/* clear kill buffer if */
		KDelete();					/*	last wasn't a kill */
	thisflag |= CFKILL;
	if (f == FALSE)
	{
		chunk = (long)(llength(curwp->w_dotp) - curwp->w_doto);
		if (chunk == 0L)
			chunk = 1L;
	}
	else if (n == 0)
	{
		chunk = (long)curwp->w_doto;
		curwp->w_doto = 0;
	}
	else if (n > 0)
	{
		chunk = (long)(llength(curwp->w_dotp) - curwp->w_doto + 1);
		nextp = lforw(curwp->w_dotp);
		while (--n)
		{
			if (nextp == curbp->b_linep)
				break;
			chunk += (long)(llength(nextp) + 1);
			nextp = lforw(nextp);
		}
	}
	else
	{
		MlWrite("Negative kill");
		return(FALSE);
	}
	return(LDelete(chunk, TRUE));
}

/**
 * Prompt and set an editor mode
 */
int AddMode(f, n)
  int	f, n;
{
	return(adjustmode(TRUE, FALSE));
}

/**
 * Prompt and delete an editor mode
 */
int DelMode(f, n)
  int	f, n;
{
	return(adjustmode(FALSE, FALSE));
}

/**
 * Prompt and set a global editor mode
 */
int AddGMode(f, n)
  int	f, n;
{
	return(adjustmode(TRUE, TRUE));
}

/**
 * Prompt and delete a global editor mode
 */
int DelGMode(f, n)
  int	f, n;
{
	return(adjustmode(FALSE, TRUE));
}

/**
 * Change the editor mode status
 */
static int near adjustmode(kind, global)
  int	kind;		/* true = set,	   false = delete */
  int	global; 	/* true = global,  false = current buffer */
{
  int	i;
  int	status; 	/* error return on input */
  int	uflag;		/* was modename uppercase? */
  char	prompt[50]; /* string to prompt user with */
  char	cbuf[NPAT]; /* buffer to receive mode name into */

	prompt[0] = 0;			/* start with an empty string */

	/* build the proper prompt string */
	if (global)
		stracat(prompt,"Global ");

	stracat(prompt,"Mode to ");

	if (kind == TRUE)
		stracat(prompt, "add: ");
	else
		stracat(prompt, "delete: ");

	/* prompt the user and get an answer */
	status = MlReply(prompt, cbuf, NPAT - 1);
	if (status != TRUE)
		return(FALSE);

	/* test case: upper means foreground - lower means background */
	uflag = (*cbuf >= 'A' && *cbuf <= 'Z');

	/* make it uppercase for stracmp() */
	strupper(cbuf);

	/* test it first against the colors we know */
	for (i = 0; i < NCOLORS; i++)
	{
		if (stracmp(cbuf, cname[i]) == 0)
		{
			/* finding the match, we set the color */
			if (uflag)
			{
				if (global)
					gfcolor = i;
				else
					curwp->w_fcolor = (char)i;
			}
			else
			{
				if (global)
					gbcolor = i;
				else
					curwp->w_bcolor = (char)i;
			}

			curwp->w_flag |= WFCOLR;
			MlErase();
			return(TRUE);
		}
	}

	/* test it against the modes we know */

	for (i = 0; i < NUMMODES; i++)
	{
		if (cbuf[0] == modecode[i])
		{
			/* finding a match, we process it */
			if (kind == TRUE)
				if (global)
					gmode |= (1 << i);
				else
					curbp->b_mode |= (1 << i);
			else
				if (global)
					gmode &= ~(1 << i);
				else
					curbp->b_mode &= ~(1 << i);
			/* display new mode line */
			if (global == 0)
				UpMode();
			MlErase();			/* erase the junk */
			return(TRUE);
		}
	}

	MlWrite("No such mode");
	return(FALSE);
}

/**
 * Clear the message line (mainly for macro usage)
 */
int ClrMsg()
{
	MlForce("");
	return(TRUE);
}

/**
 * Write a string on the message line (mainly for macro usage)
 */
int WriteMsg()
{
  int	status;
  char	buf[NSTRING];

	if ((status = MlReply("Message to write: ", buf, NPAT - 1)) != TRUE)
		return(status);

	/* write the message out */
	MlForce(buf);
	return(TRUE);
}

/**
 * Move the cursor to a matching fence
 */
int GetFence(f, n)
  int	f, n;
{
  LINE	*odotp;		/* original line pointer */
  short	odoto;		/* and offset */
  int	sdir;		/* direction of search (1/-1) */
  int	count;		/* current fence level count */
  char	ch; 		/* fence type to match against */
  char 	mfence; 	/* matching fence */
  char 	c;			/* current character in scan */

	/* save the original cursor position */
	odotp = curwp->w_dotp;
	odoto = curwp->w_doto;

	/* get the current character */
	if (odoto == llength(odotp))
		ch = '\n';
	else
		ch = lgetc(odotp, odoto);

	/* setup proper matching fence */
	switch (ch)
	{
		case '(':
			mfence = ')';
			sdir = FORWARD;
			break;
		case '{':
			mfence = '}';
			sdir = FORWARD;
			break;
		case '[':
			mfence = ']';
			sdir = FORWARD;
			break;
		case ')':
			mfence = '(';
			sdir = REVERSE;
			break;
		case '}':
			mfence = '{';
			sdir = REVERSE;
			break;
		case ']':
			mfence = '[';
			sdir = REVERSE;
			break;
		default:
			MlWrite("Current character not a fence");
			PcPutC(BELL);
			return(FALSE);
	}

	/* set up for scan */
	count = 1;
	if (sdir == REVERSE)
		BackChar(FALSE, 1);
	else
		ForwChar(FALSE, 1);

	/* scan until we find it, or reach the end of file */
	while (count > 0)
	{
		if (curwp->w_doto == llength(curwp->w_dotp))
			c = '\n';
		else
			c = lgetc(curwp->w_dotp, curwp->w_doto);

		if (c == ch)
			++count;
		if (c == mfence)
			--count;

		if (sdir == FORWARD)
			ForwChar(FALSE, 1);
		else
			BackChar(FALSE, 1);

		/* at beginning (sdir == REVERSE) or end (sdir == FORWARD) of file? */
		if (Boundry(curwp->w_dotp, curwp->w_doto, sdir))
			break;
	}

	/* if count is zero, we have a match, move the sucker */
	if (count == 0)
	{
		if (sdir == FORWARD)
			BackChar(FALSE, 1);
		else
			ForwChar(FALSE, 1);
		curwp->w_flag |= WFMOVE;
		return(TRUE);
	}

	/* restore the current position */
	curwp->w_dotp = odotp;
	curwp->w_doto = odoto;
	MlWrite("Matching fence not found");
	PcPutC(BELL);
	return(FALSE);
}

/**
 * Closing fences are matched against their partners, and if on screen the
 * cursor briefly lights there.
 */
int FMatch(ch)
  char	ch; 		/* fence type to match against */
{
  LINE	*odotp;		/* original line pointer */
  short	odoto; 		/* and offset */
  LINE	*toplp;		/* top line in current window */
  int	count;		/* current fence level count */
  char 	opench; 	/* open fence */
  char 	c;			/* current character in scan */
  int	i;

	/* first get the display Update out there */
	Update(FALSE);

	/* save the original cursor position */
	odotp = curwp->w_dotp;
	odoto = curwp->w_doto;

	/* setup proper open fence for passed close fence */
	if (ch == ')')
		opench = '(';
	else if (ch == '}')
		opench = '{';
	else
		opench = '[';

	/* find the top line and set up for scan */
	toplp = curwp->w_linep->l_bp;
	count = 1;
	BackChar(FALSE, 2);

	/* scan back until we find it, or reach past the top of the window */
	while (count > 0 && curwp->w_dotp != toplp)
	{
		if (curwp->w_doto == llength(curwp->w_dotp))
			c = '\n';
		else
			c = lgetc(curwp->w_dotp, curwp->w_doto);
		if (c == ch)
			++count;
		if (c == opench)
			--count;
		BackChar(FALSE, 1);
		if (curwp->w_dotp == curwp->w_bufp->b_linep->l_fp && curwp->w_doto == 0)
			break;
	}

	/* if count is zero, we have a match, display the sucker
	 * NOTE! there is a real machine dependant timing problem here we have
	 *		 yet to solve.
	 */
	if (count == 0)
	{
		ForwChar(FALSE, 1);
		for (i = 0; i < term.t_pause; i++)
			Update(FALSE);
	}

	/* restore the current position */
	curwp->w_dotp = odotp;
	curwp->w_doto = odoto;
	return(TRUE);
}

/**
 * Ask for and insert a string into the current buffer at the current point
 */
int InsString(f, n)
  int	f;
  int	n;
{
  int	status; 			/* status return code */
  char	tstring[NPAT + 1];	/* string to add */

	/* ask for string to insert */
	status = MlReplyT("String to insert<META>: ", tstring, NPAT, metac);
	if (status != TRUE)
		return(status);

	if (f == FALSE)
		n = 1;

	if (n < 0)
		n = - n;

	/* insert it */
	while (n-- && ((status = LInstr(tstring)) == TRUE))
		;
	return(status);
}

/**
 * Ask for and overwite a string into the current buffer at the current point
 */
int OvString(f, n)
  int	f;
  int	n;
{
  int	status; 			/* status return code */
  char	tstring[NPAT + 1];	/* string to add */

	/* ask for string to insert */
	status = MlReplyT("String to overwrite<META>: ", tstring, NPAT, metac);
	if (status != TRUE)
		return(status);

	if (f == FALSE)
		n = 1;

	if (n < 0)
		n = - n;

	/* insert it */
	while (n-- && ((status = LOver(tstring)) == TRUE))
		;
	return(status);
}

/**
 * Write an 'Out of memory' message to the message line and feep
 */
void OutOfMem()
{
	MlWrite("Out of memory");
	PcPutC(BELL);
}
