/*
 *	S E A R C H . C
 *
 *	purpose:
 *		The functions in this file implement CodEd's search commands
 *
 *	note:
 *		In replaces():	the code, as it currently exists, does not support
 *						(U)ndoing the last possible substitution - this is due
 *						to the fact that replaces() exits when it reaches the
 *						end of the buffer - it is no longer in the replace mode
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <stdlib.h>
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

#define CHCASE(c)	((c) ^ 0x20)	/* toggle the case of a letter */
#define MAXCHAR		256				/* number of 8-bit ascii chars */

/*
 * local function prototypes
 */

#if MAGIC
static int near 	delins(int, char *, int);
#else
static int near 	delins(int, char *);
#endif
static int near 	fbound(int, int);
static int near 	isletter(char);
static int near 	nextch(int);
static int near 	readpattern(char *, char *, int);
static int near 	replaces(int, int, int);
static int near 	savematch(void);
static void near	checkmatch(int);
static void near	putbackch(int);

#if MAGIC
static BITMAP near	clearbits(void);
static int near 	amatch(struct mc *, int);
static int near 	biteq(int, char *);
static int near 	cclmake(char **, struct mc *);
static int near 	mceq(int, struct mc *);
static int near 	mcscanner(struct mc *, int, int);
static int near 	mcstr(void);
static int near 	rmcclear(void);
static int near 	rmcstr(void);
static void near	setbit(int, char *);
#endif

/*
 * local (module global) variable declarations
 */

static ushort	matchlen = 0;			/* length of matched string */
static ushort	mlenold  = 0;			/*	(used by replace functions) */
static LINE 	*matchline = NULLLPTR;	/* line of *start* of match */
static int		matchoff = 0;			/* offset of *start* of match */
static LINE 	*scurline = NULLLPTR;	/* current scanned line */
static int		scuroff = 0;			/* current scanned line offset */
static LINE 	*curline = NULLLPTR;	/* current line */
static int		curoff = 0; 			/* current offset */

static int		patlenadd;
static int		deltaf[MAXCHAR], deltab[MAXCHAR];
static int		lastchfjump, lastchbjump;

/**
 * Search forward.	Get a search string from the user, and search for the
 * string.	If found, reset the dot to be just after the match string, and
 * (perhaps) repaint the display.
 */
int ForwSearch(f, n)
  int	f, n;
{
  int	status = TRUE;

	/* if n is negative, search backwards, else go get search string */
	if (n < 0)
		return(BackSearch(f, -n));

	/* Ask the user for the text of a pattern. If response is TRUE (responses
	 * other than FALSE are possible), search for the pattern for as long as
	 * n is positive (n == 0 will go through once, which is just fine).
	 */
	if ((status = readpattern("Search", &pat[0], TRUE)) == TRUE)
	{
		do
		{
#if MAGIC
			if ((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
				status = mcscanner(&mcpat[0], FORWARD, PTEND);
			else
#endif
				status = Scanner(&pat[0], FORWARD, PTEND);
		} while ((--n > 0) && status);

		/* complain if no match */
		checkmatch(status);
	}
	return(status);
}

/**
 * Search forward for a previously acquired search string. If found, reset the
 * dot to be just after the match string, and (perhaps) repaint the display.
 */
int ForwHunt(f, n)
  int	f, n;
{
  int	status = TRUE;

	if (n < 0)				/* search backwards */
		return(BackHunt(f, -n));

	/* Make sure a pattern exists, or that we didn't switch
	 * into MAGIC mode until after we entered the pattern.
	 */
	if (pat[0] == NULLCHR)
	{
		MlWrite("No pattern set");
		return(FALSE);
	}
#if MAGIC
	if ((curwp->w_bufp->b_mode & MDMAGIC) != 0 && mcpat[0].mc_type == MCNIL)
	{
		if (!mcstr())
			return(FALSE);
	}
#endif

	/* Search for the pattern for as long as n is positive (n == 0 will go
	 * through once, which is just fine).
	 */
	do
	{
#if MAGIC
		if ((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
			status = mcscanner(&mcpat[0], FORWARD, PTEND);
		else
#endif
			status = Scanner(&pat[0], FORWARD, PTEND);
	} while ((--n > 0) && status);

	/* complain if no match */
	checkmatch(status);

	return(status);
}

/**
 * Reverse search.	Get a search string from the user, and search, starting at
 * dot and proceeding toward the front of the buffer. If found dot is left
 * pointing at the first character of the pattern (the last character that was
 * matched).
 */
int BackSearch(f, n)
  int	f, n;
{
  int	status = TRUE;

	/* if n is negative, search forwards, else go get search string */
	if (n < 0)
		return(ForwSearch(f, -n));

	/* Ask the user for the text of a pattern.	If response is TRUE (responses
	 * other than FALSE are possible), search for the pattern for as long as
	 * n is positive (n == 0 will go through once, which is just fine).
	 */
	if ((status = readpattern("Reverse search", &pat[0], TRUE)) == TRUE)
	{
		do
		{
#if MAGIC
			if ((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
				status = mcscanner(&tapcm[0], REVERSE, PTBEG);
			else
#endif
				status = Scanner(&tap[0], REVERSE, PTBEG);
		} while ((--n > 0) && status);

		/* complain if no match */
		checkmatch(status);
	}
	return(status);
}

/**
 * Reverse search for a previously acquired search string, starting at dot and
 * proceeding toward the front of the buffer. If found dot is left pointing at
 * the first character of the pattern (the last character that was matched).
 */
int BackHunt(f, n)
  int	f, n;
{
  int	status = TRUE;

	if (n < 0)
		return(ForwHunt(f, -n));

	/* Make sure a pattern exists, or that we didn't switch
	 * into MAGIC mode until after we entered the pattern.
	 */
	if (tap[0] == NULLCHR)
	{
		MlWrite("No pattern set");
		return(FALSE);
	}
#if MAGIC
	if ((curwp->w_bufp->b_mode & MDMAGIC) != 0 && tapcm[0].mc_type == MCNIL)
	{
		if (!mcstr())
			return(FALSE);
	}
#endif

	/* Go search for it for as long as n is positive (n == 0 will go through
	 * once, which is just fine).
	 */
	do
	{
#if MAGIC
		if ((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
			status = mcscanner(&tapcm[0], REVERSE, PTBEG);
		else
#endif
			status = Scanner(&tap[0], REVERSE, PTBEG);
	} while ((--n > 0) && status);

	/* complain if no match */
	checkmatch(status);

	return(status);
}

/**
 * Let the user know he's SOL
 */
static void near checkmatch(status)
  int		status;
{
	if (status != TRUE)
		MlWrite("Not found");
}

#if MAGIC
/**
 * Search for a Meta-pattern in either direction.  If found, reset the dot to
 * be at the start or just after the match string, and (perhaps) repaint the
 * display.
 */
static int near mcscanner(mcpatrn, direct, beg_or_end)
  MC	*mcpatrn;		/* pointer into pattern */
  int	direct; 		/* which way to go */
  int	beg_or_end; 	/* put point at begin or end of pattern */
{
	/* If we are going in reverse, then the 'end' is actually the beginning
	 * of the pattern.	Toggle it.
	 */
	beg_or_end ^= direct;

	/* Save the old matchlen length, in case it is very different (closure)
	 * from the old length. This is important for query-replace undo command.
	 */
	mlenold = matchlen;

	/* setup local scan pointers to global dot */
	scurline = curwp->w_dotp;
	scuroff = curwp->w_doto;

	/* scan each character until we hit the head link record */
	while (!Boundry(scurline, scuroff, direct))
	{
		/* Save the current position in case we need to restore it on a match,
		 * and initialize matchlen to zero in case we are doing a search for
		 * replacement.
		 */
		matchline = scurline;
		matchoff = scuroff;
		matchlen = 0;

		if (amatch(mcpatrn, direct))
		{
			/* A SUCCESSFULL MATCH!!! Reset the global dot pointers. */
			if (beg_or_end == PTEND)	/* at end of string */
			{
				curwp->w_dotp = scurline;
				curwp->w_doto = scuroff;
			}
			else						/* at beginning of string */
			{
				curwp->w_dotp = matchline;
				curwp->w_doto = matchoff;
			}
			curwp->w_flag |= WFMOVE;	/* flag that we have moved */
			savematch();
			return(TRUE);
		}
		/* advance the cursor */
		nextch(direct);
	}

	return(FALSE);						/* we could not find a match */
}

/**
 * Search for a Meta-pattern in either direction.  Based on the recursive
 * routine amatch() (for "anchored match") in Kernighan & Plauger's "Software
 * Tools".
 */
static int near amatch(mcptr, direct)
  MC	*mcptr; 		/* string to scan for */
  int	direct; 		/* which way to go */
{
  int	c;				/* character at current position */
  int	nchars;

	/* save current location */
	curline = scurline;
	curoff = scuroff;

	/* The beginning-of-line and end-of-line metacharacters do not compare
	 * against characters, they compare against positions. BOL is guaranteed
	 * to be at the start of the pattern for forward searches, and at the end
	 * of the pattern for reverse searches.  The reverse is true for EOL. So,
	 * for a start, we check for them on entry.
	 */
	if (mcptr->mc_type == BOL)
	{
		if (scuroff != 0)
			return(FALSE);
		mcptr++;
	}

	if (mcptr->mc_type == EOL)
	{
		if (scuroff != llength(scurline))
			return(FALSE);
		mcptr++;
	}

	while (mcptr->mc_type != MCNIL)
	{
		c = nextch(direct);

		if (mcptr->mc_type & CLOSURE)
		{
			/* Try to match as many characters as possible against the current
			 * Meta-character.	A NewLine never matches a closure.
			 */
			nchars = 0;
			while (c != '\n' && mceq(c, mcptr))
			{
				c = nextch(direct);
				nchars++;
			}

			/* We are now at the character that made us fail.  Try to match the
			 * rest of the pattern. Shrink the closure by one for each failure.
			 * Since closure matches ZERO or more occurrences of a pattern, a
			 * match may start even if the previous loop matched no characters.
			 */
			mcptr++;

			for (;;)
			{
				putbackch(direct);
				if (amatch(mcptr, direct))
				{
					matchlen += nchars;
					goto success;
				}
				if (nchars-- == 0)
					goto failure;
			}
		}
		else							/* not CLOSURE */
		{
			/* The only way we'd get a BOL metacharacter at this point is at 
			 * the end of the reversed pattern. The only way we'd get an EOL 
			 * metacharacter here is at the end of a regular pattern. So if we 
			 * match one or the other, and are at the appropriate position, we 
			 * are guaranteed success (since the next pattern character has to 
			 * be MCNIL). Before we report success, however, we back up by one 
			 * character, so as to leave the cursor in the correct position.  
			 * For example, a search for ")$" will leave the cursor at the end 
			 * of the line, while a search for ")<NL>" will leave the cursor at 
			 * the beginning of the next line.	This follows the notion that 
			 * the Meta-character '$' (and likewise '^') match positions, not 
			 * characters. 
			 */ 

			if (mcptr->mc_type == BOL)
			{
				if (scuroff == llength(scurline))
				{
					putbackch(direct);
					goto success;
				}
				else
					goto failure;
			}

			if (mcptr->mc_type == EOL)
			{
				if (scuroff == 0)
				{
					putbackch(direct);
					goto success;
				}
				else
					goto failure;
			}

			/* not BOL nor EOL, so go through Meta-character equal function */
			if (!mceq(c, mcptr))
				goto failure;
		}

		/* increment the length counter and advance the pattern pointer */
		matchlen++;
		mcptr++;
	}

	/* A SUCCESSFUL MATCH!!! */

success:

	return(TRUE);

	/* NO LUCK THIS TIME!!! */

failure:

	/* restore original position */
	scurline = curline;
	scuroff = curoff;
	return(FALSE);
}
#endif

/**
 * Search for a pattern in either direction.  If found, reset the dot to be at
 * the start (reverse search) or just after the match string (forward search),
 * and (perhaps) repaint the display.
 */
int Scanner(patrn, direct, beg_or_end)
  uchar *patrn; 	/* string to scan for */
  int	direct; 	/* which way to go */
  int	beg_or_end; /* put point at begin or end of pattern */
{
  int	c;			/* character at current position */
  uchar *patptr;	/* pointer into pattern */
  int	emf;		/* MDEXACT mode flag? */
  int	jump;		/* next offset */

	/* If we are going in reverse, then the 'end' is actually the beginning of
	 * the pattern.  Toggle it.
	 */
	beg_or_end ^= direct;

	/* set up local pointers to global dot */
	curline = curwp->w_dotp;
	curoff = curwp->w_doto;

	/* set up current exact mode flag */
	emf = curwp->w_bufp->b_mode & MDEXACT;

	/* scan each character until we hit the head link record */
	jump = patlenadd;
	while (!fbound(jump, direct))
	{
		/* save the current position in case we match the search string */
		matchline = curline;
		matchoff = curoff;

		/* setup scanning pointers */
		scurline = curline;
		scuroff = curoff;
		patptr = &patrn[0];

		/* scan the pattern for a match */
		while ((c = (uint)*patptr++) != NULLCHR)
		{
			if (!Eq(c, nextch(direct), emf))
			{
				jump = (direct == FORWARD) ? lastchfjump : lastchbjump;
				goto fail;
			}
		}
    
		/* A SUCCESSFULL MATCH!!! reset the global dot pointers */
		if (beg_or_end == PTEND)		/* at end of string */
		{
			curwp->w_dotp = scurline;
			curwp->w_doto = scuroff;
		}
		else							/* at beginning of string */
		{
			curwp->w_dotp = matchline;
			curwp->w_doto = matchoff;
		}

		curwp->w_flag |= WFMOVE;		/* flag that we have moved */
		savematch();
		return(TRUE);

fail:;									/* continue to search */
	}
	return(FALSE);
}

/**
 * Return information depending on whether we have hit a Boundry (and may
 * therefore search no further) or if a trailing character of the search
 * string has been found.  See Boundry() for search restrictions.
 */
static int near fbound(jump, dir)
  int	jump;
  int	dir;
{
  int	spare;

	if (dir == FORWARD)
	{
		while (jump != 0)
		{
			curoff += jump;
			spare = curoff - llength(curline);

			while (spare > 0)
			{
				curline = lforw(curline);		/* skip to next line */
				curoff = spare - 1;
				spare = curoff - llength(curline);
				if (curline == curbp->b_linep)
					return(TRUE);				/* hit end of buffer */
			}

			if (spare == 0)
				jump = deltaf[(int)'\n'];
			else
				jump = deltaf[(int)lgetc(curline, curoff)];
		}

		/* the last character matches, so back up to start of possible match */
		curoff -= patlenadd;

		while (curoff < 0)
		{
			curline = lback(curline);			/* skip back a line */
			curoff += llength(curline) + 1;
		}

	}
	else										/* REVERSE */
	{
		jump++; 								/* allow for offset in reverse*/
		while (jump != 0)
		{
			curoff -= jump;
			while (curoff < 0)
			{
				curline = lback(curline);		/* skip back a line */
				curoff += llength(curline) + 1;
				if (curline == curbp->b_linep)
					return(TRUE);				/* hit end of buffer */
			}

			if (curoff == llength(curline))
				jump = deltab[(int)'\n'];
			else
				jump = deltab[(int)lgetc(curline, curoff)];
		}

		/* the last character matches, so back up to start of possible match */
		curoff += matchlen;
		spare = curoff - llength(curline);
		while (spare > 0)
		{
			curline = lforw(curline);			/* skip back a line */
			curoff = spare - 1;
			spare = curoff - llength(curline);
		}
	}
	return(FALSE);
}

/**
 * Settting up search delta forward and delta backward tables.	The reverse
 * search string and string lengths are set here, for table initialization
 * and for substitution purposes.  The default for any character to jump is
 * the pattern length.
 */
void SetJmpTbl(apat)
  char	apat[];
{
  int	i;
    
	RevStrCpy(tap, apat);
	patlenadd = (mlenold = matchlen = stralen(apat)) - 1;

	for (i = 0; i < MAXCHAR; i++)
	{
		deltaf[i] = matchlen;
		deltab[i] = matchlen;
	}

	/* now put in the characters in the pattern, duplicating the CASE */
	for (i = 0; i < patlenadd; i++)
	{
		if (isletter(apat[i]))
			deltaf[(uint)CHCASE((uint)apat[i])] = patlenadd - i;
		deltaf[(uint)apat[i]] = patlenadd - i;

		if (isletter(tap[i]))
			deltab[(uint)CHCASE((uint)tap[i])] = patlenadd - i;
		deltab[(uint)tap[i]] = patlenadd - i;
	}

	/* The last character will have the pattern length unless there are
	 * duplicates of it.  Get the number to jump from the arrays delta,
	 * and overwrite with zeroes in delta duplicating the CASE.
	 */
	lastchfjump = patlenadd + deltaf[(uint)apat[patlenadd]];
	lastchbjump = patlenadd + deltab[(uint)apat[0]];

	if (isletter(apat[patlenadd]))
		deltaf[(uint)CHCASE(apat[patlenadd])] = 0;
	deltaf[(int)apat[patlenadd]] = 0;

	if (isletter(apat[0]))
		deltab[(uint)CHCASE(apat[0])] = 0;
	deltab[(int)apat[0]] = 0;
}

/**
 * See if a char is a letter ('A'-'Z' or 'a'-'z').	True if so, false otherwise.
 */
static int near isletter(c)
  char	c;
{
	return((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') ? TRUE : FALSE);
}

/**
 * Compare two characters.	The "bc" comes from the buffer, "pc" from the
 * pattern.  If we are not in EXACT mode, fold out the case.
 */
int Eq(bc, pc, emf)
  int	bc, pc; 		/* two characters to compare */
  int	emf;			/* MDEXACT mode flag */
{
	if (!emf)
	{
		bc = (bc >= 'a' && bc <= 'z') ? bc ^ 0x20 : bc; 	/* toupper(bc) */
		pc = (pc >= 'a' && pc <= 'z') ? pc ^ 0x20 : pc; 	/* toupper(pc) */
	}
	return(bc == pc);
}

/**
 * Read a pattern.	Stash it in apat.  If it is the search string, create the
 * reverse pattern and the magic pattern, assuming we are in MAGIC mode (and
 * defined that way). Apat is not updated if the user types in an empty line.
 * If the user typed an empty line, and there is no old pattern, it is an
 * error. Display the old pattern, in the style of Jeff Lomicka. There is some
 * do-it-yourself control expansion.  Change to using <META> to delimit the
 * end-of-pattern to allow <NL>s in the search string. 
 */
static int near readpattern(prompt, apat, srch)
  char	*prompt;
  char	apat[];
  int	srch;
{
  int	status;
  char	tpat[NPAT + 20];

	strcpycat(tpat, prompt, " [");				/* build new prompt string */
	ExpandP(&apat[0], &tpat[stralen(tpat)], NPAT / 2);	/* add old pattern */
	stracat(tpat, "]<META>: ");

	/* Read a pattern.	Either we get one, or we just get the META charater,
	 * and use the previous pattern. Then, if it's the search string, make a
	 * reversed pattern. Then make the Meta-pattern, if we are defined that way.
	 */
	if ((status = MlReplyT(tpat, tpat, NPAT, metac)) == TRUE)
	{
		stracpy(apat, tpat);

		/* if we are doing the search string, set the delta tables */
		if (srch)
			SetJmpTbl(apat);

#if MAGIC
		/* only make the Meta-pattern if in magic mode, since the pattern in
		 * question might have an invalid Meta combination
		 */
		if ((curwp->w_bufp->b_mode & MDMAGIC) == 0)
		{
			McClear();
			rmcclear();
		}
		else
			status = srch ? mcstr() : rmcstr();
#endif
	}
	else if (status == FALSE && apat[0] != 0)	/* old one */
		status = TRUE;

	return(status);
}

/**
 * We found the pattern?  Let's save it away.
 */
static int near savematch()
{
  char	*ptr;		/* pointer to last match string */
  int	i;

	/* free any existing match string, then attempt to allocate a new one */
	if (patmatch != NULLCPTR)
		free(patmatch);

	ptr = patmatch = (char *)malloc(matchlen + 1);

	if (ptr == NULLCPTR)
	{
		OutOfMem();
		return(FALSE);
	}

	scuroff = matchoff;
	scurline = matchline;
	for (i = 0; i < matchlen; i++)
		*ptr++ = (char)nextch(FORWARD);
	*ptr = NULLCHR;

	return(TRUE);
}

/**
 * Reverse string copy
 */
int RevStrCpy(rvstr, str)
  char	*rvstr, *str;
{
  int	i;

	str += (i = stralen(str));
	while (i-- > 0)
		*rvstr++ = *--str;
	*rvstr = NULLCHR;
	return(TRUE);
}

/**
 * Search and replace
 */
int SReplace(f, n)
  int	f, n;
{
	return(replaces(FALSE, f, n));
}

/**
 * Search and replace with query
 */
int QReplace(f, n)
  int	f, n;
{
	return(replaces(TRUE, f, n));
}

/**
 * Search for a string and replace it with another string. Query might be
 * enabled (according to kind).
 */
static int near replaces(kind, f, n)
  int	kind;			/* Query enabled flag */
  int	f;				/* default flag */
  int	n;				/* # of repetitions wanted */
{
  int	status; 		/* success flag on pattern inputs */
  int	rlength;		/* length of replacement string */
  int	numsub; 		/* number of substitutions */
  int	nummatch;		/* number of found matches */
  int	nlflag; 		/* last char of search string a <NL>? */
  int	nlrepl; 		/* was a replace done on the last line? */
  char	c;				/* input char for query */
  char	tpat[NPAT]; 	/* temporary to hold search pattern */
  LINE	*origline;		/* original dot position and offset */
  int	origoff;		/*	(for 'quit back to original position' option) */
  LINE	*lastline;		/* position of last replace and offset */
  int	lastoff;		/*	(for 'quit back to last change' option) */
  char	buf[NSTRING];

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */

	/* check for negative repetitions */
	if (f && n < 0)
		return(FALSE);

	/* ask the user for the text of a pattern */
	if ((status = readpattern(	(
										kind == FALSE
									?
										"Replace"
									:
										"Query replace"
								)
								,&pat[0], TRUE)) != TRUE)
		return(status);

	/* check for a null REPLACE string */
	if (pat[0] == 0)
	{
		MlWrite("Can't replace a NULL string");
		return(FALSE);
	}

	/* ask for the replacement string */
	if ((status = readpattern("with", &rpat[0], FALSE)) != TRUE)
		return(status);

	/* find the length of the replacement string */
	rlength = stralen(&rpat[0]);

	/* set up flags so we can make sure not to do a recursive replace on the
	 * last line
	 */
	nlflag = (pat[matchlen - 1] == '\n');
	nlrepl = FALSE;

	if (kind)
	{
		/* build query replace question string */
		stracpy(tpat, "Replace '");
		ExpandP(&pat[0], &tpat[stralen(tpat)], NPAT / 3);
		stracat(tpat, "' with '");
		ExpandP(&rpat[0], &tpat[stralen(tpat)], NPAT / 3);
		stracat(tpat, "'? ");

		/* initialize last replaced pointers */
		lastline = NULLLPTR;
		lastoff = 0;
	}

	/* save original dot position, init the number of matches and
	 * substitutions, and scan through the file
	 */
	origline = curwp->w_dotp;
	origoff = curwp->w_doto;
	numsub = 0;
	nummatch = 0;

	while ((f == FALSE || n > nummatch) && (nlflag == FALSE || nlrepl == FALSE))
	{
		/* Search for the pattern. If we search with a regular expression,
		 * matchlen is reset to the true length of the matched string.
		 */
#if MAGIC
		if ((magical && curwp->w_bufp->b_mode & MDMAGIC) != 0)
		{
			if (!mcscanner(&mcpat[0], FORWARD, PTBEG))
				break;
		}
		else
#endif
			if (!Scanner(&pat[0], FORWARD, PTBEG))
				break;						/* all done */

		 ++nummatch;						/* increment # of matches */

		/* check if we are on the last line */
		nlrepl = (lforw(curwp->w_dotp) == curwp->w_bufp->b_linep);

		/* check for query */
		if (kind)
		{
			/* get the query */
pprompt:
			MlWrite(&tpat[0]);
qprompt:
			Update(TRUE);				/* show the proposed place to change */
			c = (char)ExtChToCh(TermGetCh());	/*	 and input */
			MlWrite("");				/*	 and clear it */

			/* and respond appropriately */
			switch (c)
			{
				case 'y':		/* yes, substitute */
				case 'Y':
					break;
				case 'n':		/* no, onword */
				case 'N':
					ForwChar(FALSE, 1);
					continue;
				case 'd':		/* yes/stop asking */
				case 'D':
					kind = FALSE;
					break;
				case 'u':		/* undo last and re-prompt */
				case 'U':
					/* restore old position */
					if (lastline == NULLLPTR)
					{
						/* there is nothing to undo */
						PcPutC(BELL);
						goto pprompt;
					}
					curwp->w_dotp = lastline;
					curwp->w_doto = lastoff;
					lastline = NULLLPTR;
					lastoff = 0;

					/* delete the new string */
					BackChar(FALSE, rlength);
#if MAGIC
					if (!delins(rlength, patmatch, FALSE))
#else
					if (!delins(rlength, patmatch))
#endif
						return(FALSE);

					/* record one less substitution, backup, save our place,
					 * and reprompt
					 */
					--numsub;
					BackChar(FALSE, mlenold);
					matchline = curwp->w_dotp;
					matchoff = curwp->w_doto;
					goto pprompt;
				case 'l':		/* Quit back to last change */
				case 'L':
					/* restore previous position (if there is one) */
					if (lastline != NULLLPTR)
					{
						curwp->w_dotp = lastline;
						curwp->w_doto = lastoff;
						curwp->w_flag |= WFMOVE;
					}
					goto quitqr;
				case 'b':		/* Quit back (FALLING THRU TO NEXT CASE!!) */
				case 'B':
					/* restore old position */
					curwp->w_dotp = origline;
					curwp->w_doto = origoff;
					curwp->w_flag |= WFMOVE;
				case 'q':		/* Quit and stay here */
				case 'Q':
quitqr:
					MlWrite("Quitting");
					return(FALSE);
				default:		/* bitch and beep */
					PcPutC(BELL);
					MlWrite	(

	"(Y)es, (N)o, (D)o all, (U)ndo last, (Q)uit here, Quit (B)egin, Quit (L)ast"

							);
					goto qprompt;
			}
		}

		/* if this is the point origin, flag so we a can reset it */
		if (curwp->w_dotp == origline)
		{
			origline = NULLLPTR;
			lastline = curwp->w_dotp->l_bp;
		}

		/* delete the sucker, and insert its replacement */
#if MAGIC
		status = delins(matchlen, &rpat[0], TRUE);
#else
		status = delins(matchlen, &rpat[0]);
#endif
		if (origline == NULLLPTR)
		{
			origline = lastline->l_fp;
			origoff = 0;
		}
		if (status != TRUE)
			return(status);

		/* Save our position, since we may undo this.  If we are not querying,
		 * check to make sure that we didn't replace an empty string (possible
		 * in MAGIC mode), because we'll enter the land of 'loopus infinitus'.
		 */
		if (kind)
		{
			lastline = curwp->w_dotp;
			lastoff = curwp->w_doto;
		}
		else if (matchlen == 0)
		{
			MlWrite("Empty string replaced.  Stopping.");
			return(FALSE);
		}

		numsub++;		/* increment # of substitutions */
	}

	/* and report the results */
	strfmt(buf, "%d substitutions", numsub);
	MlWrite(buf);

	/* clear the keyboard buffer (in case too many (Y)eses were pressed */
	kbflush();

	return(TRUE);
}

/**
 * Delete a specified length from the current point then either insert the
 * string directly, or make use of replacement Meta-array.
 */
#if MAGIC
static int near delins(dlength, instr, use_meta)
#else
static int near delins(dlength, instr)
#endif
  int	dlength;
  char	*instr;
#if MAGIC
  int	use_meta;
#endif
{
  int	status;
#if MAGIC
  RMC	*rmcptr;
#endif

	/* zap what we gotta, and insert its replacement */
	if ((status = LDelete((long)dlength, FALSE)) != TRUE)
		MlWrite("Error while deleting");
	else
#if MAGIC
		if ((rmagical && use_meta) && (curwp->w_bufp->b_mode & MDMAGIC) != 0)
		{
			rmcptr = &rmcpat[0];
			while (rmcptr->mc_type != MCNIL && status == TRUE)
			{
				if (rmcptr->mc_type == LITCHAR)
					status = LInstr(rmcptr->rstr);
				else
					status = LInstr(patmatch);
				rmcptr++;
			}
		}
		else
#endif
			status = LInstr(instr);
	return(status);
}

/**
 * Expand control key sequences for output
 */
int ExpandP(srcstr, deststr, maxlength)
  char	*srcstr;		/* string to expand */
  char	*deststr;		/* destination of expanded string */
  int	maxlength;		/* maximum chars in destination */
{
  uchar c;				/* current char to translate */

	/* scan through the string */
	while ((c = *srcstr++) != 0)
	{
		if (c == '\n')					/* it's a NewLine */
		{
			*deststr++ = '<';
			*deststr++ = 'N';
			*deststr++ = 'L';
			*deststr++ = '>';
			maxlength -= 4;
		}
		else if (c < 0x20 || c == 0x7f) /* control character */
		{
			*deststr++ = '^';
			*deststr++ = c ^ 0x40;
			maxlength -= 2;
		}
		else							/* any other character */
		{
			*deststr++ = c;
			maxlength--;
		}

		/* check for maxlength */
		if (maxlength < 4)
		{
			*deststr++ = '$';
			*deststr = NULLCHR;
			return(FALSE);
		}
	}
	*deststr = NULLCHR;
	return(TRUE);
}

/**
 * Return information depending on whether we may search no further.  Beginning
 * of file and end of file are the obvious cases, but we may want to add further
 * optional Boundry restrictions in the future, a' la VMS EDT.	At the moment,
 * just return TRUE or FALSE depending on if a Boundry is hit (ouch).
 */
int Boundry(cline, coff, dir)
  LINE	*cline;
  int	coff;
  int	dir;
{
	if (dir == FORWARD)
		return(coff == llength(cline) && (lforw(cline) == curbp->b_linep));
	else
		return(coff == 0 && (lback(cline) == curbp->b_linep));
}

/**
 * Retrieve the next/previous character in the buffer, and advance/retreat the
 * point. The order in which this is done is significant, and depends upon the
 * direction of the search.  Forward searches look at the current character and
 * move, reverse searches move and look at the character.
 */
static int near nextch(dir)
  int	dir;
{
  int	c;

	if (dir == FORWARD)
	{
		if (scuroff == llength(scurline))	/* if at EOL */
		{
			scurline = lforw(scurline); 	/* skip to next line */
			scuroff = 0;
			c = '\n';						/* and return a <NL> */
		}
		else
			c = lgetc(scurline, scuroff++); /* get the char */
	}
	else									/* dir == REVERSE */
	{
		if (scuroff == 0)
		{
			scurline = lback(scurline);
			scuroff = llength(scurline);
			c = '\n';
		}
		else
			c = lgetc(scurline, --scuroff);
	}
	return(c);
}

/**
 * Put back the line and offset values to their previous values.  Used when a
 * nextch() returns a char that does not match the pattern.
 */
static void near putbackch(dir)
  int	dir;
{
	if (dir == REVERSE)
	{
		if (scuroff == llength(scurline))	/* if at EOL */
		{
			scurline = lforw(scurline); 	/* skip to next line */
			scuroff = 0;
		}
		else
			++scuroff;
	}
	else									/* dir == FORWARD */
	{
		if (scuroff == 0)
		{
			scurline = lback(scurline);
			scuroff = llength(scurline);
		}
		else
			--scuroff;
	}
}

#if MAGIC
/**
 * Set up the 'magic' array.  The closure symbol is taken as a literal
 * character when:  1) it is the first character in the pattern, and
 * 2) when preceded by a symbol that does not allow closure, such as a
 * NewLine, beginning of line symbol, or another closure symbol.
 */
static int near mcstr()
{
  MC	*mcptr, *rtpcm;
  char	*patptr;
  int	mj;
  int	pchr;
  int	status = TRUE;
  int	does_closure = FALSE;

	/* If we had metacharacters in the MC array previously, free up any bitmaps
	 * that may have been allocated.
	 */
	if (magical)
		McClear();

	mj = 0;
	mcptr = &mcpat[0];
	patptr = &pat[0];

	while (((pchr = (int)*patptr) != NULL) && status)
	{
		switch (pchr)
		{
			case MC_BCCL:
				status = cclmake(&patptr, mcptr);
				magical = TRUE;
				does_closure = TRUE;
				break;
			case MC_BOL:
				if (mj != 0)
					goto litcase;
				mcptr->mc_type = BOL;
				magical = TRUE;
				break;
			case MC_EOL:
				if (*(patptr + 1) != NULLCHR)
					goto litcase;
				mcptr->mc_type = EOL;
				magical = TRUE;
				break;
			case MC_ANY:
				mcptr->mc_type = ANY;
				magical = TRUE;
				does_closure = TRUE;
				break;
			case MC_CLOSURE:
				/* Does the closure symbol mean closure here? If so, back up to
				 * the previous element and indicate it is enclosed.
				 */
				if (!does_closure)
					goto litcase;
				mj--;
				mcptr--;
				mcptr->mc_type |= CLOSURE;
				magical = TRUE;
				does_closure = FALSE;
				break;
			case MC_ESC:	/* Note: no break between MC_ESC case and default */
				if (*(patptr + 1) != NULLCHR)
				{
					pchr = *++patptr;
					magical = TRUE;
				}
			default:
litcase:		mcptr->mc_type = LITCHAR;
				mcptr->u.lchar = pchr;
				does_closure = (pchr != '\n');
				break;
		}
		mcptr++;
		patptr++;
		mj++;
	}

	/* close off the Meta-string */
	mcptr->mc_type = MCNIL;

	/* Set up the reverse array, if status is good. Please note the structure
	 * assignment - your compiler may not like that. If the status is not good,
	 * nil out the Meta-pattern. The only way the status would be bad is from
	 * the cclmake() routine, and the bitmap for that member is guaranteed to be
	 * freed.  So we stomp a MCNIL value there, and call McClear() to free any
	 * other bitmaps.
	 */
	if (status)
	{
		rtpcm = &tapcm[0];
		while (--mj >= 0)
		{
			*rtpcm++ = *--mcptr;
		}
		rtpcm->mc_type = MCNIL;
	}
	else
	{
		(--mcptr)->mc_type = MCNIL;
		McClear();
	}
	return(status);
}

/**
 * Set up the replacement 'magic' array.  Note that if no Meta-characters are
 * encountered in the replacement string, the array is never actually created
 * - we will just use the character array rpat[] as the replacement string.
 */
static int near rmcstr()
{
  RMC	*rmcptr;
  char	*patptr;
  int	status = TRUE;
  int	mj;

	patptr = &rpat[0];
	rmcptr = &rmcpat[0];
	mj = 0;
	rmagical = FALSE;

	while (*patptr && status == TRUE)
	{
		switch (*patptr)
		{
			case MC_DITTO:
				/* If there were non-magical characters in the string before
				 * reaching this character, plunk it in the replacement array
				 * before processing the current Meta-character.
				 */
				if (mj != 0)
				{
					rmcptr->mc_type = LITCHAR;
					if ((rmcptr->rstr = (char *)malloc(mj + 1)) == NULLCPTR)
					{
						OutOfMem();
						return(FALSE);
					}
					strancpy(rmcptr->rstr, patptr - mj, mj + 1);
					rmcptr++;
					mj = 0;
				}
				rmcptr->mc_type = DITTO;
				rmcptr++;
				rmagical = TRUE;
				break;
			case MC_ESC:
				rmcptr->mc_type = LITCHAR;

				/* We malloc mj plus two here, instead of one, because we have
				 * to count the current character.
				 */
				if ((rmcptr->rstr = (char *)malloc(mj + 2)) == NULLCPTR)
				{
					OutOfMem();
					return(FALSE);
				}

				strancpy(rmcptr->rstr, patptr - mj, mj + 2);

				/* If MC_ESC is not the last character in the string, find out
				 * what it is escaping, and overwrite the last character with it
				 */
				if (*(patptr + 1) != NULLCHR)
					*((rmcptr->rstr) + mj) = *++patptr;

				rmcptr++;
				mj = 0;
				rmagical = TRUE;
				break;
			default:
				mj++;
		}
		patptr++;
	}

	if (rmagical && mj > 0)
	{
		rmcptr->mc_type = LITCHAR;
		if ((rmcptr->rstr = (char *)malloc(mj + 1)) == NULLCPTR)
		{
			OutOfMem();
			return(FALSE);
		}
		strancpy(rmcptr->rstr, patptr - mj, mj + 1);
		rmcptr++;
	}
	rmcptr->mc_type = MCNIL;
	return(TRUE);
}

/**
 * Free up any CCL bitmaps, and MCNIL the MC search arrays
 */
int McClear()
{
  MC	*mcptr;

	mcptr = &mcpat[0];

	while (mcptr->mc_type != MCNIL)
	{
		if	(
					(mcptr->mc_type & MASKCL) == CCL
				||
					(mcptr->mc_type & MASKCL) == NCCL
			)
			free(mcptr->u.cclmap);
		mcptr++;
	}
	mcpat[0].mc_type = tapcm[0].mc_type = MCNIL;
	magical = FALSE;
	return(TRUE);
}

/**
 * Free up any strings, and MCNIL the RMC array
 */
static int near rmcclear()
{
  RMC	*rmcptr;

	rmcptr = &rmcpat[0];

	while (rmcptr->mc_type != MCNIL)
	{
		if (rmcptr->mc_type == LITCHAR)
			free(rmcptr->rstr);
		rmcptr++;
	}
	rmcpat[0].mc_type = MCNIL;
	return(TRUE);
}

/**
 * Meta-character equality with a character
 */
static int near mceq(bc, mt)
  int	bc;
  MC	*mt;
{
  int	result;
  char	buf[NSTRING];

	switch (mt->mc_type & MASKCL)
	{
		case LITCHAR:
			result = Eq(bc, mt->u.lchar, curwp->w_bufp->b_mode & MDEXACT);
			break;
		case ANY:
			result = (bc != '\n');
			break;
		case CCL:
			if (!(result = biteq(bc, mt->u.cclmap)))
			{
				if	(
							!(curwp->w_bufp->b_mode & MDEXACT)
						&&
							(isletter(bc))
					)
				{
					result = biteq(CHCASE(bc), mt->u.cclmap);
				}
			}
			break;
		case NCCL:
			result = !biteq(bc, mt->u.cclmap);
			if	(
						!(curwp->w_bufp->b_mode & MDEXACT)
					&&
						(isletter(bc))
				)
			{
				result &= !biteq(CHCASE(bc), mt->u.cclmap);
			}
			break;
		default:
			strfmt(buf, "mceq: what is %d?", mt->mc_type);
			MlWrite(buf);
			result = FALSE;
			break;

	}
	return(result);
}

/**
 * Create the bitmap for the character class. ppatptr is left pointing to the
 * end-of-character-class character, so that a loop may automatically increment
 * with safety.
 */
static int near cclmake(ppatptr, mcptr)
  char		**ppatptr;
  MC		*mcptr;
{
  BITMAP	bmap;
  char		*patptr;
  int		pchr, ochr;

	if ((bmap = clearbits()) == (BITMAP)NULLCPTR)
	{
		OutOfMem();
		return(FALSE);
	}

	mcptr->u.cclmap = bmap;
	patptr = *ppatptr;

	/* Test the initial character(s) in ccl for special cases - negate ccl, or
	 * an end ccl character as a first character.  Anything else gets set in
	 * the bitmap.
	 */
	if (*++patptr == MC_NCCL)
	{
		patptr++;
		mcptr->mc_type = NCCL;
	}
	else
		mcptr->mc_type = CCL;

	if ((ochr = *patptr) == MC_ECCL)
	{
		MlWrite("No characters in character class");
		free(bmap);
		return(FALSE);
	}
	else
	{
		if (ochr == MC_ESC)
			ochr = *++patptr;
		setbit(ochr, bmap);
		patptr++;
	}

	while (ochr != NULLCHR && (pchr = *patptr) != MC_ECCL)
	{
		switch (pchr)
		{
			/* range character loses its meaning if it is the last character
			 * in the class
			 */
			case MC_RCCL:
				if (*(patptr + 1) == MC_ECCL)
					setbit(pchr, bmap);
				else
				{
					pchr = *++patptr;
					while (++ochr <= pchr)
						setbit(ochr, bmap);
				}
				break;
			case MC_ESC:	/* Note: no break between case MC_ESC and default */
				pchr = *++patptr;
			default:
				setbit(pchr, bmap);
				break;
		}
		patptr++;
		ochr = pchr;
	}

	*ppatptr = patptr;

	if (ochr == NULLCHR)
	{
		MlWrite("Character class not ended");
		free(bmap);
		return(FALSE);
	}
	return(TRUE);
}

/**
 * Is the character in the bitmap?
 */
static int near biteq(bc, cclmap)
  int		bc;
  BITMAP	cclmap;
{
	if (bc >= HICHAR)
		return(FALSE);
	return((*(cclmap + (bc >> 3)) & BIT(bc & 7)) ? TRUE : FALSE);
}

/**
 * Allocate and zero out a CCL bitmap
 */
static BITMAP near clearbits()
{
  BITMAP	cclstart, cclmap;
  int		i;

	if ((cclmap = cclstart = (BITMAP)malloc(BMSIZE)) != (BITMAP)NULLCPTR)
		for (i = 0; i < BMSIZE; i++)
			*cclmap++ = 0;
	return(cclstart);
}

/**
 * Set a bit (ON only) in the bitmap
 */
static void near setbit(bc, cclmap)
  int		bc;
  BITMAP	cclmap;
{
	if (bc < HICHAR)
		*(cclmap + (bc >> 3)) |= BIT(bc & 7);
}
#endif
