/*
 *	S P A W N . C
 *
 *	purpose:
 *		Various DOS access commands for CodEd.
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <stdlib.h>
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

struct pblock			/* EXEC parameter block */
{
	short	envptr;		/* short pointer to environment string */
	char	*cline;		/* long pointer to command line */
	char	*fcb1;		/* long pointer to FCB at PSP + 0x5c */
	char	*fcb2;		/* long pointer to FCB at PSP + 0x6c */
};
typedef	struct pblock	PBLOCK;

/*
 * local function prototypes
 */

static int near		oneliner(char *);
static int near		execprog(char *);

/*
 * local (module global) variable declarations
 */

static char		vtype[NBUFN];	/* size of sres[] in def.h */
static char		f1[38];			/* FCB1 (not initialized) */
static char		f2[38];			/* FCB2 (not initialized) */
static PBLOCK	parmblk;

static union REGS		rg;		/* registers for DOS call */
static struct SREGS		srg;	/* segment registers for DOS calls */

/**
 * Spawn a DOS shell (uses COMSPEC)
 */
int DosShell(f, n)
  int	f, n;
{
  int	err, savcle;

	savcle = clexec;
	clexec = TRUE;
	err = oneliner("");
	clexec = savcle;
	return(err);
}

/**
 * Run a one-liner in a DOS shell
 */
int Spawn(f, n)
  int	f, n;
{
  int	status;
  char	line[NLINE];

	if ((status = MlReply("Spawn: ", line, NLINE)) != TRUE)
		return(status);
	return(oneliner(line));
}

/**
 * Run a one-liner as a child process
 */
int Exec(f, n)
  int	f, n;
{
  int	status;
  char	line[NLINE];

	if ((status = MlReply("Exec: ", line, NLINE)) != TRUE)
		return(status);
	return(execprog(line));
}

/**
 * Pipe a one-liner into a buffer
 */
int Pipe(f, n)
  int	f, n;
{
  int			status;				/* return status */
  WINDOW		*wp; 				/* pointer to new window */
  BUFFER		*bp; 				/* pointer to buffer to zot */
  char			line[NLINE];		/* command line send to shell & env. str */
  char			filnam[NSTRING];
  static char	tmpfil[] = "tmp.ce";

	if (syenvstr("TMP", line, NLINE) == RC_FAIL)
		filnam[0] = 0;
	else
	{	stracpy(filnam, line);
		if (filnam[stralen(filnam) - 1] != '\\')
			stracat(filnam, "\\");
	}
	stracat(filnam, tmpfil);

	/* get the command to Pipe in */
	if ((status = MlReply("Pipe: ", line, NLINE)) != TRUE)
		return(status);

	/* get rid of the command output buffer if it exists */
	if ((bp = FindBuf(tmpfil, FALSE, 0)) != FALSE)
	{
		/* try to make sure we are off screen */
		wp = wheadp;
		while (wp != NULLWPTR)
		{
			if (wp->w_bufp == bp)
			{
				OnlyWind(TRUE, 1);
				break;
			}
			wp = wp->w_wndp;
		}
		if (ZotBuf(bp) != TRUE)
			return(FALSE);
	}

	stracat(line, " >");
	stracat(line, filnam);
	oneliner(line);

	/* split the current window to make room for the command output */
	if (!SplitWind(FALSE, 1))
		return(FALSE);

	/* and read the stuff in */
	if (!GetFile(filnam))
		return(FALSE);

	/* set to VIEW mode, clear the file name, and Update all mode lines */
	curwp->w_bufp->b_mode |= MDVIEW;
	stracpy(curwp->w_bufp->b_fname, "");
	UpMode();

	/* and get rid of the temporary file */
	drunlink(filnam);
	return(TRUE);
}

/**
 * Make the actual system call, then mark the screen as garbage so a full
 * repaint is done.  If we are interactive, wait for a single character
 * to be typed before continuing.
 */
static int near oneliner(cmd)
  char			*cmd;
{
  char			*shell;			/* DOS command processor */
  char			cmndl[NSTRING];	/* constructed command line */
  char			*cmnd;			/* pointer to cmndl[] */
  static char	bob[] = " /c ";	/* current switch char plus other stuff */

	/* detect current switch character and set us up to use it */
	rg.h.ah = 0x37;				/* get setting data */
	rg.h.al = 0x00;				/* get switch character */
	intdos(&rg, &rg);
	bob[1] = rg.h.dl;			/* Update bob[] with the current switchar */

	if (syenvstr("COMSPEC", cmndl, NSTRING) == RC_FAIL)
		return(FALSE);

	while (*cmd == ' ' || *cmd == '\t')
		++cmd;

	cmnd = cmndl;
	if (*cmd)
	{
		stracat(cmnd, bob);
		stracat(cmnd, cmd);
	}
	return(execprog(cmnd));
}

/**
 * EXEC a command line
 */
static int near execprog(cmd)
  char	*cmd;			/* command line to execute */
{
  char	*sp;			/* temp string pointer */
  char	prog[NSTRING];	/* program filespec */
  char	tail[NSTRING];	/* command tail with length byte */
  int	err;			/* DOS error code */

	stracpy(vtype, sres);	/* save current display mode */
	PcClose();				/* reset to 25 line display mode */

	/* strip leading white space, if any */
	while (*cmd == ' ' || *cmd == '\t')
		++cmd;

	/* grab the program name */
	sp = prog;
	while (*cmd && (*cmd != ' ') && (*cmd != '\t'))
		*sp++ = *cmd++;
	*sp = 0;

	/* strip white space between program name and command line */
	while (*cmd && ((*cmd == ' ') || (*cmd == '\t')))
		cmd++;

	/* first char is command line length; remainder is <CR> terminated */
	tail[0] = (char)(stralen(cmd));
	strcpycat(&tail[1], cmd, "\r");

	/* look up the program on the path trying various extentions */
	if ((sp = FileLook(prog, FALSE)) == NULLCPTR)
	{
		if ((sp = FileLook(stracat(prog, ".exe"), FALSE)) == NULLCPTR)
		{
			stracpy(&prog[stralen(prog) - 4], ".com");
			if ((sp = FileLook(prog, FALSE)) == NULLCPTR)
			{
				err = -1;			/* flag it so we return FALSE */
				goto errexit;		/* go reset the video mode and go home */
			}
		}
	}
	stracpy(prog, sp);

	segread(&srg);

	/* set up the EXEC parameter block */
	parmblk.envptr = 0;				/* make child inherit parent's environment*/
	parmblk.cline = tail;			/* parameter line pointer */
	parmblk.fcb1 = f1;				/* point to a blank FCB */
	parmblk.fcb2 = f2;				/* point to a blank FCB */

	/* and make the call */
	rg.h.ah = 0x4b;					/* EXEC Load or Execute a Program */
	rg.h.al = 0x00;					/* load and execute function subcode */

	srg.ds = FP_SEG(prog);
	rg.x.dx = FP_OFF(prog);
	srg.es = FP_SEG(&parmblk);
	rg.x.bx = FP_OFF(&parmblk);

	intdosx(&rg, &rg, &srg);
	if (rg.x.cflag == 0)			/* successful execution */
	{
		rg.h.ah = 0x4d;				/* get child process return code */
		intdos(&rg, &rg);
		err = rg.x.ax;				/* and save it */
	}
	else
		err = -_doserrno;			/* failed child call (return as negative) */

	/* if we are interactive, pause here */
	if (!clexec)
	{
		MlPutS("\r\n\n[End]");
		TermGetCh();
	}

errexit:

	sgarbf = TRUE;
	PcRez(vtype);					/* restore display mode */
	return(err < 0 ? FALSE : TRUE);
}
