/*
 *	V A R . H
 *
 *	purpose:
 *		Environment and user variable definitions for CodEd.
 */

/* structure to hold user variables and their definitions */

struct uvar
{
	char		u_name[NVSIZE + 1]; 	/* name of user variable */
	char	   *u_value;				/* value (string) */
};
typedef struct uvar 	UVAR;

/* current user variables (this structure will probably change) */

#define MAXVARS 63

UVAR			uv[MAXVARS + 1];		/* user variables */

/* list of recognized environment variables */

char *envars[] =
{
    "cbufname",     /* current buffer name */
    "cfname",       /* current file name */
    "cmode",        /* mode of current buffer */
    "curchar",      /* current character under the cursor */
    "curcol",       /* current column pos of cursor */
    "curline",      /* current line in file */
    "cwline",       /* current screen line in window */
    "debug",        /* macro debugging */
    "discmd",       /* display commands on command line */
    "disinp",       /* display command line input characters */
    "gmode",        /* global modes */
    "hardtab",      /* current hard tab size */
    "hjump",        /* horizontal screen jump size */
    "hscroll",      /* horizontal scrolling flag */
    "kill",         /* kill buffer (read only) */
    "lastkey",      /* last keyboard char struck */
    "line",         /* text of current line */
    "lwidth",       /* width of current line */
    "match",        /* last matched magic pattern */
    "pagelen",      /* number of lines used by editor */
    "pending",      /* type ahead pending flag */
    "replace",      /* replacement pattern */
    "rval",         /* child process return value */
    "safesave",     /* safe save flag */
    "search",       /* search pattern */
    "softtab",      /* current soft tab size */
    "sres",         /* current screen resolution */
    "status",       /* returns the status of the last command */
    "target",       /* target for line moves */
    "tpause",       /* length to pause for paren matching */
    "wline",        /* # of lines in current window */
};

#define NEVARS	sizeof(envars) / sizeof(char *)

/* and its preprocesor definitions */

#define EVCBUFNAME  0
#define EVCFNAME    1
#define EVCMODE     2
#define EVCURCHAR   3
#define EVCURCOL    4
#define EVCURLINE   5
#define EVCWLINE    6
#define EVDEBUG     7
#define EVDISCMD    8
#define EVDISINP    9
#define EVGMODE     10
#define EVHARDTAB   11
#define EVHJUMP     12
#define EVHSCROLL   13
#define EVKILL      14
#define EVLASTKEY   15
#define EVLINE      16
#define EVLWIDTH    17
#define EVMATCH     18
#define EVPAGELEN   19
#define EVPENDING   20
#define EVREPLACE   21
#define EVRVAL      22
#define EVSAFESAVE  23
#define EVSEARCH    24
#define EVSOFTTAB   25
#define EVSRES      26
#define EVSTATUS    27
#define EVTARGET    28
#define EVTPAUSE    29
#define EVWLINE     30

/* list of recognized user functions */

struct ufunc
{
	char	   *f_name; 	/* name of function */
	int 		f_type; 	/* 1 = monamic, 2 = dynamic */
};
typedef struct ufunc	UFUNC;

#define NILNAMIC	0
#define MONAMIC 	1
#define DYNAMIC 	2
#define TRINAMIC	3

UFUNC funcs[] =
{
    "abs", MONAMIC,     /* absolute value of a number */
    "add", DYNAMIC,     /* add two numbers together */
    "and", DYNAMIC,     /* logical and */
    "asc", MONAMIC,     /* char to integer conversion */
    "ban", DYNAMIC,     /* bitwise AND */
    "bin", MONAMIC,     /* lookup what function name is bound to a key */
    "bno", MONAMIC,     /* bitwise NOT */
    "bor", DYNAMIC,     /* bitwise OR */
    "bxo", DYNAMIC,     /* bitwise XOR */
    "cat", DYNAMIC,     /* concatanate string */
    "chr", MONAMIC,     /* integer to char conversion */
    "div", DYNAMIC,     /* division */
    "env", MONAMIC,     /* retrieve a system environment var */
    "equ", DYNAMIC,     /* logical equality check */
    "exi", MONAMIC,     /* check if a file exists */
    "fin", MONAMIC,     /* look for a file on the path */
    "gre", DYNAMIC,     /* logical greater than */
    "gtk", NILNAMIC,    /* get 1 character */
    "ind", MONAMIC,     /* evaluate indirect value */
    "lef", DYNAMIC,     /* left string(string, len) */
    "len", MONAMIC,     /* string length */
    "les", DYNAMIC,     /* logical less than */
    "low", MONAMIC,     /* lower case string */
    "mid", TRINAMIC,    /* mid string(string, pos, len) */
    "mod", DYNAMIC,     /* mod */
    "neg", MONAMIC,     /* negate */
    "not", MONAMIC,     /* logical not */
    "or",  DYNAMIC,     /* logical or */
    "rig", DYNAMIC,     /* right string(string, pos) */
    "seq", DYNAMIC,     /* string logical equality check */
    "sgr", DYNAMIC,     /* string logical greater than */
    "sin", DYNAMIC,     /* find the index of one string in another */
    "sle", DYNAMIC,     /* string logical less than */
    "sub", DYNAMIC,     /* subtraction */
    "tim", DYNAMIC,     /* multiplication */
    "tru", MONAMIC,     /* truth of the universe logical test */
    "upp", MONAMIC,     /* uppercase string */
};

#define NFUNCS	sizeof(funcs) / sizeof(UFUNC)

/* and its preprocesor definitions */

#define UFABS       0
#define UFADD       1
#define UFAND       2
#define UFASCII     3
#define UFBAND      4
#define UFBIND      5
#define UFBNOT      6
#define UFBOR       7
#define UFBXOR      8
#define UFCAT       9
#define UFCHR       10
#define UFDIV       11
#define UFENV       12
#define UFEQUAL     13
#define UFEXIST     14
#define UFFIND      15
#define UFGREATER   16
#define UFGTKEY     17
#define UFIND       18
#define UFLEFT      19
#define UFLENGTH    20
#define UFLESS      21
#define UFLOWER     22
#define UFMID       23
#define UFMOD       24
#define UFNEG       25
#define UFNOT       26
#define UFOR        27
#define UFRIGHT     28
#define UFSEQUAL    29
#define UFSGREAT    30
#define UFSINDEX    31
#define UFSLESS     32
#define UFSUB       33
#define UFTIMES     34
#define UFTRUTH     35
#define UFUPPER     36
