/*
 *	W O R D . C
 *
 *	purpose:
 *		The routines in this file implement commands that work a word
 *		at a time.
 */

#include <stdio.h>			/* Turbo C++ verison 1.0 */
#include <dos.h>

#include <pdefs.h>			/* PforCe version 1.05 (includes jwg.h) */
#include <pf_ansi.h>

#include "struct.h"
#include "def.h"
#include "protos.h"

/*
 * local function prototypes
 */

static int near		inword(void);

/**
 * Move the cursor backward by "n" words. All of the details of motion are
 * performed by the "BackChar" and "ForwChar" routines. Error if you try to
 * move beyond the buffers.
 */
int BackWord(f, n)
  int	f, n;
{
	if (n < 0)
		return(ForwWord(f, -n));
	if (BackChar(FALSE, 1) == FALSE)
		return(FALSE);
	while (n--)
	{
		while (inword() == FALSE)
		{
			if (BackChar(FALSE, 1) == FALSE)
				return(FALSE);
		}
		while (inword() != FALSE)
		{
			if (BackChar(FALSE, 1) == FALSE)
				return(FALSE);
		}
	}
	return(ForwChar(FALSE, 1));
}

/**
 * Move the cursor forward by the specified number of words. All of the motion
 * is done by "ForwChar". Error if you try and move beyond the buffer's end.
 */
int ForwWord(f, n)
  int	f, n;
{
	if (n < 0)
		return(BackWord(f, -n));
	while (n--)
	{
		while (inword() == TRUE)
		{
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
		}

		while (inword() == FALSE)
		{
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
		}
	}
	return(TRUE);
}

/**
 * Move the cursor forward by the specified number of words. As you move,
 * convert any characters to upper case. Error if you try and move beyond the
 * end of the buffer.
 */
int UpperWord(f, n)
  int	f, n;
{
  int	c;

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if (n < 0)
		return(FALSE);
	while (n--)
	{
		while (inword() == FALSE)
		{
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
		}
		while (inword() != FALSE)
		{
			c = lgetc(curwp->w_dotp, curwp->w_doto);
			if (c >= 'a' && c <= 'z')
			{
				c ^= 0x20;
				lputc(curwp->w_dotp, curwp->w_doto, (char)c);
				LChange(WFHARD);
			}
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
		}
	}
	return(TRUE);
}

/**
 * Move the cursor forward by the specified number of words. As you move
 * convert characters to lower case. Error if you try and move over the end
 * of the buffer.
 */
int LowerWord(f, n)
  int	f, n;
{
  int	c;

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if (n < 0)
		return(FALSE);
	while (n--)
	{
		while (inword() == FALSE)
		{
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
		}
		while (inword() != FALSE)
		{
			c = lgetc(curwp->w_dotp, curwp->w_doto);
			if (c >= 'A' && c <= 'Z')
			{
				c ^= 0x20;
				lputc(curwp->w_dotp, curwp->w_doto, (char)c);
				LChange(WFHARD);
			}
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
		}
	}
	return(TRUE);
}

/**
 * Move the cursor forward by the specified number of words. As you move
 * convert the first character of the word to upper case, and subsequent
 * characters to lower case. Error if you try and move past the end of the
 * buffer.
 */
int CapWord(f, n)
  int	f, n;
{
  int	c;

	if (curbp->b_mode & MDVIEW) /* don't allow this command if */
		return(ReadOnly());		/* we are in read only mode */
	if (n < 0)
		return(FALSE);
	while (n--)
	{
		while (inword() == FALSE)
		{
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
		}
		if (inword() != FALSE)
		{
			c = lgetc(curwp->w_dotp, curwp->w_doto);
			if (c >= 'a' && c <= 'z')
			{
				c ^= 0x20;
				lputc(curwp->w_dotp, curwp->w_doto, (char)c);
				LChange(WFHARD);
			}
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
			while (inword() != FALSE)
			{
				c = lgetc(curwp->w_dotp, curwp->w_doto);
				if (c >= 'A' && c <= 'Z')
				{
					c ^= 0x20;
					lputc(curwp->w_dotp, curwp->w_doto, (char)c);
					LChange(WFHARD);
				}
				if (ForwChar(FALSE, 1) == FALSE)
					return(FALSE);
			}
		}
	}
	return(TRUE);
}

/**
 * Kill forward by "n" words. Remember the location of dot. Move forward by
 * the right number of words. Put dot back where it was and issue the kill
 * command for the right number of characters. With a zero argument, just
 * kill one word and no whitespace.
 */
int DelFWord(f, n)
  int	f, n;
{
  LINE	*dotp;	/* original cursor line */
  short	doto;	/*		and row */
  int	c;		/* temp char */
  long	size;	/* # of chars to delete */

	/* don't allow this command if we are in read only mode */
	if (curbp->b_mode & MDVIEW)
		return(ReadOnly());

	/* ignore the command if there is a negative argument */
	if (n < 0)
		return(FALSE);

	/* clear the kill buffer if last command wasn't a kill */
	if ((lastflag & CFKILL) == 0)
		KDelete();
	thisflag |= CFKILL; 	/* this command is a kill */

	/* save the current cursor position */
	dotp = curwp->w_dotp;
	doto = curwp->w_doto;

	/* figure out how many characters to give the axe */
	size = 0L;

	/* get us into a word */
	while (inword() == FALSE)
	{
		if (ForwChar(FALSE, 1) == FALSE)
			return(FALSE);
		++size;
	}

	if (n == 0)
	{									/* skip one word, no whitespace! */
		while (inword() == TRUE)
		{
			if (ForwChar(FALSE, 1) == FALSE)
				return(FALSE);
			++size;
		}
	}
	else
	{
		/* skip n words */
		while (n--)
		{
			/* if we are at EOL; skip to the beginning of the next */
			while (curwp->w_doto == llength(curwp->w_dotp))
			{
				if (ForwChar(FALSE, 1) == FALSE)
					return(FALSE);
				++size;
			}
	    
			/* move forward till we are at the end of the word */
			while (inword() == TRUE)
			{
				if (ForwChar(FALSE, 1) == FALSE)
					return(FALSE);
				++size;
			}
	    
			/* if there are more words, skip the interword stuff */
			if (n != 0)
				while (inword() == FALSE)
				{
					if (ForwChar(FALSE, 1) == FALSE)
						return(FALSE);
					++size;
				}
		}
	    
		/* skip whitespace and newlines */
		while	(
						(curwp->w_doto == llength(curwp->w_dotp))
					||
						(c = lgetc(curwp->w_dotp, curwp->w_doto)) == ' '
					||
						c == '\t'
				)
		{
			if (ForwChar(FALSE, 1) == FALSE)
				break;
			++size;
		}
	}

	/* restore the original position and delete the words */
	curwp->w_dotp = dotp;
	curwp->w_doto = doto;
	return(LDelete(size, TRUE));
}

/**
 * Kill backwards by "n" words. Move backwards by the desired number of words,
 * counting the characters. When dot is finally moved to its resting place,
 * fire off the kill command.
 */
int DelBWord(f, n)
  int	f, n;
{
  long	size;

	/* don't allow this command if we are in read only mode */
	if (curbp->b_mode & MDVIEW)
		return(ReadOnly());

	/* ignore the command if there is a nonpositive argument */
	if (n <= 0)
		return(FALSE);

	/* clear the kill buffer if last command wasn't a kill */
	if ((lastflag & CFKILL) == 0)
		KDelete();
	thisflag |= CFKILL; 	/* this command is a kill */

	if (BackChar(FALSE, 1) == FALSE)
		return(FALSE);

	size = 0L;
	while (n--)
	{
		while (inword() == FALSE)
		{
			if (BackChar(FALSE, 1) == FALSE)
				return(FALSE);
			++size;
		}
		while (inword() != FALSE)
		{
			++size;
			if (BackChar(FALSE, 1) == FALSE)
				goto bckdel;
		}
	}
	if (ForwChar(FALSE, 1) == FALSE)
		return(FALSE);
bckdel:
	return(LDelete(size, TRUE));
}

/**
 * Return TRUE if the character at dot is a character that is considered to be
 * part of a word. The word character list is hard coded. Should be settable.
 */
static int near inword()
{
  int	c;

	if (curwp->w_doto == llength(curwp->w_dotp))
		return(FALSE);
	c = lgetc(curwp->w_dotp, curwp->w_doto);
	if (c >= 'a' && c <= 'z')
		return(TRUE);
	if (c >= 'A' && c <= 'Z')
		return(TRUE);
	if (c >= '0' && c <= '9')
		return(TRUE);
	return(FALSE);
}
